/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE:
* 
* CU_Array.cc
* 
* AUTHOR: 
* 
* Stephan R.A. Deibel
* 
* CREATION DATE: 
* 
* 9/25/95
* 
* DESCRIPTION: 
* 
* Simple C++ array class implementation.
* 
* NOTES:
*
* Not yet tested!
*
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
* 
* 25Sep95  Stephan   Created.
* 
*/

#ifdef dos
#include "CUArray.hh"
#else
#include "CU_Array.hh"
#endif
#include <string.h>


/****************************/
/****** PUBLIC METHODS ******/
/****************************/


/****************************************************************************
* FUNCTION:
*
* CUC_Array::CUC_Array()
*
* DESCRIPTION:
*
* Constructor for this class
*
* PARAMETERS:
*
* tUInt4                       -- Size of elements in array
*
*/

CUC_Array::CUC_Array(tUInt4 item_size)
{
  fArray = NULL;
  fNumItems = 0;
  fItemSize = item_size;
  fAllocated = 0;
  fGrowthIncr = 64;
}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::~CUC_Array()
*
* DESCRIPTION:
*
* Destructor for this class
*
*/

CUC_Array::~CUC_Array()
{
  if (fArray != NULL)
    CU_Free(fArray);
#ifdef DEBUG
  fArray = NULL;
  fNumItems = 0;
  fAllocated = 0;
#endif
}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::GetNumItems() const
*
* DESCRIPTION:
*
* Get number of items on the array
*
* RETURNS:
*
* tUInt4                       -- The number
*
*/

tUInt4 CUC_Array::GetNumItems() const
{
  return fNumItems;
}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::SetNumItems()
*
* DESCRIPTION:
*
* Set number of items on the array, blanking any exposed memory
* out to NULL.
*
* PARAMETERS:
*
* tUInt4                       -- The number of items
*
* RETURNS:
*
* tUInt4                       -- New number of items (=input or 0 if
*                                 failed).
*
*/

tUInt4 CUC_Array::SetNumItems(tUInt4 n)
{
  /* Grow if needed */
  if (GrowMemory(n) == 0)
    return 0;

  /* Blank out any exposed memory */
  if (n > fNumItems)
    CU_MemSet((void *) ((tUInt1 *) fArray + fNumItems * fItemSize), 0, 
              (tUInt4) (n - fNumItems) * fItemSize);

  /* Set new size */
  fNumItems = n;

  return n;
}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::SetItem()
*
* DESCRIPTION:
*
* Copies the provided item into the array buffer.
*
* PARAMETERS:
*
* tUInt4                       -- The item number (1 = first; must be
*                                 <= value returned by GetNumItems())
* void *                       -- Buffer space containing item value.
*
*/

void CUC_Array::SetItem(tUInt4 n, void *new_item)
{
  GOODMEM(new_item);
  ASSERT(n <= fNumItems);
  CU_MemCpy((void *)((tUInt1 *) fArray + (n - 1) * fItemSize), 
            new_item, fItemSize);
}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::GetItem() const
*
* DESCRIPTION:
*
* Copies the desired item to the provided buffer.
*
* PARAMETERS:
*
* tUInt4                       -- The item number (1 = first; must be
*                                 <= value returned by GetNumItems())
* void *                       -- Buffer space for return value.
*
*/

void CUC_Array::GetItem(tUInt4 n, void *item) const
{
  GOODMEM(item);
  ASSERT(0 < n && n <= fNumItems);
  CU_MemCpy(item, (void *)((tUInt1 *)fArray + (n - 1) * fItemSize), fItemSize);
}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::GetItemPtr()
*
* DESCRIPTION:
*
* Gets a pointer to the desired item in the array
*
* PARAMETERS:
*
* tUInt4                       -- The item number (1 = first; must be
*                                 <= value returned by GetNumItems())
*
* RETURNS:
*
* void *                       -- Pointer to array data; this is the
*                                 actual array element, which may be
*                                 changed in place as desired as long
*                                 as the array is not changed through
*                                 other CUC_Array() calls; after that,
*                                 the return value becomes invalid!
*
*/

void *CUC_Array::GetItemPtr(tUInt4 n)
{
  ASSERT(0 < n && n <= fNumItems);
  return (void *) ((tUInt1 *) fArray + (n - 1) * fItemSize);
}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::Append()
*
* DESCRIPTION:
*
* Append the given item to the list.
*
* PARAMETERS:
*
* void *                       -- Buffer space containing item value.
*
* RETURNS:
*
* tUInt4                       -- The item's position (0 = memory error)
*
*/

tUInt4 CUC_Array::Append(void *new_item)
{
  return Insert(fNumItems + 1, new_item);
}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::Prepend()
*
* DESCRIPTION:
*
* Prepend the given item to the list.
*
* PARAMETERS:
*
* void *                       -- Buffer space containing item value.
*
* RETURNS:
*
* tUInt4                       -- The item's position (0 = memory error)
*
*/

tUInt4 CUC_Array::Prepend(void *new_item)
{
  return Insert(1, new_item);
}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::Insert()
*
* DESCRIPTION:
*
* Insert the given item into the list, moving items out of the way
* as needed.
*
* PARAMETERS:
*
* tUInt4                       -- Insertion position (must be <= value
*                                 returned from GetNumItems() + 1); 1 = first.
* void *                       -- Buffer space containing item value.
*
* RETURNS:
*
* tUInt4                       -- The item's position (0 = memory error)
*
*/

tUInt4 CUC_Array::Insert(tUInt4 n, void *new_item)
{
  ASSERT(n <= fNumItems + 1);

  /* Grow if needed */
  if (GrowMemory(fNumItems + 1) == 0)
    return 0;

  /* Move items out of the way */
  if (n <= fNumItems)
    CU_MemMove((void *)((tUInt1 *)fArray + n * fItemSize), 
               (void *)((tUInt1 *)fArray + (n - 1) * fItemSize), 
               (fNumItems - n + 1) * fItemSize);

  /* Place item */
  fNumItems = fNumItems + 1;
  SetItem(n, new_item);

  /* Return position */
  return n;
}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::Remove()
*
* DESCRIPTION:
*
* Remove the given item(s) from the list, moving items to fill the gap
* as needed.
*
* PARAMETERS:
*
* tUInt4                       -- Deletion position (must be <= value
*                                 returned from GetNumItems())
* tUInt4                       -- Number of items to remove (may be 0)
*
* RETURNS:
*
* void
*
*/

void CUC_Array::Remove(tUInt4 pos, tUInt4 num)
{
  ASSERT(pos <= fNumItems + 1);

  /* Check degenerate case */
  if (num == 0)
    return;

  /* Move items to close the gap */
  CU_MemMove((void *)((tUInt1 *) fArray + (pos - 1) * fItemSize), 
             (void *)((tUInt1 *) fArray + ((pos - 1) + num) * fItemSize), 
             (fNumItems - (pos + num) + 1) * fItemSize);

  /* Decrement num items */
  fNumItems = fNumItems - num;
}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::Copy() const
*
* DESCRIPTION:
*
* Allocates and returns a copy of the array.
*
* RETURNS:
*
* CUC_Array *                   -- A pointer to the copy
*			        -- NULL if fails
*
*/

CUC_Array *CUC_Array::Copy() const
{
  CUC_Array *ret_array = NULL;
  void *tmp_item = NULL;

  /* Create & Initialize new array */
  ret_array = new CUC_Array( fItemSize );
  if (ret_array == NULL) {
    return( NULL );
  }

  ret_array->SetGrowthIncr( fGrowthIncr );
  if ( !ret_array->SetNumItems(fNumItems) ) {
    delete ret_array;
    return( NULL );
  }

  /* Copy the Array data */
  tmp_item = CU_Malloc( fItemSize );
  for (tUInt4 i = 1; i <= fNumItems; i ++) {
    GetItem( i, tmp_item );
    ret_array->SetItem( i, tmp_item );
  }
  
  delete tmp_item;
  
  /* Done */
  return( ret_array );

}


/****************************************************************************
* FUNCTION:
*
* CUC_Array::SetGrowthIncr()
*
* DESCRIPTION:
*
* Set the memory growth increment used when array's internal
* memory usage is automatically expanded.
*
* The array's implementation internally allocates memory on an
* as-needed basis.  Instead of reallocating memory item-by-item
* (which is slow), memory is allocated increments of a set block size.
* The default value for the growth block size is 64, but should be 
* increased for better performance if Insert(), Append(), or 
* Prepend() are called repeatedly many times.
*
* It is best to use SetNumItems() to completely avoid repeatedly
* reallocating memory.  If this is done, the growth increment
* set here is irrelevant.
*
* PARAMETERS:
*
* tUInt4                       -- New increment size, in bytes
*
*/

void CUC_Array::SetGrowthIncr(tUInt4 n)
{
  fGrowthIncr = n;
}


/****************************/
/***** PRIVATE METHODS ******/
/****************************/

/****************************************************************************
* FUNCTION:
*
* CUC_Array::GrowMemory()
*
* DESCRIPTION:
*
* Grow the internal memory allocation for this array to make room for
* at least the given number of items.
*
* Memory will be reallocated using block sizes of 64 by default,
* or whatever value was last set with SetGrowthIncr().
*
* PARAMETERS:
*
* tUInt4                       -- Number of items to accomodate
*
* RETURNS:
*
* tUInt4                       -- Total memory allocated after call
*                                 (0 indicates memory error, leaving
*                                 allocation at previous level)
*
*/

tUInt4 CUC_Array::GrowMemory(tUInt4 num_items)
{
  tUInt4 space_needed;
  tUInt4 total_length;

  if (num_items * fItemSize > fAllocated) {
    /* Figure increment needed */
    space_needed = num_items * fItemSize;
    total_length = space_needed + (fGrowthIncr - (space_needed % fGrowthIncr));
/*
  tUInt4 incr;
    incr = ((num_items * fItemSize) - fAllocated) / fGrowthIncr;
    incr = (fGrowthIncr * incr) + 1;
    */

    /* Allocated new array */
    if (fArray == NULL) {
      fArray = (void *) CU_Malloc(total_length);
      if (fArray == NULL)
        return 0;
    }

    /* Reallocated existing array */
    else {
      void *new_mem;

      new_mem = CU_Realloc(fArray, total_length);
      if (new_mem == NULL)
        return 0;
      else
        fArray = new_mem;
    }

    /* Increment recorded size */
    fAllocated = total_length;
  }

  return fAllocated;
}


/* End of CU_Array.cc */

