/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE: CU_CStream.c
* 
* AUTHOR: 
* 
* Jaime Requena
*
* CREATION DATE: 
*
* May 1997
* 
* DESCRIPTION: 
* 
* Streams implementations with slightly better error checking
* than the usual.
* 
* NOTES: 
* 
* 
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
* 
* $Log: CU_CStream.c,v $
* Revision 1.3  1998/07/15 05:20:18  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
 * Revision 1.2  1997/11/10  16:17:11  jrequena
 * made minor change to write string
 *
 * Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
 * Converting to CVS
 *
* Revision 1.1  1997/05/02 21:14:07  jrequena
* Initial revision
*
*
*/

#include "CU_CStream.h"
#include "CU_Debug.h"
#include <string.h>


/***************************************************************************
* 
* FUNCTION: 
* 
* CU_InitStream()
* 
* DESCRIPTION: 
* 
* Initializes CUt_Stream structure.
* 
* PARAMETERS: 
* 
* CUt_Stream*              -- The stream
* 
* RETURNS: 
* 
* NOTES:
*
* 
*/

void
CU_InitStream(CUt_Stream* stream)
{
  if (stream != NULL) {
    stream->name = NULL;
    stream->file = NULL;
  }
}


/***************************************************************************
* 
* FUNCTION: 
* 
* open()
* 
* DESCRIPTION: 
* 
* Open a file into the file stream; closes any previously open file
* 
* PARAMETERS: 
* 
* const char *               -- The file name
* 
* RETURNS: 
* 
* NOTES:
*
* No file is opened if there is any kind of error
* 
*/
void
CU_OpenStream( CUt_Environment *ev, CUt_Stream* stream, const char *filename )
{
  ASSERT(stream != NULL);
  ASSERT(filename != NULL);

  // Set file name
  stream->name = (char *) CU_Malloc(strlen(filename) + 1);
  if (stream->name == NULL)
    ev->error = kOutOfMemory;

  // Open file
  strcpy( stream->name, filename );
  stream->file = fopen(stream->name, "rb");  
  if (stream->file == NULL) 
    ev->error = kFileIOError;

}
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_CloseStream()
* 
* DESCRIPTION: 
* 
* Close the file associated with this stream, if any
* 
* PARAMETERS: 
* 
* None.
* 
* RETURNS: 
* 
* void.
* 
*/

void 
CU_CloseStream(CUt_Stream *stream)
{
  if (stream->file != NULL) {
    fclose( stream->file );
    stream->file = NULL;
  }
  if (stream->name != NULL) {
    CU_Free( stream->name );
    stream->name = NULL;
  }
}
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_GetStream()
* 
* DESCRIPTION: 
* 
* Get a byte from the file
* 
* PARAMETERS: 
* 
* None.
* 
* RETURNS: 
* 
* tInt2                       -- The byte, in the form of a tInt2;
*                                EOF on failure or at end of file.
* 
*/

tInt2 
CU_GetStream(const CUt_Stream *stream)
{
  ASSERT(stream->file != NULL);
  return fgetc(stream->file);
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CU_ReadStream()
* 
* DESCRIPTION:
* 
* Read the specified number of bytes from the file
* 
* PARAMETERS:
* 
* void *                      -- Buffer for returned bytes
* tUInt4                      -- Number of bytes to read
* 
* RETURNS: 
* 
* tUInt4                      -- Number of bytes actually read
* 
*/

tUInt4
CU_ReadStream(const CUt_Stream *stream, void *dest,tUInt4 n)
{
  ASSERT(stream->file != NULL);
  return fread(dest, 1, n, stream->file);
}
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_SeedgStream()
* 
* DESCRIPTION: 
* 
* Seek to the given position in the file (0 = start)
* 
* PARAMETERS: 
* 
* tUInt4                        -- The position
* 
* RETURNS: 
* 
* tError                        -- kNoError or kFileIOError
* 
*/

tError 
CU_SeekgStream(const CUt_Stream *stream, tUInt4 pos)
{
  if (fseek( stream->file, pos, SEEK_SET ) != 0)
    return kFileIOError;
  else
    return kNoError;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CU_TellgStream()
* 
* DESCRIPTION: 
* 
* Get the current location in the file
* 
* PARAMETERS: 
* 
* None.
* 
* RETURNS: 
* 
* tUInt4                     -- The current position in file
* 
*/

tUInt4 
CU_TellgStream(const CUt_Stream *stream)
{
  return( ftell(stream->file) );
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_StreamGetString()
* 
* DESCRIPTION: 
* 
* Get string from stream.
* 
* PARAMETERS: 
* 
* None.
* 
* RETURNS: 
* 
* 
*/


void CU_StreamGetString(CUt_Environment* ev, const CUt_Stream *stream, 
			char* string, tUInt2 length)
{
  char* error;

  ASSERT(stream != NULL);
  ASSERT(stream->file != NULL);
  ASSERT(string != NULL);

  error = fgets(string, length, stream->file);
  if (error == NULL)
    ev->error = kFileIOError;
  
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CU_StreamWriteString()
* 
* DESCRIPTION: 
* 
* Write string to stream.
* 
* PARAMETERS: 
* 
* None.
* 
* RETURNS: 
* 
* 
*/

void
CU_StreamWriteString(CUt_Environment* ev, const CUt_Stream *stream, const char* string)
{
  tInt2 error = 0;

  ASSERT(stream != NULL);
  ASSERT(stream->file != NULL);

  error = fputs(string, stream->file);
  if (error < 0)
    ev->error = kFileIOError;

}



/* end of CU_Stream.cc */

