/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE: CU_Debug.c
* 
* AUTHOR: 
*
* Stephan R.A. Deibel
* YongJoon Lee
* John Ehresman
* 
* CREATION DATE: 
* 
* 8/1/93
*
* VERSION: 
* 
* 6/13/94
*
* DESCRIPTION: 
* 
* Basic debugging utilities for cross-platform development that replace
* and augments some of the standard C "assert" code with generally
* safer alternatives.
*
* For usage instructions, read the embedded comments in the header file.
* 
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
* 
* 3/11/94  Stephan   Added header comments and fixed some formatting problems
*                    in the Mac implementation's output.
* 4/4/94   YJLee     Fixed the printf output to shorts, because the Mac
*                    debugger prints out garbage long values.
* 4/5/94   YJLee     Changed printf output to short ints for line nums to 
*                    fix Mac problem.
* 3/15/94  Stephan   Fixed text for UNFINISHED call -- was incorrect.
* 3/12/94  Stephan   Fixed formatting of outputted messages for Mac.
* 6/13/94  Stephan   Cosmetic / commenting
* 7/4/94   Stephan   Added UNTESTED and DEBUGMSG calls.
* 8/25/95  Stephan   Added GOODMEM call.
* 
*/

#ifdef dos
#include "CUBasics.h"
#include "CUFail.h"
#include "CUDebug.h"
#else
#include "CU_Basics.h"
#include "CU_Failure.h"
#include "CU_Debug.h"
#include "CU_Text.h"
#endif /*dos */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef THINK_C
#include <pascal.h>
#endif /* THINK_C */

#ifdef unix
#include <signal.h>
#include <unistd.h>
#endif /* GNUC */

#ifdef MS_WIN
#include <windows.h>
#endif /* MS_WIN */

#ifdef CUk_DebugMalloc
 #undef CU_Malloc
 #undef CU_Free
 #undef CU_Calloc
 #undef CU_Realloc
 #undef CU_MallocSet
#endif /* CUk_DebugMalloc */


/**************************************************************************/
/* GLOBALS */

/* Does this *really* need to be a global???? [probably -- jpe] */
/*#if defined( THINK_C ) || defined( __MWERKS__ ) || defined(MS_WIN)*/
static char DBg_DebugStr[256];
/*#endif*/

static  CUt_DBMallocInfo* CUg_DBMallocInfoList = NULL;


#ifdef DEBUG // See CU_Debug.h

/****************************************************************************
* FUNCTION:
*
* _DB_Assert
*
* DESCRIPTION:
*
* Enter resident debugger and print an message indicating that an ASSERT
* macro has encountered an invalid condition.  The file and line number
* of the ASSERT macro are given.
*
* PARAMETERS:
*
* char *          -- The file name in which the error occurred
* unsigned int    -- The line number on which the error occurred.
*
*/

void 
_DB_Assert(const char *file_name, unsigned int line_num)
{
#if defined(THINK_C) || (defined(__MWERKS__) && !defined(MS_WIN))
  sprintf(DBg_DebugStr, "ASSERT FAILED:  %s, line %hu", file_name,
          line_num); 
  CU_CStringToPString( DBg_DebugStr );
  DebugStr ((unsigned char *)DBg_DebugStr);
  ExitToShell();
#endif /* THINK_C*/

#ifdef unix
  /* Print error message to standard error */
  fprintf(stderr, "ASSERT FAILED:  %s, line %hu\n", file_name,
          line_num); 

  /* Send an interrupt signal to the current process */
  kill(getpid(), SIGINT);
#endif /* unix */

#ifdef MS_WIN
  int tmp_int;
  char tmp_str[20];
  char dbg_str1[20];
  char dbg_str2[12];  
 
  dbg_str1[0] = 'A';
  dbg_str1[1] = 'S';
  dbg_str1[2] = 'S';
  dbg_str1[3] = 'E';
  dbg_str1[4] = 'R';
  dbg_str1[5] = 'T';
  dbg_str1[6] = ':';
  dbg_str1[7] = ' ';
  dbg_str1[8] = 0;
  
  dbg_str2[0] = ' ';
  dbg_str2[1] = 0;

#if defined(MSVC) || defined(MS_VC)
  _itoa(line_num, tmp_str, 10);
#else
  itoa(line_num, tmp_str, 10);
#endif /* MSVC || MS_VC */

  memcpy( DBg_DebugStr, (void *)dbg_str1, strlen(dbg_str1) );
  tmp_int = strlen(dbg_str1);
  memcpy( &DBg_DebugStr[tmp_int], (void *)file_name, strlen(file_name) );
  tmp_int += strlen(file_name);
  memcpy( &DBg_DebugStr[tmp_int], (void *)dbg_str2, strlen(dbg_str2) );
  tmp_int += strlen(dbg_str2);  
  memcpy( &DBg_DebugStr[tmp_int], (void *)tmp_str, strlen(tmp_str) );
  tmp_int += strlen(tmp_str);
  DBg_DebugStr[tmp_int] = 0;    
  
  /*  
  sprintf(DBg_DebugStr, "ASSERT FAILED:  %s, line %hu", file_name,
          line_num); 
  */
  FatalAppExit(0, DBg_DebugStr);  
#endif /* MS_WIN */
} /* Assert */


/****************************************************************************
* FUNCTION:
*
* _DB_Unimplemented
*
* DESCRIPTION:
*
* Enter resident debugger and print an message indicating that an
* UNIMPLEMENTED macro has been encountered.
*
* PARAMETERS:
*
* char *          -- The file name in which the error occurred
* unsigned int    -- The line number on which the error occurred.
*
*/

void  
_DB_Unimplemented(const char *file_name, unsigned int line_num)
{
#if defined(THINK_C) || (defined(__MWERKS__) && !defined(MS_WIN))
  sprintf(DBg_DebugStr, "UNIMPLEMENTED:  %s, line %hu",
          file_name, line_num); 
  CU_CStringToPString( DBg_DebugStr );
  DebugStr ((unsigned char *)DBg_DebugStr);
  ExitToShell();
#endif /* THINK_C*/

#ifdef unix
  fprintf(stderr, "UNIMPLEMENTED:  %s, line %hu\n",
          file_name, line_num); 

  /* send an interrupt signal to the current process */
  kill(getpid(), SIGINT);
#endif /* unix */

#ifdef MS_WIN
  int tmp_int;
  char tmp_str[20];
  char dbg_str1[20];
  char dbg_str2[12];  
 
  dbg_str1[0] = 'U';
  dbg_str1[1] = 'N';
  dbg_str1[2] = 'I';
  dbg_str1[3] = 'M';
  dbg_str1[4] = 'P';
  dbg_str1[5] = 'L';
  dbg_str1[6] = 'E';
  dbg_str1[7] = 'M';
  dbg_str1[8] = 'E';
  dbg_str1[9] = 'N';
  dbg_str1[10] = 'T';
  dbg_str1[11] = 'E';
  dbg_str1[12] = 'D';
  dbg_str1[13] = ':';
  dbg_str1[14] = ' ';
  dbg_str1[15] = 0;
  
  dbg_str2[0] = ' ';
  dbg_str2[1] = 0;
  
#if defined(MSVC) || defined(MS_VC)
  _itoa(line_num, tmp_str, 10);
#else
  itoa(line_num, tmp_str, 10);
#endif /* MSVC || MS_VC */

  memcpy( DBg_DebugStr, (void *)dbg_str1, strlen(dbg_str1) );
  tmp_int = strlen(dbg_str1);
  memcpy( &DBg_DebugStr[tmp_int], (void *)file_name, strlen(file_name) );
  tmp_int += strlen(file_name);
  memcpy( &DBg_DebugStr[tmp_int], (void *)dbg_str2, strlen(dbg_str2) );
  tmp_int += strlen(dbg_str2);  
  memcpy( &DBg_DebugStr[tmp_int], (void *)tmp_str, strlen(tmp_str) );
  tmp_int += strlen(tmp_str);
  DBg_DebugStr[tmp_int] = 0;    
  
  FatalAppExit(0, DBg_DebugStr);

#endif /* MS_WIN */

} /* Unimplemented */


/****************************************************************************
* FUNCTION:
*
* _DB_Unfinished
*
* DESCRIPTION:
*
* Similar to UNIMPLEMENTED except it posts a message to stdout and will
* not halt the program afterwards.
*
* PARAMETERS:
*
* char *          -- The file name in which the error occurred
* unsigned int    -- The line number on which the error occurred.
*
*/

void  
_DB_Unfinished(const char *file_name, unsigned int line_num)
{
#if defined(THINK_C) || (defined(__MWERKS__) && !defined(MS_WIN))
  sprintf(DBg_DebugStr, "Unfinished:  %s, line %hu\n", 
          file_name, line_num);  
  fprintf(stderr, "%s", DBg_DebugStr);
#endif /* THINK_C*/

#ifdef unix
  fprintf(stderr, "Unfinished:  %s, line %hu\n",
          file_name, line_num); 
#endif /* unix */

#ifdef MS_WIN
  int tmp_int;
  char tmp_str[20];
  char dbg_str1[20];
  char dbg_str2[12];  

  dbg_str1[0] = 'U';
  dbg_str1[1] = 'n';
  dbg_str1[2] = 'f';
  dbg_str1[3] = 'i';
  dbg_str1[4] = 'n';
  dbg_str1[5] = 'i';
  dbg_str1[6] = 's';
  dbg_str1[7] = 'h';
  dbg_str1[8] = 'e';
  dbg_str1[9] = 'd';
  dbg_str1[10] = ':';
  dbg_str1[11] = ' ';
  dbg_str1[12] = 0;
  
  dbg_str2[0] = ' ';
  dbg_str2[1] = 0;
  
#if defined(MSVC) || defined(MS_VC)
  _itoa(line_num, tmp_str, 10);
#else
  itoa(line_num, tmp_str, 10);
#endif /* MSVC || MS_VC */

  memcpy( DBg_DebugStr, (void *)dbg_str1, strlen(dbg_str1) );
  tmp_int = strlen(dbg_str1);
  memcpy( &DBg_DebugStr[tmp_int], (void *)file_name, strlen(file_name) );
  tmp_int += strlen(file_name);
  memcpy( &DBg_DebugStr[tmp_int], (void *)dbg_str2, strlen(dbg_str2) );
  tmp_int += strlen(dbg_str2);  
  memcpy( &DBg_DebugStr[tmp_int], (void *)tmp_str, strlen(tmp_str) );
  tmp_int += strlen(tmp_str);
  DBg_DebugStr[tmp_int] = '\n';    
  DBg_DebugStr[++tmp_int] = '\r';    
  DBg_DebugStr[++tmp_int] = 0;    
  
  OutputDebugString(DBg_DebugStr);
  
#endif /* MS_WIN */

} /* Unfinished */


/****************************************************************************
* FUNCTION:
*
* _DB_Untested
*
* DESCRIPTION:
*
* Similar to UNFINISHED except it conveys different semantics.
*
* PARAMETERS:
*
* char *          -- The file name in which the error occurred
* unsigned int    -- The line number on which the error occurred.
*
*/

void 
_DB_Untested(const char *file_name, unsigned int line_num)
{
#if defined( THINK_C ) || (defined(__MWERKS__) && !defined(MS_WIN))
  sprintf(DBg_DebugStr, "Untested:  %s, line %hu\n", 
          file_name, line_num);  
  fprintf(stderr, "%s", DBg_DebugStr);
#endif /* THINK_C*/

#ifdef unix
  fprintf(stderr, "Untested:  %s, line %hu\n",
          file_name, line_num); 
#endif /* unix */

#ifdef MS_WIN 
  int tmp_int;
  char tmp_str[20];
  char dbg_str1[20];
  char dbg_str2[12];  
 
  dbg_str1[0] = 'U';
  dbg_str1[1] = 'n';
  dbg_str1[2] = 't';
  dbg_str1[3] = 'e';
  dbg_str1[4] = 's';
  dbg_str1[5] = 't';
  dbg_str1[6] = 'e';
  dbg_str1[7] = 'd';
  dbg_str1[8] = ':';
  dbg_str1[9] = 0;

  dbg_str2[0] = ' ';
  dbg_str2[1] = 0;
  
#if defined(MSVC) || defined(MS_VC)
  _itoa(line_num, tmp_str, 10);
#else
  itoa(line_num, tmp_str, 10);
#endif /* MSVC || MS_VC */

  memcpy( DBg_DebugStr, (void *)dbg_str1, strlen(dbg_str1) );
  tmp_int = strlen(dbg_str1);
  memcpy( &DBg_DebugStr[tmp_int], (void *)file_name, strlen(file_name) );
  tmp_int += strlen(file_name);
  memcpy( &DBg_DebugStr[tmp_int], (void *)dbg_str2, strlen(dbg_str2) );
  tmp_int += strlen(dbg_str2);  
  memcpy( &DBg_DebugStr[tmp_int], (void *)tmp_str, strlen(tmp_str) );
  tmp_int += strlen(tmp_str);
  DBg_DebugStr[tmp_int] = '\n';    
  DBg_DebugStr[++tmp_int] = '\r';    
  DBg_DebugStr[++tmp_int] = 0;    
  
  OutputDebugString(DBg_DebugStr);
#endif /* MS_WIN */

} /* Untested */


/****************************************************************************
* FUNCTION:
*
* _DB_DebugMsg
*
* DESCRIPTION:
*
* Print a message (along with file name and line number) to stderr.
*
* PARAMETERS:
*
* char *          -- The file name in which the message was generated
* unsigned int    -- The line number at which the message was generated
* char *          -- The message
*
*/

void 
_DB_DebugMsg(const char *file_name, unsigned int line_num, const char *msg)
{
#if defined( THINK_C ) || (defined(__MWERKS__) && !defined(MS_WIN))
  sprintf(DBg_DebugStr, "%s: %s, line %hu\n", msg, file_name, line_num);  
  fprintf(stderr, "%s", DBg_DebugStr);
#endif /* THINK_C*/

#ifdef unix
  fprintf(stderr, "%s: %s, line %hu\n", msg, file_name, line_num); 
#endif /* unix */

#ifdef MS_WIN 
  int tmp_int;
  char tmp_str[20];
  const char *dbg_str1;
  char dbg_str2[12];

  dbg_str1 = msg;

  dbg_str2[0] = ' ';
  dbg_str2[1] = 0;

#if defined(MSVC) || defined(MS_VC)
  _itoa(line_num, tmp_str, 10);
#else
  itoa(line_num, tmp_str, 10);
#endif /* MSVC || MS_VC */

  memcpy( DBg_DebugStr, (void *)dbg_str1, strlen(dbg_str1) );
  tmp_int = strlen(dbg_str1);
  memcpy( &DBg_DebugStr[tmp_int], (void *)file_name, strlen(file_name) );
  tmp_int += strlen(file_name);
  memcpy( &DBg_DebugStr[tmp_int], (void *)dbg_str2, strlen(dbg_str2) );
  tmp_int += strlen(dbg_str2);  
  memcpy( &DBg_DebugStr[tmp_int], (void *)tmp_str, strlen(tmp_str) );
  tmp_int += strlen(tmp_str);
  DBg_DebugStr[tmp_int] = '\n';    
  DBg_DebugStr[++tmp_int] = '\r';    
  DBg_DebugStr[++tmp_int] = 0;    
  
  OutputDebugString(DBg_DebugStr);
#endif /* MS_WIN */

} /* DebugMsg */


/****************************************************************************
* FUNCTION:
*
* _DB_GoodMem
*
* DESCRIPTION:
*
* Checks if the given pointer points to good memory; prints an error
* message and halts if not.
*
* PARAMETERS:
*
* char *          -- The file name in which the message was generated
* unsigned int    -- The line number at which the message was generated
* void *          -- The pointer to check
*
*/

void  
_DB_GoodMem(const char *file_name, unsigned int line_num, const void *ptr)
{
#if defined( THINK_C ) || (defined(__MWERKS__) && !defined(MS_WIN))
  if (ptr != NULL)
    return;

  sprintf(DBg_DebugStr, "PTR TO BAD MEMORY:  %s, line %hu",
          file_name, line_num); 
  CU_CStringToPString( DBg_DebugStr );
  DebugStr ((unsigned char *)DBg_DebugStr);
  ExitToShell();
#endif /* THINK_C*/

#ifdef unix
  if (ptr != NULL)
    return;

  fprintf(stderr, "PTR TO BAD MEMORY:  %s, line %hu\n",
          file_name, line_num); 

  /* send an interrupt signal to the current process */
  kill(getpid(), SIGINT);
#endif /* unix */

#ifdef MS_WIN
  /* THIS IS NUTS!  See commented out section below ... what is wrong with */
  /* that?!? */
  int tmp_int;
  char tmp_str[20];
  char dbg_str1[20] = "BADMPTR: ";
  char dbg_str2[12] = ", ln ";  
 
  if (ptr != NULL)
    return;

  dbg_str1[0] = 'P';
  dbg_str1[1] = 'T';
  dbg_str1[2] = 'R';
  dbg_str1[3] = ' ';
  dbg_str1[4] = 'T';
  dbg_str1[5] = 'O';
  dbg_str1[6] = ' ';
  dbg_str1[7] = 'B';
  dbg_str1[8] = 'A';
  dbg_str1[9] = 'D';
  dbg_str1[10] = ' ';
  dbg_str1[11] = 'M';
  dbg_str1[12] = 'E';
  dbg_str1[13] = 'M';
  dbg_str1[14] = 'O';
  dbg_str1[15] = 'R';
  dbg_str1[16] = 'Y';
  dbg_str1[17] = ':';
  dbg_str1[18] = ' ';
  dbg_str1[19] = 0;
  
  dbg_str2[0] = ' ';
  dbg_str2[1] = 0;
  
#if defined(MSVC) || defined(MS_VC)
  _itoa(line_num, tmp_str, 10);
#else
  itoa(line_num, tmp_str, 10);
#endif /* MSVC || MS_VC */

  memcpy( DBg_DebugStr, (void *)dbg_str1, strlen(dbg_str1) );
  tmp_int = strlen(dbg_str1);
  memcpy( &DBg_DebugStr[tmp_int], (void *)file_name, strlen(file_name) );
  tmp_int += strlen(file_name);
  memcpy( &DBg_DebugStr[tmp_int], (void *)dbg_str2, strlen(dbg_str2) );
  tmp_int += strlen(dbg_str2);  
  memcpy( &DBg_DebugStr[tmp_int], (void *)tmp_str, strlen(tmp_str) );
  tmp_int += strlen(tmp_str);
  DBg_DebugStr[tmp_int] = 0;    
  
  FatalAppExit(0, DBg_DebugStr);

/*
  Stephan's solution; what's wrong with this?!? :
  char tmp_str[20] = "";

  if (ptr != NULL)
    return;

  _itoa(line_num, tmp_str, 10);

  strcpy(DBg_DebugStr, "PTR TO BAD MEMORY: ");
  strcat(DBg_DebugStr, file_name);
  strcat(DBg_DebugStr, ", line ");
  strcat(DBg_DebugStr, tmp_str);
  strcat(DBg_DebugStr, "\n\r");

  FatalAppExit(0, DBg_DebugStr);
*/
#endif /* MS_WIN */
} /* GoodMem */

#endif // DEBUG


/****************************************************************************
* FUNCTION:
*
* _DB_Malloc
*
* DESCRIPTION:
*
* Performs a malloc and stores the location of the caller in a linked list
*
* PARAMETERS:
*
* char *          -- The file name in which the message was generated
* unsigned int    -- The line number at which the message was generated
* tUInt4 size     -- Size of memory to malloc
*
* NOTE:
*
* This is to be used only in debugging mode, so we depend on GOODMEM stopping
* the program when passed a NULL ptr.  Assumes that file_name is a ptr to
* a static string and will not disappear
*
*/

void* 
_DB_Malloc(const char *file_name, unsigned int line_num, unsigned long size)
{
  void* ptr;
  CUt_DBMallocInfo* info;

  /* Allocate the memory the caller wants */
  ptr = CU_Malloc(size);
  GOODMEM(ptr);

  /* Allocate a new info structure */
  info = (CUt_DBMallocInfo*) CU_Malloc(sizeof(CUt_DBMallocInfo));
  GOODMEM(info);

  /* Fill in the info structure */
  info->file_name = file_name;
  info->line_num = line_num;
  info->size = size;
  info->ptr = ptr;

  /* Put info in linked list */
  info->next = CUg_DBMallocInfoList;
  CUg_DBMallocInfoList = info;

  return ptr;    
} /* _DB_Malloc */


/****************************************************************************
* FUNCTION:
*
* _DB_Free
*
* DESCRIPTION:
*
* Frees memory and removes info about malloc from linked list
*
* PARAMETERS:
*
* char *          -- The file name in which the message was generated
* unsigned int    -- The line number at which the message was generated
* void* ptr       -- Pointer allocated with _DB_Malloc to free
*
* NOTE:
*
* This is to be used only in debugging mode, so we depend on GOODMEM stopping
* the program when passed a NULL ptr.
*
*/

void 
_DB_Free(const char *file_name, unsigned int line_num, void* ptr)
{
  CUt_DBMallocInfo* info, *prev_info;

  GOODMEM(file_name);

  /* Find ptr in linked list */
  for ( info = CUg_DBMallocInfoList, prev_info = NULL; info != NULL;
        prev_info = info, info = info->next ) {
    /* If this is the ptr's info block, free the ptr and the block */
    if ( info->ptr == ptr ) {
      if ( prev_info != NULL )
        prev_info->next = info->next;
      else
        CUg_DBMallocInfoList = info->next;
      CU_Free(ptr);
      CU_Free(info);
      return;
    }
  }

  /* If we got to here, we didn't find the ptr in our list */
  sprintf(DBg_DebugStr, "Attempt to free unknown pointer: %s, line %d\n",
          file_name, line_num);
  DEBUGMSG(DBg_DebugStr);
  
  /* Try to free ptr anyway [jpe ???] */
  CU_Free(ptr);
} /* _DB_Free */


/****************************************************************************
* FUNCTION:
*
* _DB_Malloc
*
* DESCRIPTION:
*
* Performs a malloc and stores the location of the caller in a linked list
*
* PARAMETERS:
*
* char *          -- The file name in which the message was generated
* unsigned int    -- The line number at which the message was generated
* void* ptr       -- Current ptr to the memory block
* tUInt4 size     -- Size of memory to allocate
*
* NOTE:
*
* This is to be used only in debugging mode, so we depend on GOODMEM stopping
* the program when passed a NULL ptr.  Assumes that file_name is a ptr to
* a static string and will not disappear
*
*/

void* 
_DB_Realloc(const char * file_name, tUInt4 line_num, void* ptr, tUInt4 size)
{
  void* new_ptr;
  CUt_DBMallocInfo* info;

  /* Find ptr in linked list */
  for ( info = CUg_DBMallocInfoList; info != NULL; info = info->next ) {
    /* If this is the ptr's info block, break out of for loop */
    if ( info->ptr == ptr )
      break;
  }

  /* If didn't find a info block, print warning and allocate a new one */
  if ( info == NULL ) {
    sprintf(DBg_DebugStr, "Attempt to realloc unknown pointer: %s, line %ld",
	    file_name, line_num);
    DEBUGMSG(DBg_DebugStr);

    info = (CUt_DBMallocInfo*) CU_Malloc(sizeof(CUt_DBMallocInfo));
    GOODMEM(info);
    info->next = CUg_DBMallocInfoList;
    CUg_DBMallocInfoList = info;
  }

  /* Call non-debug function */
  new_ptr = CU_Realloc(ptr, size);
  GOODMEM(new_ptr);

  /* Fill in the info structure */
  info->file_name = file_name;
  info->line_num = line_num;
  info->size = size;
  info->ptr = new_ptr;

  return new_ptr;
}


/****************************************************************************
* FUNCTION:
*
* _DB_Calloc
*
* DESCRIPTION:
*
* Performs a calloc and stores the location of the caller in a linked list
*
* PARAMETERS:
*
* char *          -- The file name in which the message was generated
* unsigned int    -- The line number at which the message was generated
* tUInt4 num      -- Number of elements to allocate
* tUInt4 size     -- Size of each element
*
* NOTE:
*
* This is to be used only in debugging mode, so we depend on GOODMEM stopping
* the program when passed a NULL ptr.  Assumes that file_name is a ptr to
* a static string and will not disappear
*
*/

void* 
_DB_Calloc(const char * file_name, tUInt4 line_num, tUInt4 num, tUInt4 size)
{
  void* ptr;
  CUt_DBMallocInfo* info;

  /* Call non-debug function */
  ptr = CU_Calloc(num, size);

  GOODMEM(ptr);

  /* Allocate a new info structure */
  info = (CUt_DBMallocInfo*) CU_Malloc(sizeof(CUt_DBMallocInfo));
  GOODMEM(info);

  /* Fill in the info structure */
  info->file_name = file_name;
  info->line_num = line_num;
  info->size = num*size;
  info->ptr = ptr;

  /* Put info in linked list */
  info->next = CUg_DBMallocInfoList;
  CUg_DBMallocInfoList = info;

  return ptr;
}



/****************************************************************************
* FUNCTION:
*
* _DB_MallocSet
*
* DESCRIPTION:
*
* Performs a mallocset and stores the location of the caller in a linked list
*
* PARAMETERS:
*
* char *          -- The file name in which the message was generated
* unsigned int    -- The line number at which the message was generated
* tUInt4 size     -- Size of memory to malloc
* tUInt1 fill     -- Value to fill memory with
*
* NOTE:
*
* This is to be used only in debugging mode, so we depend on GOODMEM stopping
* the program when passed a NULL ptr.  Assumes that file_name is a ptr to
* a static string and will not disappear
*
*/

void* 
_DB_MallocSet(const char * file_name, tUInt4 line_num, tUInt4 size, tUInt1 fill)
{
  void* ptr;
  CUt_DBMallocInfo* info;

  /* Call normal function */
  ptr = CU_MallocSet(size, fill);
  GOODMEM(ptr);

  /* Allocate a new info structure */
  info = (CUt_DBMallocInfo*) CU_Malloc(sizeof(CUt_DBMallocInfo));
  GOODMEM(info);

  /* Fill in the info structure */
  info->file_name = file_name;
  info->line_num = line_num;
  info->size = size;
  info->ptr = ptr;

  /* Put info in linked list */
  info->next = CUg_DBMallocInfoList;
  CUg_DBMallocInfoList = info;

  return ptr;
}


/****************************************************************************
* FUNCTION:
*
* CU_DumpMallocInfo
*
* DESCRIPTION:
*
* Lists all current memory allocations in the named file
*
* PARAMETERS:
*
* const char *          -- The name of the file to dump the info to
*
* RETURNS:
*
* tError -- error code: kNoError or kFileIOError
*
*/

tError 
CU_DumpDBMallocInfo(const char* file_name)
{
  FILE* fp;
  CUt_DBMallocInfo* info;

  /* Error check */
  GOODMEM(file_name);
  
  /* Open file and write info */
  fp = fopen(file_name, "w");
  if ( fp == NULL )
    return kFileIOError;

  for ( info = CUg_DBMallocInfoList; info != NULL; info = info->next )
    fprintf(fp, "%ld bytes allocated at %s, line %d\n", info->size,
            info->file_name, info->line_num);

  fclose(fp);

  return kNoError;
}


/* end of CU_Debug.c */
