/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE: CU_Hash.c
* 
* AUTHOR: 
*
* YongJoon Lee
* John Ehresman
* Based on work of John K. Ousterhaut
* 
* CREATION DATE: 
*
* April, 1995
*
* DESCRIPTION: 
*
* Implementation of in-memory hash tables
* 
* NOTES: 
*
* No changes to the implementation, just changed names to conform to
* the Arachne programming conventions.
* 
* Based on internal tcl routines v. 7.4b2.
* The original copyright is as follows:
*   Copyright (c) 1987-1994 The Regents of the University of California.
*   Copyright (c) 1994 Sun Microsystems, Inc.
*
* See the file "tcl_license.terms" for information on usage and
* redistribution of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*
* Names have not been modified to conform to the Arachne standard.
*  
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
* 
* 11-JUN-96  ep-g  made changes to compile on Mac
*
* $Log: CU_Hash.c,v $
* Revision 1.2  1998/07/15 05:20:19  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
 * Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
 * Converting to CVS
 *
* Revision 1.15  1996/12/18 15:44:37  jrosenbe
* Ditched ^L & ^J at start of file
*
* Revision 1.14  1996/12/09  19:23:14  sdeibel
* Misc clean up
*
* Revision 1.13  1996/06/18  21:00:26  epattiso
* made changes to compile on Mac
*
* Revision 1.12  1996/06/18  20:41:11  epattiso
* made changes to compile on Mac
*
* Revision 1.11  1996/05/30  20:35:23  jehresma
* Rearranged position of EXPORT_PREFIX
*
* Revision 1.10  1996/05/21  14:32:09  jehresma
* Removed one word key option
*
* Revision 1.9  1996/04/26  18:13:08  yjlee
* Fixed header comments.
*
* Revision 1.8  1996/03/09  22:06:03  yjlee
* Typo in comment.
*
* Revision 1.7  1996/02/21  04:38:17  jehresma
* Add const to CU_FindHashEntry related functions
*
* Revision 1.6  1996/01/16  21:30:01  yjlee
* Minor comment change.
*
* Revision 1.5  1995/11/02  21:02:17  yjlee
* Added EXPORT_PREFIX constants for MS_WIN
*
* Revision 1.4  1995/10/19  15:10:59  jehresma
* Removed all calls to panic
*
* Revision 1.3  1995/10/12  22:09:20  yjlee
* Uses DEBUGMSG instead of panic when not in X.
*
* Revision 1.2  1995/10/10  17:14:52  yjlee
* Added Dealloc functions.
*
* Revision 1.1  1995/09/25  16:51:36  yjlee
* Initial revision
* 
*/

#ifdef dos
#include "CUHash.h"
#include "CUHash_p.h"
#else
#include "CU_Hash.h"
#include "CU_Hash.prv.h"
#endif


/***************************************************************************
* 
* FUNCTION: 
* 
* CU_InitHashTable 
* 
* DESCRIPTION: 
* 
* Given storage for a hash table, set up the fields to prepare
* the hash table for use.
* 
* PARAMETERS: 
* CUt_HashTable *tablePtr       -- Pointer to table record, which
*                                  is supplied by the caller.
* tUInt4 keyType		-- Type of keys to use in table:
*                                  CUk_STRING_KEYS or CUk_ARRAY_KEYS
* 
* RETURNS: 
* void
* 
* NOTES: 
* 
* TablePtr is now ready to be passed to CU_FindHashEntry and
* CU_CreateHashEntry.
* 
*/

void 
CU_InitHashTable( register CUt_HashTable *tablePtr, tUInt4 keyType ) 
{
  tablePtr->buckets = tablePtr->staticBuckets;
  tablePtr->staticBuckets[0] = tablePtr->staticBuckets[1] = 0;
  tablePtr->staticBuckets[2] = tablePtr->staticBuckets[3] = 0;
  tablePtr->numBuckets = CUk_SMALL_HASH_TABLE;
  tablePtr->numEntries = 0;
  tablePtr->rebuildSize = CUk_SMALL_HASH_TABLE*REBUILD_MULTIPLIER;
  tablePtr->downShift = 28;
  tablePtr->mask = 3;
  tablePtr->keyType = keyType;
  if (keyType == CUk_STRING_KEYS) {
    tablePtr->findProc = CU_StringFind;
    tablePtr->createProc = CU_StringCreate;
  } else {
    tablePtr->findProc = CU_ArrayFind;
    tablePtr->createProc = CU_ArrayCreate;
  }
} /* InitHashTable */


/***************************************************************************
* 
* FUNCTION: 
* CU_DeleteHashEntry 
* 
* DESCRIPTION: 
* 
* Remove a single entry from a hash table.
* 
* PARAMETERS: 
* 
* CUt_HashEntry *entryPtr       -- The entry to delete
* 
* RETURNS: 
* 
* void
* 
* NOTES: 
* 
* The entry given by entryPtr is deleted from its table and
* should never again be used by the caller.  It is up to the
* caller to free the clientData field of the entry, if that
* is relevant.
* 
*/

void 
CU_DeleteHashEntry(CUt_HashEntry *entryPtr)
{
  register CUt_HashEntry *prevPtr;

  /* Error check */
  GOODMEM(entryPtr);

  if (*entryPtr->bucketPtr == entryPtr) {
    *entryPtr->bucketPtr = entryPtr->nextPtr;
  } else {
    for (prevPtr = *entryPtr->bucketPtr; ; prevPtr = prevPtr->nextPtr) {
      if (prevPtr == NULL) {
        DEBUGMSG("malformed bucket chain in CU_DeleteHashEntry");
        BARF;
      }
      if (prevPtr->nextPtr == entryPtr) {
        prevPtr->nextPtr = entryPtr->nextPtr;
        break;
      }
    }
  }
  entryPtr->tablePtr->numEntries--;
  CU_Free(entryPtr);
} /* DeleteHashEntry */
 

/***************************************************************************
* 
* FUNCTION: 
* CU_DeallocHashEntry 
* 
* DESCRIPTION: 
* 
* Remove a single entry from a hash table.  Frees clientData as well.
* 
* PARAMETERS: 
* 
* CUt_HashEntry *entryPtr       -- The entry to delete
* 
* RETURNS: 
* 
* void
* 
* NOTES: 
* 
* The entry given by entryPtr is deleted from its table and
* should never again be used by the caller.
* 
*/

void 
CU_DeallocHashEntry(CUt_HashEntry *entryPtr)
{
    register CUt_HashEntry *prevPtr;

    if (*entryPtr->bucketPtr == entryPtr) {
	*entryPtr->bucketPtr = entryPtr->nextPtr;
    } else {
	for (prevPtr = *entryPtr->bucketPtr; ; prevPtr = prevPtr->nextPtr) {
	    if (prevPtr == NULL) {
		DEBUGMSG("malformed bucket chain in CU_DeleteHashEntry");
                BARF;
	    }
	    if (prevPtr->nextPtr == entryPtr) {
		prevPtr->nextPtr = entryPtr->nextPtr;
		break;
	    }
	}
    }
    entryPtr->tablePtr->numEntries--;
    if ( entryPtr->clientData != NULL ) 
      CU_Free(entryPtr->clientData);
    CU_Free(entryPtr);
 
} /* DeallocHashEntry */
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_DeallocHashTable
*  
* DESCRIPTION: 
* 
* Free up everything associated with a hash table except for
* the record for the table itself.  Calls CU_Free on the client data as
* well.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr      -- Table to delete
* 
* RETURNS: 
* 
* void 
* 
* NOTES: 
* 
* The hash table is no longer useable.
* 
*/

void 
CU_DeallocHashTable( CUt_HashTable *tablePtr )
{
  register CUt_HashEntry *hPtr, *nextPtr;
  tUInt4 i;

  /*
   * Free up all the entries in the table.
   */

  for (i = 0; i < tablePtr->numBuckets; i++) {
    hPtr = tablePtr->buckets[i];
    while (hPtr != NULL) {
      nextPtr = hPtr->nextPtr;
      if (hPtr->clientData) {
        CU_Free(hPtr->clientData); }
      CU_Free(hPtr);
      
      hPtr = nextPtr;
    }
  }

  /*
   * Free up the bucket array, if it was dynamically allocated.
   */

  if (tablePtr->buckets != tablePtr->staticBuckets) {
    CU_Free(tablePtr->buckets);
  }

  /*
   * Arrange for panic/DEBUGMSGs if the table is used again without
   * re-initialization.
   */

  tablePtr->findProc = CU_BogusFind;
  tablePtr->createProc = CU_BogusCreate;
} /* DeallocHashTable */
 
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_DeleteHashTable
*  
* DESCRIPTION: 
* 
* Free up everything associated with a hash table except for
* the record for the table itself.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr      -- Table to delete
* 
* RETURNS: 
* 
* void 
* 
* NOTES: 
* 
* The hash table is no longer useable.
* 
*/

void 
CU_DeleteHashTable( CUt_HashTable *tablePtr )
{
  register CUt_HashEntry *hPtr, *nextPtr;
  tUInt4 i;

  /*
   * Free up all the entries in the table.
   */

  for (i = 0; i < tablePtr->numBuckets; i++) {
    hPtr = tablePtr->buckets[i];
    while (hPtr != NULL) {
      nextPtr = hPtr->nextPtr;
      CU_Free(hPtr);
      hPtr = nextPtr;
    }
  }
  
  /*
   * Free up the bucket array, if it was dynamically allocated.
   */

  if (tablePtr->buckets != tablePtr->staticBuckets) {
    CU_Free(tablePtr->buckets);
  }

  /*
   * Arrange for panic/DEBUGMSGs if the table is used again without
   * re-initialization.
   */

  tablePtr->findProc = CU_BogusFind;
  tablePtr->createProc = CU_BogusCreate;
} /* DeleteHashTable */
 
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_FirstHashEntry
*  
* DESCRIPTION: 
* 
* Locate the first entry in a hash table and set up a record
* that can be used to step through all the remaining entries
* of the table.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr      -- Table to search
* CUt_HashSearch *searchPtr     -- Place to store information about
*                                  progress through the table
* 
* RETURNS: 
* 
* The return value is a pointer to the first entry in tablePtr,
* or NULL if tablePtr has no entries in it.  The memory at
* searchPtr is initialized so that subsequent calls to
* CU_NextHashEntry will return all of the entries in the table,
* one at a time.
* 
* NOTES: 
* 
* 
*/

CUt_HashEntry * 
CU_FirstHashEntry( CUt_HashTable *tablePtr, CUt_HashSearch *searchPtr )
{
  searchPtr->tablePtr = tablePtr;
  searchPtr->nextIndex = 0;
  searchPtr->nextEntryPtr = NULL;
  return CU_NextHashEntry(searchPtr);
} /* FirstHashEntry */

 

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_NextHashEntry
* 
* DESCRIPTION: 
* 
* Once a hash table enumeration has been initiated by calling
* CU_FirstHashEntry, this procedure may be called to return
* successive elements of the table.
* 
* PARAMETERS: 
* 
* Ut_HashSearch *searchPtr      -- Place to store information about
*                                  progress through the table.  Must
*                                  have been initialized by calling
*                                  CU_FirstHashEntry
* 
* RETURNS: 
* 
* The return value is the next entry in the hash table being
* enumerated, or NULL if the end of the table is reached.
* 
* NOTES: 
* 
* 
*/

CUt_HashEntry * 
CU_NextHashEntry( register CUt_HashSearch *searchPtr )
{
  CUt_HashEntry *hPtr;
  
  while (searchPtr->nextEntryPtr == NULL) {
    if (searchPtr->nextIndex >= searchPtr->tablePtr->numBuckets) {
      return NULL;
    }
    searchPtr->nextEntryPtr =
      searchPtr->tablePtr->buckets[searchPtr->nextIndex];
    searchPtr->nextIndex++;
  }
  hPtr = searchPtr->nextEntryPtr;
  searchPtr->nextEntryPtr = hPtr->nextPtr;
  return hPtr;
}



/***************************************************************************
* 
* FUNCTION: 
* 
* CU_HashStats
*  
* DESCRIPTION: 
* 
* Return statistics describing the layout of the hash table
* in its hash buckets.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr       -- Table for which to produce stats
* 
* RETURNS: 
* 
* The return value is a malloc-ed string containing information
* about tablePtr.  It is the caller's responsibility to free
* this string.
* 
* NOTES: 
* 
* 
*/
 
char * 
CU_HashStats( CUt_HashTable *tablePtr )
{
#define NUM_COUNTERS 10
  tUInt4 count[NUM_COUNTERS], overflow, j;
  tUInt4 i;
  double average, tmp;
  register CUt_HashEntry *hPtr;
  char *result, *p;
  
  /*
   * Compute a histogram of bucket usage.
   */
  
  for (i = 0; i < NUM_COUNTERS; i++) {
    count[i] = 0;
  }
  overflow = 0;
  average = 0.0;
  for (i = 0; i < tablePtr->numBuckets; i++) {
    j = 0;
    for (hPtr = tablePtr->buckets[i]; hPtr != NULL; hPtr = hPtr->nextPtr) {
      j++;
    }
    if (j < NUM_COUNTERS) {
      count[j]++;
    } else {
      overflow++;
    }
    tmp = j;
    average += (tmp+1.0)*(tmp/tablePtr->numEntries)/2.0;
  }
  
  /*
   * Print out the histogram and a few other pieces of information.
   */
  
  result = (char *) CU_Malloc((unsigned) ((NUM_COUNTERS*60) + 300));
  if ( result == NULL )
    return NULL;
  sprintf(result, "%ld entries in table, %ld buckets\n",
          tablePtr->numEntries, tablePtr->numBuckets);
  p = result + strlen(result);
  for (i = 0; i < NUM_COUNTERS; i++) {
    sprintf(p, "number of buckets with %ld entries: %ld\n",
            i, count[i]);
    p += strlen(p);
  }
  sprintf(p, "number of buckets with %d or more entries: %ld\n",
          NUM_COUNTERS, overflow);
  p += strlen(p);
  sprintf(p, "average search distance for entry: %.1f", average);
  return result;
  
} /* HashStats */


/**************************************************************************/
/* PRIVATE FUNCTIONS */


/***************************************************************************
* 
* FUNCTION: 
* 
* CU_HashString
* 
* DESCRIPTION: 
* 
* Compute a one-word summary of a text string, which can be
* used to generate a hash index.
* 
* PARAMETERS: 
* 
* 
* RETURNS: 
* 
* The return value is a one-word summary of the information in
* string.
* 
* NOTES: 
* 
* 
*/

static tUInt4
CU_HashString(const char *string)
{
  register tUInt4 result;
  register tInt4 c;

  /*
   * I tried a zillion different hash functions and asked many other
   * people for advice.  Many people had their own favorite functions,
   * all different, but no-one had much idea why they were good ones.
   * I chose the one below (multiply by 9 and add new character)
   * because of the following reasons:
   *
   * 1. Multiplying by 10 is perfect for keys that are decimal strings,
   *    and multiplying by 9 is just about as good.
   * 2. Times-9 is (shift-left-3) plus (old).  This means that each
   *    character's bits hang around in the low-order bits of the
   *    hash value for ever, plus they spread fairly rapidly up to
   *    the high-order bits to fill out the hash value.  This seems
   *    works well both for decimal and non-decimal strings.
   */
  
  result = 0;
  while (1) {
    c = *string;
    string++;
    if (c == 0) {
      break;
    }
    result += (result<<3) + c;
  }

  return result;

} /* CU_HashString */


/***************************************************************************
* 
* FUNCTION: 
* 
* CU_StringFind
* 
* DESCRIPTION: 
* 
* Given a hash table with string keys, and a string key, find
* the entry with a matching key.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr       -- Table in which to lookup entry
* char *key                     --  Key to use to find matching entry
* 
* RETURNS: 
* 
* The return value is a token for the matching entry in the
* hash table, or NULL if there was no matching entry.
* 
* NOTES: 
* 
* 
*/

static CUt_HashEntry * 
CU_StringFind(const CUt_HashTable *tablePtr, const void *key )
{
  register CUt_HashEntry *hPtr;
  register const char *p1, *p2;
  tInt4 index;

  index = CU_HashString((char *)key) & tablePtr->mask;

  /*
   * Search all of the entries in the appropriate bucket.
   */

  for (hPtr = tablePtr->buckets[index]; hPtr != NULL;
       hPtr = hPtr->nextPtr) {
    for (p1 = (char *)key, p2 = hPtr->key.string; ; p1++, p2++) {
      if (*p1 != *p2) {
        break;
      }
      if (*p1 == '\0') {
        return hPtr;
      }
    }
  }

  return NULL;

} /* StringFind */
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_StringCreate
* 
* DESCRIPTION: 
* 
* Given a hash table with string keys, and a string key, find
* the entry with a matching key.  If there is no matching entry,
* then create a new entry that does match.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr       -- Table in which to lookup entry
* char *key                     -- Key to use to find or create matching
*                                  entry
* tInt4 *newPtr                 -- Store info here telling whether a new
*                                  entry was created.
* 
* RETURNS: 
* 
* The return value is a pointer to the matching entry.  If this
* is a newly-created entry, then *newPtr will be set to a non-zero
* value;  otherwise *newPtr will be set to 0.  If this is a new
* entry the value stored in the entry will initially be 0.
* 
* NOTES: 
* 
* A new entry may be added to the hash table.
* 
*/

static CUt_HashEntry *
CU_StringCreate( CUt_HashTable *tablePtr, const void *key, tInt4 *newPtr)
{
  register CUt_HashEntry *hPtr;
  register const char *p1, *p2;
  tInt4 index;
  
  index = CU_HashString((char *)key) & tablePtr->mask;
  
  /*
   * Search all of the entries in this bucket.
   */
  
  for (hPtr = tablePtr->buckets[index]; hPtr != NULL;
       hPtr = hPtr->nextPtr) {
    for (p1 = (char *)key, p2 = hPtr->key.string; ; p1++, p2++) {
      if (*p1 != *p2) {
        break;
      }
      if (*p1 == '\0') {
        *newPtr = 0;
        return hPtr;
      }
    }
  }
  
  /*
   * Entry not found.  Add a new one to the bucket.
   */
  
  *newPtr = 1;
  hPtr = (CUt_HashEntry *) CU_Malloc((tUInt4)
         (sizeof(CUt_HashEntry) + strlen((char *)key) - (sizeof(hPtr->key) -1)));
  if ( hPtr == NULL )
    return NULL;
  
  hPtr->tablePtr = tablePtr;
  hPtr->bucketPtr = &(tablePtr->buckets[index]);
  hPtr->nextPtr = *hPtr->bucketPtr;
  hPtr->clientData = 0;
  strcpy(hPtr->key.string, (char *)key);
  *hPtr->bucketPtr = hPtr;
  tablePtr->numEntries++;
  
  /*
   * If the table has exceeded a decent size, rebuild it with many
   * more buckets.
   */
  
  if (tablePtr->numEntries >= tablePtr->rebuildSize) {
    CU_RebuildTable(tablePtr);
  }
  return hPtr;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_OneWordFind
* 
* DESCRIPTION: 
* 
* Given a hash table with one-word keys, and a one-word key, find
* the entry with a matching key.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr       -- Table in which to lookup entry.
* register char *key            -- Key to use to find matching entry.
* 
* RETURNS: 
* 
* The return value is a token for the matching entry in the
* hash table, or NULL if there was no matching entry.
* 
* NOTES: 
* 
* 
*/
#ifdef OLD
static CUt_HashEntry *
CU_OneWordFind(const CUt_HashTable *tablePtr, const void *key )
{
  register CUt_HashEntry *hPtr;
  tInt4 index;
  
  index = RANDOM_INDEX(tablePtr, key);
  
  /*
   * Search all of the entries in the appropriate bucket.
   */
  
  for (hPtr = tablePtr->buckets[index]; hPtr != NULL;
       hPtr = hPtr->nextPtr) {
    if (hPtr->key.oneWordValue == key) {
      return hPtr;
    }
  }
    return NULL;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_OneWordCreate
* 
* DESCRIPTION: 
* 
* Given a hash table with one-word keys, and a one-word key, find
* the entry with a matching key.  If there is no matching entry,
* then create a new entry that does match.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr       -- Table in which to lookup entry.
* register char *key            -- Key to use to find or create matching
*                                  entry.
* tInt4 *newPtr                 -- Store info here telling whether a new
*                                  entry was created.
* 
* RETURNS: 
* 
* The return value is a pointer to the matching entry.  If this
* is a newly-created entry, then *newPtr will be set to a non-zero
* value;  otherwise *newPtr will be set to 0.  If this is a new
* entry the value stored in the entry will initially be 0.
* 
* NOTES: 
* 
* A new entry may be added to the hash table.
* 
*/

static CUt_HashEntry *
CU_OneWordCreate( CUt_HashTable *tablePtr, const void *key, tInt4 *newPtr )
{
  register CUt_HashEntry *hPtr;
  tInt4 index;
  
  index = RANDOM_INDEX(tablePtr, key);
  
  /*
   * Search all of the entries in this bucket.
   */
  
  for (hPtr = tablePtr->buckets[index]; hPtr != NULL;
       hPtr = hPtr->nextPtr) {
    if (hPtr->key.oneWordValue == key) {
      *newPtr = 0;
      return hPtr;
    }
  }
  
  /*
   * Entry not found.  Add a new one to the bucket.
   */
  
  *newPtr = 1;
  hPtr = (CUt_HashEntry *) CU_Malloc(sizeof(CUt_HashEntry));
  if ( hPtr == NULL )
    return NULL;
  hPtr->tablePtr = tablePtr;
  hPtr->bucketPtr = &(tablePtr->buckets[index]);
  hPtr->nextPtr = *hPtr->bucketPtr;
  hPtr->clientData = 0;
  hPtr->key.oneWordValue = (char*) key;
  *hPtr->bucketPtr = hPtr;
  tablePtr->numEntries++;
  
  /*
   * If the table has exceeded a decent size, rebuild it with many
   * more buckets.
   */
  
  if (tablePtr->numEntries >= tablePtr->rebuildSize) {
    CU_RebuildTable(tablePtr);
  }
  return hPtr;
}
#endif

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_ArrayFind
* 
* DESCRIPTION: 
* 
* Given a hash table with array-of-int keys, and a key, find
* the entry with a matching key.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr       -- Table in which to lookup entry.
* char *key                     -- Key to use to find matching entry.
* 
* RETURNS: 
* 
* The return value is a token for the matching entry in the
* hash table, or NULL if there was no matching entry.
* 
* NOTES: 
* 
*/

static CUt_HashEntry *
CU_ArrayFind(const CUt_HashTable *tablePtr, const void *key )
{
  register CUt_HashEntry *hPtr;
  const tInt1 *arrayPtr = (const tInt1 *) key;
  const tInt1 *iPtr1, *iPtr2;
  tInt4 index, count;
  
  /* Error check */
  GOODMEM(tablePtr);
  GOODMEM(key);
  
  for (index = 0, count = tablePtr->keyType, iPtr1 = arrayPtr;
       count > 0; count--, iPtr1++) {
    index += *iPtr1;
  }
  index = RANDOM_INDEX(tablePtr, index);
  
  /*
   * Search all of the entries in the appropriate bucket.
   */
  
  for (hPtr = tablePtr->buckets[index]; hPtr != NULL;
       hPtr = hPtr->nextPtr) {
    for (iPtr1 = arrayPtr, iPtr2 = hPtr->key.bytes,
         count = tablePtr->keyType; ; count--, iPtr1++, iPtr2++) {
      if (count == 0) {
        return hPtr;
      }
      if (*iPtr1 != *iPtr2) {
        break;
      }
    }
  }
  return NULL;
}



/***************************************************************************
* 
* FUNCTION: 
* 
* CU_ArrayCreate
* 
* DESCRIPTION: 
* 
* Given a hash table with n-byte keys, and a n-byte key, find
* the entry with a matching key.  If there is no matching entry,
* then create a new entry that does match.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr       -- Table in which to lookup entry
* register char *key            -- Key to use to find or create matching
*                                  entry
* tInt4 *newPtr                 -- Store info here telling whether a new
*                                  entry was created
* 
* RETURNS: 
* 
* The return value is a pointer to the matching entry.  If this
* is a newly-created entry, then *newPtr will be set to a non-zero
* value;  otherwise *newPtr will be set to 0.  If this is a new
* entry the value stored in the entry will initially be 0.
* 
* NOTES: 
* 
* A new entry may be added to the hash table.
* 
*/

static CUt_HashEntry *
CU_ArrayCreate( CUt_HashTable *tablePtr, const void *key, tInt4 *newPtr)
{
  register CUt_HashEntry *hPtr;
  const tInt1 *arrayPtr = (const tInt1 *) key;
  const tInt1 *iPtr1, *iPtr2;
  tInt1 *entry_val;
  tInt4 index, count;
  
  for (index = 0, count = tablePtr->keyType, iPtr1 = arrayPtr;
       count > 0; count--, iPtr1++) {
    index += *iPtr1;
  }
  index = RANDOM_INDEX(tablePtr, index);
  
  /*
   * Search all of the entries in the appropriate bucket.
   */
  
  for (hPtr = tablePtr->buckets[index]; hPtr != NULL;
       hPtr = hPtr->nextPtr) {
    for (iPtr1 = arrayPtr, iPtr2 = hPtr->key.bytes,
         count = tablePtr->keyType; ; count--, iPtr1++, iPtr2++) {
      if (count == 0) {
        *newPtr = 0;
        return hPtr;
      }
      if (*iPtr1 != *iPtr2) {
        break;
      }
    }
  }
  
  /*
   * Entry not found.  Add a new one to the bucket.
   */
  
  *newPtr = 1;
  hPtr = (CUt_HashEntry *) CU_Malloc(sizeof(CUt_HashEntry)
                                     + (tablePtr->keyType - 4));
  if ( hPtr == NULL )
    return NULL;
  hPtr->tablePtr = tablePtr;
  hPtr->bucketPtr = &(tablePtr->buckets[index]);
  hPtr->nextPtr = *hPtr->bucketPtr;
  hPtr->clientData = 0;
  for (iPtr1 = arrayPtr, entry_val = hPtr->key.bytes, count = tablePtr->keyType;
       count > 0; count--, iPtr1++, entry_val++) {
    *entry_val = *iPtr1;
  }
  *hPtr->bucketPtr = hPtr;
  tablePtr->numEntries++;
  
  /*
   * If the table has exceeded a decent size, rebuild it with many
   * more buckets.
   */
  
  if (tablePtr->numEntries >= tablePtr->rebuildSize) {
    CU_RebuildTable(tablePtr);
  }
  return hPtr;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_BogusFind
* 
* DESCRIPTION: 
* 
* This procedure is invoked when an CU_FindHashEntry is called
* on a table that has been deleted.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr       -- Table in which to lookup entry
* char *key                     -- Key to use to find matching entry
* 
* RETURNS: 
* 
* If panic/DEBUGMSG returns (which panic it shouldn't) this procedure returns
* NULL.
* 
* NOTES: 
* 
* Generates a panic/DEBUGMSG.
* 
*/

static CUt_HashEntry *
CU_BogusFind( const CUt_HashTable *tablePtr, const void *key )
{
  UNUSED(tablePtr);
  UNUSED(key);

  DEBUGMSG("called CU_FindHashEntry on deleted table");
  BARF;

  return NULL;
}
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_BogusCreate
* 
* DESCRIPTION: 
* 
* This procedure is invoked when an CU_CreateHashEntry is called
* on a table that has been deleted.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr       -- Table in which to lookup entry
* char *key                     -- Key to use to find or create matching
*                                  entry
* tInt4 *newPtr                 -- Store info here telling whether a new
*                                  entry was created.
* 
* RETURNS: 
* 
* If panic/DEBUGMSG returns (which panic it shouldn't) this procedure returns
* NULL.
* 
* NOTES: 
* 
* Generates a panic/DEBUGMSG.
* 
*/

static CUt_HashEntry *
CU_BogusCreate( CUt_HashTable *tablePtr, const void *key, tInt4 *newPtr )
{
  UNUSED(tablePtr);
  UNUSED(key);
  UNUSED(newPtr);

  DEBUGMSG("called CU_CreateHashEntry on deleted table");
  BARF;

  return NULL;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_RebuildTable
* 
* DESCRIPTION: 
* 
* This procedure is invoked when the ratio of entries to hash
* buckets becomes too large.  It creates a new table with a
* larger bucket array and moves all of the entries into the
* new table.
* 
* PARAMETERS: 
* 
* CUt_HashTable *tablePtr       -- Table to enlarge.
* 
* RETURNS: 
* 
* None.
* 
* NOTES: 
* 
* Memory gets reallocated and entries get re-hashed to new
* buckets.
* 
*/

static void
CU_RebuildTable( CUt_HashTable *tablePtr )
{
  tInt4 oldSize, count, index;
  CUt_HashEntry **oldBuckets;
  register CUt_HashEntry **oldChainPtr, **newChainPtr;
  register CUt_HashEntry *hPtr;

  oldSize = tablePtr->numBuckets;
  oldBuckets = tablePtr->buckets;

  /*
   * Allocate and initialize the new bucket array, and set up
   * hashing constants for new array size.
   */

  tablePtr->numBuckets *= 4;
  tablePtr->buckets = (CUt_HashEntry **) CU_Malloc((tUInt4)
                    (tablePtr->numBuckets * sizeof(CUt_HashEntry *)));
  if ( tablePtr->buckets == NULL ) {
    ASSERT(1 == 0);
    tablePtr->buckets = oldBuckets;
    return;
  }

  for (count = tablePtr->numBuckets, newChainPtr = tablePtr->buckets;
       count > 0; count--, newChainPtr++) {
    *newChainPtr = NULL;
  }
  tablePtr->rebuildSize *= 4;
  tablePtr->downShift -= 2;
  tablePtr->mask = (tablePtr->mask << 2) + 3;

  /*
   * Rehash all of the existing entries into the new bucket array.
   */

  for (oldChainPtr = oldBuckets; oldSize > 0; oldSize--, oldChainPtr++) {
    for (hPtr = *oldChainPtr; hPtr != NULL; hPtr = *oldChainPtr) {
      *oldChainPtr = hPtr->nextPtr;
      if (tablePtr->keyType == CUk_STRING_KEYS) {
        index = CU_HashString(hPtr->key.string) & tablePtr->mask;
      } else {
        register tInt1 *iPtr;
        tUInt4 count;
        
        for (index = 0, count = tablePtr->keyType,
             iPtr = hPtr->key.bytes; count > 0; count--, iPtr++) {
          index += *iPtr;
        }
        index = RANDOM_INDEX(tablePtr, index);
      }
      hPtr->bucketPtr = &(tablePtr->buckets[index]);
      hPtr->nextPtr = *hPtr->bucketPtr;
      *hPtr->bucketPtr = hPtr;
    }
  }
  
  /*
   * Free up the old bucket array, if it was dynamically allocated.
   */
  
  if (oldBuckets != tablePtr->staticBuckets) {
    CU_Free(oldBuckets);
  }

} /* RebuildTable */


/* end of CU_Hash.c */




