/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE:
* 
* CU_Sequence.cc
* 
* AUTHOR: 
* 
* Stephan R.A. Deibel
* John Ehresman
* YongJoon Lee
* 
* CREATION DATE: 
* 
* Oct 1996
* 
* DESCRIPTION: 
* 
* Sequence construct support.  Compatible with CORBA sequences.
* 
* NOTES:
*
*
* MODIFICATIONS: 
* 
* $Log: CU_Sequence.c,v $
* Revision 1.4  1998/07/15 05:20:20  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
* Revision 1.3  1998/07/06 15:44:35  jehresma
* Defined CUt_StringSeq and support functions.  Defined CU_GetFilenamesInDir
* and CU_JoinPathname.
*
* Revision 1.2  1998/03/06 23:31:34  jehresma
* Fixed bug in CU_SeqRemove
*
 * Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
 * Converting to CVS
 *
* Revision 1.12  1997/06/14 19:12:34  sdeibel
* Bug fix to CU_SeqRemove : was copying too much!
*
* Revision 1.11  1997/03/03 23:40:42  sdeibel
* Added CU_SeqAppendArray
*
// Revision 1.10  1997/02/11  21:21:25  sdeibel
// Changed Remove not to have return error: never fails!
//
// Revision 1.9  1996/12/17  17:47:32  sdeibel
// Added Sequence Remove
//
// Revision 1.8  1996/12/05  23:46:59  sdeibel
// Mac port
//
// Revision 1.7  1996/11/25  20:03:07  sdeibel
// Added insert and sorting functions
//
// Revision 1.6  1996/11/11  22:54:43  sdeibel
// Re-added SeqPrepend()
//
// Revision 1.5  1996/10/24  21:23:40  jehresma
// Converted to C file
//
// Revision 1.3  1996/10/16  16:52:19  jehresma
// Fixed sequence append bug and changed memcpy calls to CU_MemCpy
//
// Revision 1.2  1996/10/10  18:26:59  sdeibel
// Misc
//
// Revision 1.1  1996/10/10  18:10:44  sdeibel
// Initial revision
//
*
*/

#include "CU_Sequence.h"
#include "CU_Text.h"


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqInit()
* 
* DESCRIPTION:
*
* Initializes a new sequence
*
* PARAMETERS: 
* 
* void *       -- The sequence to initialize; needs to be a sequence type
* 
*/

void  
CU_SeqInit(void * seq)
{
  CUt_GenericSeq * sequence;

  GOODMEM(seq);
  sequence = (CUt_GenericSeq *) seq;

  sequence->_maximum = 0;
  sequence->_length = 0;
  sequence->_buffer = NULL;
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqCreate()
* 
* DESCRIPTION:
*
* Creates a new sequence on the heap and initializes it
*
* RETURNS: 
*
* void *       -- Pointer to newly allocated sequence, or NULL if out of
*                 memory
* 
*/

void * 
CU_SeqCreate()
{
  CUt_GenericSeq *sequence;

  sequence = (CUt_GenericSeq *) CU_Malloc(sizeof(CUt_GenericSeq));
  if ( sequence == NULL )
    return NULL;

  CU_SeqInit(sequence);

  return sequence;
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqClear()
* 
* DESCRIPTION:
*
* Clears a sequence, freeing memory of all elements
*
* PARAMETERS: 
* 
* void *       -- The sequence to destroy; needs to be a sequence type
* 
* NOTES:
*
* Only does a shallow deallocation of the elements in the sequence
* 
*/

void 
CU_SeqClear(void * seq)
{
  CUt_GenericSeq * sequence;

  GOODMEM(seq);
  sequence = (CUt_GenericSeq *) seq;

  if ( sequence->_buffer != NULL )
    CU_Free(sequence->_buffer);
  sequence->_buffer = NULL;
  sequence->_maximum = 0;
  sequence->_length = 0;
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqDestroy()
* 
* DESCRIPTION:
*
* Frees all memory associated with a sequence
*
* PARAMETERS: 
* 
* void *       -- The sequence to free; needs to be a sequence type
* 
* NOTES:
*
* Only does a shallow deallocation of the elements in the sequence
* 
*/

void 
CU_SeqDestroy(void * seq)
{
  GOODMEM(seq);

  CU_SeqClear(seq);
  CU_Free(seq);
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqExpand()
* 
* DESCRIPTION:
*
* Expand memory for the sequence to include at least max_size elements and 
* init any new space allocated with zeros.  If sequence is longer than 
* new_max, this doesn't do anything.
*
* PARAMETERS: 
* 
* void *      -- The sequence to expand; needs to be a sequence type
* tUInt4      -- New maximum size of the sequence
* size_t      -- Size of a sequence element
* 
* RETURNS:
*
* tError      -- kNoError or kOutOfMemory if failed in operation (in which
*                case the sequence buffer will be set to NULL!)
*
* NOTES:
*
* Note that the _length field is not changed in the struct; the user must
* set that as appropriate.
*
* Users should expand sequences by a block size so they avoid reallocating
* on every new element added
* 
*/

tError 
CU_SeqExpand(void * seq, tUInt4 new_max, size_t el_size)
{
  CUt_GenericSeq* sequence;

  GOODMEM(seq);
  sequence = (CUt_GenericSeq *) seq;

  if ( new_max > sequence->_maximum ) {
    if ( sequence->_buffer == NULL )
      sequence->_buffer = CU_Malloc(el_size * new_max);
    else
      sequence->_buffer = CU_Realloc(sequence->_buffer, el_size*new_max);

    if ( sequence->_buffer == NULL )
      return kOutOfMemory;

    CU_MemSet((tUInt1 *) sequence->_buffer + sequence->_maximum * el_size, 0, 
           (new_max - sequence->_maximum) * el_size);

    sequence->_maximum = new_max;
  }
  
  return kNoError;
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqSet()
* 
* DESCRIPTION:
*
* Set the value of sequence element index to the given value
*
* PARAMETERS: 
* 
* void *      -- The sequence to modify; needs to be a sequence type
* tUInt4      -- Index of the element to set; 0 = first one
* void *      -- Pointer to data value
* size_t      -- Size of a sequence element
* 
* NOTES:
*
* Should probably expand by a block size so reallocation doesn't
* occur with each element!
*
*/

tError 
CU_SeqSet(void * seq, tUInt4 index, void * data, size_t el_size)
{
  CUt_GenericSeq* sequence;

  GOODMEM(seq);

  sequence = (CUt_GenericSeq *) seq;
   
  if ( CU_SeqExpand(seq, index + 1, el_size) == kOutOfMemory )
    return kOutOfMemory;

  CU_MemCpy(((tUInt1 *) sequence->_buffer + index * el_size), data, el_size);

  if ( index + 1 > sequence->_length )
    sequence->_length = index + 1;

  return kNoError;
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqInsert()
* 
* DESCRIPTION:
*
* Insert the given value into the sequence.  The sequence is expanded
* and any overlapping data is moved out of the way before insertion.
*
* PARAMETERS: 
* 
* void *      -- The sequence to modify; needs to be a sequence type
* tUInt4      -- Index of the element to set; 0 = first one
* void *      -- Pointer to data value
* size_t      -- Size of a sequence element
* 
* NOTES:
*
* Should probably expand by a block size so reallocation doesn't
* occur with each element!
*
*/

tError 
CU_SeqInsert(void * seq, tUInt4 index, void * data, size_t el_size)
{
  CUt_GenericSeq* sequence;

  GOODMEM(seq);

  sequence = (CUt_GenericSeq *) seq;

  /* Expand, if necessary */
  if ( CU_SeqExpand(seq, CU_MAX(sequence->_length + 1, index + 1), 
		    el_size) == kOutOfMemory )
    return kOutOfMemory;

  /* Move data out of the way */
  CU_MemMove(((tUInt1 *) sequence->_buffer + (index + 1) * el_size),
	     ((tUInt1 *) sequence->_buffer + index * el_size),
	     (sequence->_length - index) * el_size);

  /* Set item */
  CU_MemCpy(((tUInt1 *) sequence->_buffer + index * el_size), data, el_size);
  sequence->_length = CU_MAX(sequence->_length + 1, index + 1);

  return kNoError;
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqRemove()
* 
* DESCRIPTION:
*
* Remove the given entry from the sequence.  The sequence is adjusted
* so elements remain contiguous in memory.  Will never fail.
*
* PARAMETERS: 
* 
* void *      -- The sequence to modify; needs to be a sequence type
* tUInt4      -- Index of the element to set; 0 = first one
* size_t      -- Size of a sequence element
* 
*/

void
CU_SeqRemove(void * seq, tUInt4 index, size_t el_size)
{
  CUt_GenericSeq* sequence;

  GOODMEM(seq);

  sequence = (CUt_GenericSeq *) seq;

  /* Move data into gap */
  CU_MemMove(((tUInt1 *) sequence->_buffer + index * el_size),
	     ((tUInt1 *) sequence->_buffer + (index + 1) * el_size),
	     (sequence->_length - index - 1) * el_size);

  /* Adjust length */
  sequence->_length = sequence->_length - 1;

  /* Set unused memory at end to 0 */
  CU_MemSet((tUInt1 *) sequence->_buffer + sequence->_length * el_size, 0, 
            el_size);

  return;
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqInsertSorted()
* 
* DESCRIPTION:
*
* Insert the given value into the sequence in sorted order, as defined
* by the given sort key and controlled by the given sort function.
* The sequence is expanded and overlapping data is moved out of the
* way before insertion.
*
* PARAMETERS: 
* 
* void *          -- The sequence to modify; needs to be a sequence type
* void *          -- Pointer to data value
* size_t          -- Size of a sequence element
* void *          -- Pointer to insertion sort key.
* CUt_SeqSortFct  -- Function to control insertion:  Compares sort
*                    key with existing sequence elements and returns:
*                      =0  if sort key equals existing element key, 
*                      <0  if sort key is < existing element key, and
*                      >0  if sort key is > existing element key.
* tUInt4          -- Buffer space for returned position at which new 
*                    element was inserted; 0 = first one; may be NULL.
* 
* RETURNS:
*
* tError          -- kNoError or kOutOfMemory.
*
* NOTES:
*
* Should probably expand by a block size so reallocation doesn't
* occur with each element!
*
*/

tError CU_SeqInsertSorted(void *seq, void *data, size_t el_size,
			  void *sort_key, CUt_SeqSortFct sort_fct,
			  tUInt4 *pos)
{
  tUInt4 loc;

  GOODMEM(seq);
  GOODMEM(sort_key);
  GOODMEM(sort_fct);

  CU_SeqFindSorted(seq, el_size, sort_key, sort_fct, &loc);
  if (pos != NULL)
    *pos = loc;

  return CU_SeqInsert(seq, loc, data, el_size);
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqFindSorted()
* 
* DESCRIPTION:
*
* Find the given sort key in the sorted sequence, using the given
* sort function to control the search.
*
* PARAMETERS: 
* 
* void *          -- The sequence to modify; needs to be a sequence type
* size_t          -- Size of a sequence element
* void *          -- Pointer to insertion sort key.
* CUt_SeqSortFct  -- Function to control insertion:  Compares sort
*                    key with existing sequence elements and returns:
*                      =0  if sort key equals existing element key, 
*                      <0  if sort key is < existing element key, and
*                      >0  if sort key is > existing element key.
* tUInt4 *        -- Buffer space for returned position of element in
*                    sequence; set to appropriate insertion point
*                    if element was not found; 0 = first one; may be NULL.
* 
* RETURNS:
*
* tBoolean        -- TRUE if found.
*
*/

tBoolean CU_SeqFindSorted(void *seq, size_t el_size, void *sort_key,
			  CUt_SeqSortFct sort_fct, tUInt4 *pos)
{
  CUt_GenericSeq *sequence;
  tUInt4 upper, lower, middle;
  tInt4 cmp_result;
  tBoolean found = FALSE;

  /* Error check */
  GOODMEM(seq);
  GOODMEM(sort_key);
  GOODMEM(sort_fct);

  /* Type cast */
  sequence = (CUt_GenericSeq *) seq;

  /* Do a binary search for the key */
  lower = 1;
  upper = sequence->_length;
  middle = 0;
  while ( lower <= upper && !found) {
    middle = lower + (upper - lower) / 2;
    cmp_result = sort_fct(sort_key, 
      (tUInt1 *) sequence->_buffer + el_size * (middle - 1));
    if ( cmp_result == 0 )
      found = TRUE;
    else if ( cmp_result < 0 )
      upper = middle - 1;
    else
      lower = middle + 1;
  }

  /* Set return position value, if requested */
  if (pos != NULL) {
    if ( middle > 0 ) {
      if (cmp_result <= 0)
	*pos = middle - 1;
      else
	*pos = middle;
    }
    else
      *pos = 0;
  }
  return found;
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqAppend()
* 
* DESCRIPTION:
*
* Append a new value on to the end of a sequence; this does a
* shallow copy of the data
*
* PARAMETERS: 
* 
* void *       -- The sequence to modify; needs to be a sequence type
* void *       -- Pointer to data value
* size_t       -- Size of a sequence element
* 
*/

tError 
CU_SeqAppend(void * seq, void * data, size_t el_size)
{
  CUt_GenericSeq* sequence;

  GOODMEM(seq);

  sequence = (CUt_GenericSeq *) seq;

  return CU_SeqSet(seq, sequence->_length, data, el_size);
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqAppendArray()
* 
* DESCRIPTION:
*
* Append an array of new values on to the end of a sequence; this does a
* shallow copy of the data
*
* PARAMETERS: 
* 
* void *       -- The sequence to modify; needs to be a sequence type
* tUInt4       -- The number of items in array
* void *       -- Pointer to array of items
* size_t       -- Size of a single sequence element
* 
*/

tError 
CU_SeqAppendArray(void * seq, tUInt4 num, void * data, size_t el_size)
{
  CUt_GenericSeq* sequence;

  GOODMEM(seq);

  sequence = (CUt_GenericSeq *) seq;
   
  if ( CU_SeqExpand(seq, sequence->_length + num, el_size) == kOutOfMemory )
    return kOutOfMemory;

  CU_MemCpy(((tUInt1 *) sequence->_buffer + sequence->_length * el_size), 
	    data, el_size * num);

  sequence->_length += num;

  return kNoError;
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqAppendSeq()
* 
* DESCRIPTION:
*
* Append a new sequence on to the end of a sequence; this does a
* shallow copy of the data
*
* PARAMETERS: 
* 
* void *       -- The sequence to modify; needs to be a sequence type
* void *       -- Pointer to sequence to append; needs to be a sequence type
* size_t       -- Size of a sequence element
* 
*/

tError  
CU_SeqAppendSeq(void* dest_seq, void* src_seq, size_t el_size)
{
  CUt_GenericSeq *src;
  CUt_GenericSeq *dest;

  GOODMEM(dest_seq);
  GOODMEM(src_seq);

  UNTESTED;

  src = (CUt_GenericSeq *) src_seq;
  dest = (CUt_GenericSeq *) dest_seq;

  /* Nothing to do if seq to append from is empty */
  if ( src->_length == 0 )
    return kNoError;

  if (CU_SeqExpand(dest_seq, dest->_length + src->_length, el_size)
      == kOutOfMemory)
    return kOutOfMemory;
  
  CU_MemCpy(((tUInt1 *) dest->_buffer + dest->_length * el_size), 
            ((tUInt1 *) src->_buffer), src->_length * el_size);
  
  dest->_length = dest->_length + src->_length;

  return kNoError;
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqPrepend()
* 
* DESCRIPTION:
*
* Prepend a new value on to the beginning of a sequence; this does a
* shallow copy of the data
*
* PARAMETERS: 
* 
* void *       -- The sequence to modify; needs to be a sequence type
* void *       -- Pointer to data value
* size_t       -- Size of a sequence element
* 
*/

tError 
CU_SeqPrepend(void * seq, void * data, size_t el_size)
{
  CUt_GenericSeq* sequence;

  GOODMEM(seq);

  sequence = (CUt_GenericSeq *) seq;

  if ( CU_SeqExpand(seq, sequence->_length + 1, el_size) == kOutOfMemory )
    return kOutOfMemory;

  CU_MemMove(((tUInt1 *) sequence->_buffer + el_size),
	    ((tUInt1 *) sequence->_buffer), sequence->_length * el_size);
  CU_MemCpy(((tUInt1 *) sequence->_buffer), data, el_size);

  sequence->_length++;

  return kNoError;
}


/***************************************************************************
* 
* FUNCTION:
*
* CU_SeqCopy()
* 
* DESCRIPTION:
*
* Create a copy of the given sequence; this makes a copy of the buffer
* but will not follow pointers.
*
* PARAMETERS: 
* 
* const void *          -- The sequence to copy; 
*                            needs to be a sequence type
* void *                -- The destination sequence; 
*                            needs to be a sequence type
* size_t                -- Size of a sequence element
* 
*
* RETURNS:
*
* tError                -- kNoError or kOutOfMemory if failed in operation 
*                          (sequence buffer will be set to NULL on error)
*
* NOTES:
*
* The destination sequence must be empty
* 
*/

tError 
CU_SeqCopy(const void * src_seq, void * dest_seq, size_t el_size)
{
  CUt_GenericSeq * src_sequence;
  CUt_GenericSeq * dest_sequence;

  GOODMEM(src_seq);
  GOODMEM(dest_seq);

  src_sequence = (CUt_GenericSeq *) src_seq;
  dest_sequence = (CUt_GenericSeq *) dest_seq;

  ASSERT(dest_sequence->_buffer == NULL);

  *dest_sequence = *src_sequence;

  /* Now, create a copy of the elements */
  if ( src_sequence->_maximum != 0 ) {
    GOODMEM(src_sequence->_buffer);
    dest_sequence->_buffer = CU_Malloc(src_sequence->_maximum * el_size);
    if ( dest_sequence->_buffer == NULL )
      return kOutOfMemory;

   CU_MemCpy(dest_sequence->_buffer, src_sequence->_buffer,
             src_sequence->_maximum * el_size);
  }     
  
  return kNoError;
}


/****************************************************************************
* FUNCTION:
*
* CU_StringSeqInit()
*
* DESCRIPTION:
*
* Initialize a string sequence
*
* PARAMETERS:
*
* CUt_StringSeq* seq -- Sequence to initialize
* 
* RETURNS:
*
* void
*
*/
void CU_StringSeqInit(CUt_StringSeq* seq)
{
  CU_SeqInit(seq);
}


/****************************************************************************
* FUNCTION:
*
* CU_StringSeqClear()
*
* DESCRIPTION:
*
* Clear a string sequence
*
* PARAMETERS:
*
* CUt_StringSeq* seq -- Sequence to clear
* 
* RETURNS:
*
* void
*
*/
void CU_StringSeqClear(CUt_StringSeq* seq)
{
  tUInt4 i;

  for ( i = 0; i < seq->_length; i++ ) {
    if ( seq->_buffer[i] != NULL )
      CU_Free(seq->_buffer[i]);
  }

  CU_SeqClear(seq);
}


/****************************************************************************
* FUNCTION:
*
* CU_StringSeqAppendItem()
*
* DESCRIPTION:
*
* Append a string to a string sequence
*
* PARAMETERS:
*
* CUt_StringSeq* seq -- Sequence to append to
* const char* str    -- String to append
* 
* RETURNS:
*
* tError -- kNoError or kOutOfMemory
*
*/
tError CU_StringSeqAppendItem(CUt_StringSeq* seq, const char* str)
{
  char* tmp = NULL;
  tError err = kNoError;
  
  if ( str != NULL ) {
    tmp = CU_StrDup(str);
    if ( tmp == NULL )
      err = kOutOfMemory;
  }

  if ( err == kNoError )
    err = CU_SeqAppend(seq, &tmp, sizeof(tmp));

  if ( err != kNoError && tmp != NULL )
    CU_Free(tmp);

  return err;
}


/* end of CU_Sequence.cc */

