/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE: CU_Stream.cc
* 
* AUTHOR: 
* 
* YongJoon Lee
* Jaime Requena
* John Ehresman
* Stephan Deibel
*
* CREATION DATE: 
*
* Aug 1995
* 
* DESCRIPTION: 
* 
* Streams implementations with slightly better error checking
* than the usual.
* 
* NOTES: 
* 
* 
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
* 
* $Log: CU_Stream.cpp,v $
* Revision 1.6  1998/12/01 23:18:51  bsloan
* Port to MetroWerks CodeWarrior Pro4 for Windows
*
* Revision 1.5  1998/08/05 20:33:03  jehresma
* Implemented simple file output stream class
*
* Revision 1.4  1998/07/15 05:20:21  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
* Revision 1.3  1998/06/10 16:15:10  jehresma
* Fixed memory leak
*
// Revision 1.2  1997/12/04  17:06:09  jehresma
// Minor changes to suppress warnings
//
// Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
// Converting to CVS
//
* Revision 1.16  1997/09/15 21:58:42  jehresma
* Added virtual operators to OStream
*
* Revision 1.15  1997/07/08 21:52:20  sdeibel
* Bug fix to outputting NULL strings
*
* Revision 1.14  1997/07/01  15:11:58  sdeibel
* Added char text output operator, removed tInt1 numerical output
* operator (caller should typecast to tInt2).
*
* Revision 1.13  1996/12/18  15:45:37  jrosenbe
* Ditched ^L & ^J at start of file
*
* Revision 1.12  1996/12/06  19:37:09  jehresma
* Added write method to CUC_OStrStream
*
* Revision 1.11  1996/11/08 17:37:09  jehresma
* Reimplemented << operators for int's to eliminate use of ltoa
*
* Revision 1.10  1996/09/24 06:59:37  jehresma
* Changes for Visual C++
*
* Revision 1.9  1996/09/23  23:03:07  sdeibel
* Misc
*
* Revision 1.8  1996/09/23  17:16:11  sdeibel
* Added OStream and OStrStream (written by Jaimed; merged by me)
*
* Revision 1.7  1996/04/12  20:39:16  sdeibel
* Fixied read() to return correct value for file streams
*
* Revision 1.6  1996/04/11  22:34:31  sdeibel
* Made seekg() more lenient in generic stream class:  If seek past end,
* it sets curr to end instead of failing.
*
* Revision 1.5  1996/04/10  21:38:10  sdeibel
* Added better error checking; errors were being ignored and
* not passed out!
*
* Revision 1.4  1996/03/20  00:20:06  jehresma
* Added const to file stream constructor
*
* Revision 1.3  1996/01/25  16:26:59  jont
* fixed typo bug in CUC_IFStream::CUC_IFStream()
*
*/

#include "CU_Stream.hh"
#include "CU_Debug.h"
#include <string.h>


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IStream::CUC_IStream
* 
* DESCRIPTION: 
* 
* Constructor for this class
* 
* PARAMETERS: 
* 
* None.
* 
*/

CUC_IStream::CUC_IStream()
{
  fBuff = NULL;
  fBase = NULL;
  fCPtr = NULL;
  fEPtr = NULL;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IStream::~CUC_IStream()
* 
* DESCRIPTION: 
* 
* Destructor for this class
* 
*/

CUC_IStream::~CUC_IStream()
{
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IStream::get()
* 
* DESCRIPTION: 
* 
* Get a byte from the stream
* 
* PARAMETERS: 
* 
* None.
* 
* RETURNS: 
* 
* tInt2                     -- The byte as a tInt2; EOF when at end
*                              of stream
* 
*/

tInt2 
CUC_IStream::get()
{
  tUInt1 ret_val;

  // Check if at end
  if ( fCPtr > fEPtr )
    return EOF;

  // Return one byte
  ret_val = *fCPtr;
  fCPtr++;

  // Done
  return( ret_val ); 
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IStream::read()
* 
* DESCRIPTION: 
* 
* Read a number of bytes from the stream
* 
* PARAMETERS: 
* 
* void *                   -- Buffer space for bytes read
* tUInt4                   -- Number of bytes to read
* 
* RETURNS: 
* 
* tUInt4                   -- Number of bytes actually read;
*                             0 when at end of stream
* 
*/

tUInt4 
CUC_IStream::read(void *dest, tUInt4 n)
{
  tUInt4 num;

  GOODMEM( dest );

  // Compute number of bytes to read, up to number available
  num = CU_MIN(n, (tUInt4) (fEPtr - fCPtr + 1));

  // Read bytes
  if (num > 0) {
    CU_MemCpy(dest, fCPtr, num);
    fCPtr += num;
  }

  // Done
  return num;
}
  

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IStream::seekg()
* 
* DESCRIPTION: 
*
* Seek to the given position on the stream 
* 
* PARAMETERS: 
* 
* tUInt4                       -- The position (0 = start); if beyond end 
*                                 of stream, the end is used.
* 
* RETURNS: 
* 
* tError                       -- kNoError in all cases
* 
*/

tError
CUC_IStream::seekg(tUInt4 pos)
{
  if (fBuff + pos > fEPtr) {
    fCPtr = fEPtr;
    return kNoError;
  }
  else {
    fCPtr = fBuff + pos;
    return kNoError;
  }
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IStream::tellg()
* 
* DESCRIPTION: 
* 
* Get the current position on the stream, relative to starting location
* 
* PARAMETERS: 
* 
* None.
* 
* RETURNS: 
* 
* tUInt4                      -- The position (0 = at start)
* 
*/
  
tUInt4 
CUC_IStream::tellg()
{
  return(fCPtr - fBuff);
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IFStream::CUC_IFStream()
* 
* DESCRIPTION: 
*
* Input file stream constructor
* 
* PARAMETERS: 
*
* None.
* 
*/

CUC_IFStream::CUC_IFStream()
{
  fName = NULL;
  fFile = NULL;
}
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IFStream::~CUC_IFStream
* 
* DESCRIPTION: 
* 
* Destructor for this class
* 
*/

CUC_IFStream::~CUC_IFStream()
{
  close();
}
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IFStream::open()
* 
* DESCRIPTION: 
* 
* Open a file into the file stream; closes any previously open file
* 
* PARAMETERS: 
* 
* const char *               -- The file name
* 
* RETURNS: 
* 
* tError                     -- kNoError, kOutOfMemory, or kFileIOError
* 
* NOTES:
*
* No file is opened if there is any kind of error
* 
*/

tError
CUC_IFStream::open( const char *filename )
{
  tError err = kNoError;

  // Close existing file
  close();

  // Set file name
  fName = (char *) CU_Malloc(strlen(filename) + 1);
  if (fName == NULL) {
    err = kOutOfMemory;
    goto exception;
  }

  // Open file
  strcpy( fName, filename );
  fFile = fopen(fName, "rb");  
  if (fFile == NULL) {
    err = kFileIOError;
    goto exception;
  }

  // Success
  return kNoError;

  // Failure
exception:
  close();
  return err;
}
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IFStream::close()
* 
* DESCRIPTION: 
* 
* Close the file associated with this stream, if any
* 
* PARAMETERS: 
* 
* None.
* 
* RETURNS: 
* 
* void.
* 
*/

void 
CUC_IFStream::close()
{
  if (fFile != NULL) {
    fclose( fFile );
    fFile = NULL;
  }
  if (fName != NULL) {
    CU_Free( fName );
    fName = NULL;
  }
}
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IFStream::get()
* 
* DESCRIPTION: 
* 
* Get a byte from the file
* 
* PARAMETERS: 
* 
* None.
* 
* RETURNS: 
* 
* tInt2                       -- The byte, in the form of a tInt2;
*                                EOF on failure or at end of file.
* 
*/

tInt2 
CUC_IFStream::get()
{
  ASSERT(fFile != NULL);
  return fgetc(fFile);
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IFStream::read()
* 
* DESCRIPTION: 
* 
* Read the specified number of bytes from the file
* 
* PARAMETERS: 
* 
* void *                      -- Buffer for returned bytes
* tUInt4                      -- Number of bytes to read
* 
* RETURNS: 
* 
* tUInt4                      -- Number of bytes actually read
* 
*/

tUInt4
CUC_IFStream::read(void *dest,tUInt4 n)
{
  ASSERT(fFile != NULL);
  return fread(dest, 1, n, fFile);
}
 

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IFStream::seekg()
* 
* DESCRIPTION: 
* 
* Seek to the given position in the file (0 = start)
* 
* PARAMETERS: 
* 
* tUInt4                        -- The position
* 
* RETURNS: 
* 
* tError                        -- kNoError or kFileIOError
* 
*/

tError 
CUC_IFStream::seekg(tUInt4 pos)
{
  if (fseek( fFile, pos, SEEK_SET ) != 0)
    return kFileIOError;
  else
    return kNoError;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IFStream::tellg()
* 
* DESCRIPTION: 
* 
* Get the current location in the file
* 
* PARAMETERS: 
* 
* None.
* 
* RETURNS: 
* 
* tUInt4                     -- The current position in file
* 
*/

tUInt4 
CUC_IFStream::tellg()
{
  return( ftell(fFile) );
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IStrStream::CUC_IStrStream()
* 
* DESCRIPTION: 
* 
* Constructor for this class
* 
* PARAMETERS: 
* 
* char *                     -- The buffer for the stream; the
*                               memory ownership is retained by the
*                               caller, and the buffer must be kept
*                               intact until after the stream 
*                               is destroyed.
* tUInt4                     -- The buffer length
* 
*/

CUC_IStrStream::CUC_IStrStream(const char *buff, tUInt4 len)
{
  fBuff = buff;
  fBase = buff;
  fEPtr = buff + len - 1;
  fCPtr = buff;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_IStrStream::~CUC_IStrStream()
* 
* DESCRIPTION: 
* 
* Destructor for the class 
* 
*/

CUC_IStrStream::~CUC_IStrStream()
{
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream::CUC_OStream
* 
* DESCRIPTION: 
* 
* Constructor for this class
* 
* PARAMETERS: 
* 
* None.
* 
*/

CUC_OStream::CUC_OStream()
{
  fError = kNoError;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream::~CUC_OStream()
* 
* DESCRIPTION: 
* 
* Destructor for this class
* 
*/

CUC_OStream::~CUC_OStream()
{
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream::CUC_OStream
* 
* DESCRIPTION: 
* 
* Constructor for this class
* 
* PARAMETERS: 
* 
* None.
*
* RETURNS:
*
* tError
* 
*/

tError CUC_OStream::GetError()
{
  return fError;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream::CUC_OStream
* 
* DESCRIPTION: 
* 
* Constructor for this class
* 
* PARAMETERS: 
* 
* None.
* 
*/

CUC_OStrStream::CUC_OStrStream()
{
  fText = CU_TXCreate();
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream::~CUC_OStream()
* 
* DESCRIPTION: 
* 
* Destructor for this class
* 
*/

CUC_OStrStream::~CUC_OStrStream()
{
  CU_TXDestroy(fText);
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStrStream::write ()
* 
* DESCRIPTION: 
* 
* Write a certain number of characters into the stream
*
* PARAMETERS: 
* 
* const char*                         -- Ptr to 1st character
* tUInt4 len                          -- Numbe of characters to add
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStrStream& CUC_OStrStream::write(const char* item, tUInt4 length)
{
  fError = CU_TXAppendText(fText, item, length);
  return *this;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* char                         -- The character
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OStrStream::operator<< (char item)
{
  fError = CU_TXAppendChar(fText, item);
  return *this;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* char*                         -- The string
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OStrStream::operator<< (char* item)
{
  if (item != NULL) 
    fError = CU_TXAppendCString(fText, item);
  return *this;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* const char*                         -- The string
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OStrStream::operator<< (const char* item)
{
  fError = CU_TXAppendCString(fText, item);
  return *this;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* tInt2                          -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OStrStream::operator<< (tInt2 item)
{
  char tmpstr[64];
  sprintf(tmpstr, "%d", item);
  fError = CU_TXAppendCString(fText, tmpstr);
  return *this;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* tInt4                         -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OStrStream::operator<< (tInt4 item)
{
  char buffer[50];
  sprintf(buffer, "%ld", (long int) item);
  fError = CU_TXAppendCString(fText, buffer);

  return *this;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* tUInt1                         -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OStrStream::operator<< (tUInt1 item)
{
  char tmpstr[64];
  sprintf(tmpstr, "%d", item);
  fError = CU_TXAppendCString(fText, tmpstr);
  return *this;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* tUInt2                          -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OStrStream::operator<< (tUInt2 item)
{
  char tmpstr[64];
  sprintf(tmpstr, "%u", item);
  fError = CU_TXAppendCString(fText, tmpstr);
  return *this;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* tUInt4                         -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OStrStream::operator<< (tUInt4 item)
{
  char buffer[50];
  sprintf(buffer, "%ld", (long unsigned int) item);
  fError = CU_TXAppendCString(fText, buffer);

  return *this;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStrStream::operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* float                         -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OStrStream::operator<< (float item)
{
  char tmpstr[64];
  sprintf(tmpstr, "%f", item);
  fError = CU_TXAppendCString(fText, tmpstr);
  return *this;
}


/***************************************************************************

* 

* FUNCTION: 

* 

* CUC_OStrStream::str()

* 

* DESCRIPTION: 

* 

* Returns string associated with stream; always returns valid ptr

*

* RETURNS: 

* 

* const char*  -- string               

* 

*/



const char* CUC_OStrStream::str() const

{ 

  if ( fText == NULL )

    return "";

  if ( fText->text == NULL )

    return "";



  return fText->text;

}





/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OFileStream::CUC_OFileStream
* 
* DESCRIPTION: 
* 
* Constructor for this class
* 
* PARAMETERS: 
* 
* None.
* 
*/

CUC_OFileStream::CUC_OFileStream()
  : fFP(NULL), fCloseFP(FALSE)
{
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OFileStream::CUC_OFileStream
* 
* DESCRIPTION: 
* 
* Constructor for this class to wrap around an open file pointer
* 
* PARAMETERS: 
* 
* FILE* fp -- File pointer to write output to; this object will not close it
* 
*/

CUC_OFileStream::CUC_OFileStream(FILE* fp)
  : fFP(fp), fCloseFP(FALSE)
{
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OFileStream::~CUC_OFileStream()
* 
* DESCRIPTION: 
* 
* Destructor for this class
* 
*/

CUC_OFileStream::~CUC_OFileStream()
{
  if ( fCloseFP && fFP != NULL )
    fclose(fFP);
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OFileStream::write ()
* 
* DESCRIPTION: 
* 
* Write a certain number of characters into the stream
*
* PARAMETERS: 
* 
* const char*                         -- Ptr to 1st character
* tUInt4 len                          -- Numbe of characters to add
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OFileStream::write(const char* item, tUInt4 length)
{
  for ( tUInt4 i = 0; i < length; i++ )
    fputc(item[i], fFP);
  return *this;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* char                         -- The character
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OFileStream::operator<< (char item)
{
  fputc(item, fFP);
  return *this;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* char*                         -- The string
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OFileStream::operator<< (char* item)
{
  if (item != NULL) 
    fputs(item, fFP);
  return *this;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* const char*                         -- The string
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OFileStream::operator<< (const char* item)
{
  fputs(item, fFP);
  return *this;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* tInt2                          -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OFileStream::operator<< (tInt2 item)
{
  fprintf(fFP, "%d", item);
  return *this;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* tInt4                         -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OFileStream::operator<< (tInt4 item)
{
  fprintf(fFP, "%ld", (long int) item);

  return *this;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* tUInt1                         -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OFileStream::operator<< (tUInt1 item)
{
  fprintf(fFP, "%d", item);
  return *this;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* tUInt2                          -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OFileStream::operator<< (tUInt2 item)
{
  fprintf(fFP, "%u", item);
  return *this;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OStream& operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* tUInt4                         -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OFileStream::operator<< (tUInt4 item)
{
  fprintf(fFP, "%ld", (long unsigned int) item);
  return *this;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CUC_OFileStream::operator << ()
* 
* DESCRIPTION: 
* 
* Output operator
*
* PARAMETERS: 
* 
* float                         -- The number
* 
* RETURNS: 
* 
* CUC_OStream&               
* 
*/

CUC_OStream& CUC_OFileStream::operator<< (float item)
{
  fprintf(fFP, "%f", item);
  return *this;
}

/* end of CU_Stream.cc */

