/***************************************************************************
*
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*
* ==========================================================================
*
* FILE: CU_Text.c
*
* AUTHOR:
*
* John Ehresman
* Edward Pattison-Gordon
* Stephan Deibel
*
* CREATION DATE: 
*
* Jan 1996, based on previous work
*
* DESCRIPTION:
*
* C language dynamic string support.  Takes care of memory management 
* when setting string to a value, appending data to it, or disposing it.
*
* NOTES:
*
*
* MODIFICATIONS:
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
*
* $Log: CU_Text.c,v $
* Revision 1.6  1998/12/01 23:18:51  bsloan
* Port to MetroWerks CodeWarrior Pro4 for Windows
*
* Revision 1.5  1998/10/08 14:53:39  swang
* Changes to run in VC++, need to re-check compatibility with gnu linux
*
* Revision 1.4  1998/08/14 22:00:26  sdeibel
* Bug fix to allow null string
*
* Revision 1.3  1998/07/15 05:20:21  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
 * Revision 1.2  1997/11/02  05:40:55  sdeibel
 * Made some of the string comparison functions a bit kinder with
 * NULL pointers.
 *
 * Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
 * Converting to CVS
 *
* Revision 1.26  1997/09/16 14:54:23  jrequena
* Added type CUt_TextSeq and function CU_TXClearSequence
*
* Revision 1.25  1997/08/12  19:17:14  jrequena
* added CU_StripString
*
* Revision 1.24  1997/06/10  21:29:05  jehresma
* Implementation of CU_StrNCaseCmp
*
* Revision 1.23  1997/03/25 19:15:07  sdeibel
* Added CU_StrCharDelete
*
* Revision 1.22  1996/10/15  16:23:15  jehresma
* Bug fix to CU_TXDeleteRange
*
* Revision 1.21  1996/10/08  21:55:42  sdeibel
* Relaxed a few GOODMEMs to work OK when str_len == 0
*
* Revision 1.20  1996/10/08  21:41:11  jehresma
* Added functions to insert and delete ranges
*
* Revision 1.19  1996/10/03  17:39:04  jehresma
* Added CU_TXUseTextPtr function
*
* Revision 1.18  1996/09/24  09:27:19  sdeibel
* Fixed silly +1 bug
*
* Revision 1.17  1996/09/24  06:29:40  sdeibel
* Misc
*
* Revision 1.16  1996/09/24  05:31:01  sdeibel
* Fixed compare bugs (more)
*
* Revision 1.15  1996/09/24  03:36:53  sdeibel
* Fixed bug in str compare fct
*
* Revision 1.14  1996/09/18  20:27:29  sdeibel
* Added conversion fct
*
* Revision 1.13  1996/09/18  14:49:06  sdeibel
* Commenting
*
* Revision 1.12  1996/06/26  22:37:09  epattiso
* added CU_CStringToNewPString and CU_PStringToNewCString
*
* Revision 1.11  1996/05/30  20:35:23  jehresma
* Rearranged position of EXPORT_PREFIX
*
* Revision 1.10  1996/03/20  23:49:16  yjlee
* Added an assert in aAppend.\
*
* Revision 1.9  1996/03/19  20:39:18  yjlee
* Minor format changes only.
*
* Revision 1.8  1996/03/19  20:36:42  yjlee
* Changed CU_TXSetTextPtr to add NULL term only _if_necessary_.
*   (for use with string constants.)
*
* Revision 1.7  1996/03/18  04:44:31  jehresma
* Fixed typecasting warnings
*
* Revision 1.6  1996/03/13  23:10:24  jehresma
* Bug fix to compare function
*
* Revision 1.5  1996/03/13  16:11:20  jehresma
* win32 round 1
*
* Revision 1.4  1996/02/19  20:18:16  sdeibel
* Removed const from CU_TXUseCString since that was technically
* incorrect; this function is problematic anyway.
*
* Revision 1.3  1996/02/19  03:16:01  jehresma
* Added CU_TXCopy
*
* Revision 1.2  1996/01/24  19:00:24  sdeibel
* Minor commenting change
*
 * Revision 1.1  1996/01/23  23:33:04  sdeibel
 * Initial revision
 *
*
*/


/**************************************************************************/
/* INCLUDES */

#ifdef dos
#include "CUText.h"
#include "CUText_p.h"
#include "CUDebug.h"
#else
#include "CU_Text.h"
#include "CU_Text.prv.h"
#include "CU_Debug.h"
#endif

#include <ctype.h>


/***************************************************************************
*
* FUNCTION:
*
* CU_PStringToCString()
*
* DESCRIPTION:
*
* Destructively converts a Pascal string to a C string.
*
* PARAMETERS:
*
* char *        -- The string
*
* RETURNS:
*
* void (the input string is altered)
*
*/

EXPORT_PREFIX void 
CU_PStringToCString (char * the_string)
{
  tUInt2  length;

  /* Error check */
  ASSERT(the_string != NULL);

  /* first byte of string contains the string length */
  length = *the_string;

  /* shift the string characters down one position,  */
  /* overwriting the length byte */
  CU_MemMove(the_string, (the_string + 1), length);

  /* place null character at end of string */
  the_string[length] = 0;

} /* CU_PStringToCString */


/***************************************************************************
*
* FUNCTION:
*
* CU_PStringToNewCString()
*
* DESCRIPTION:
*
* Creates a new C string with the same contents as the Pascal string.
*
* PARAMETERS:
*
* const char *        -- The Pascal string
*
* RETURNS:
*
* char *              -- The new C string
*
*/

EXPORT_PREFIX char * 
CU_PStringToNewCString (const char * p_string)
{
  tUInt1  length;
  char *  c_string = NULL;

  /* Error check */
  ASSERT(p_string != NULL);

  /* first byte of Pascal string contains string length */
  length = *p_string;
  
  /* create the new string */
  c_string = (char *) CU_Malloc((length + 1));
  if ((c_string == NULL))  return (NULL);

  /* copy string characters from Pascal string to C string */
  CU_MemCpy(c_string, &p_string[1], length);

  /* place null character at end of C string */
  c_string[length] = 0;
  
  return (c_string);

} /* CU_PStringToNewCString */


/***************************************************************************
*
* FUNCTION:
*
* CU_CStringToPString()
*
* DESCRIPTION:
*
* Destructively converts a C string to a Pascal string.
*
* PARAMETERS:
*
* char *        -- The string
*
* RETURNS:
*
* void (the input string is altered)
*
*/

EXPORT_PREFIX void 
CU_CStringToPString (char * the_string)
{
  tUInt1  length;                            
  char   last_char;
  char   temp_char;

  /* Error check */
  ASSERT(the_string != NULL);

  length = 0;

  if (the_string[0])
  {
     /* there's at least one character in the string */
     last_char = the_string[0];

     /* in one pass through the string: */
     /*  - determine the length of the string */
     /*  - shift it up one character */
     do
     {
		  length = length + 1;
        temp_char = the_string[length];

        /* last_char contains the character originally in [length - 1] */
        /* move that character into [length] */
        the_string[length] = last_char;

        /* save the character originally in [length] in last_char */
        last_char = temp_char;

     } while (last_char);

  };/* if */

  /* place length in first byte of string */
  the_string[0] = length;

} /* CU_CStringToPString */


/***************************************************************************
*
* FUNCTION:
*
* CU_CStringToNewPString()
*
* DESCRIPTION:
*
* Creates a new Pascal string with the same contents as the C string 
* (upto kMaxUInt1 characters).
*
* PARAMETERS:
*
* const char *        -- The C string
*
* RETURNS:
*
* char *              -- The new Pascal string
*
*/

EXPORT_PREFIX char * 
CU_CStringToNewPString (const char * c_string)
{
  tUInt4  length;                            
  char *  p_string;

  /* Error check */
  ASSERT(c_string != NULL);

  /* Pascal strings cannot be longer than kMaxUInt1 */
  length = CU_MIN(strlen(c_string), kMaxUInt1);

  /* create the new string */
  p_string = (char *) CU_Malloc((length + 1));
  if ((p_string == NULL))  return (NULL);

  /* copy string characters from Pascal string to C string */
  CU_MemCpy(&p_string[1], c_string, length);

  /* place length in first byte of string */
  p_string[0] = (char) length;

  return (p_string);

} /* CU_CStringToNewPString */


/***************************************************************************
* 
* FUNCTION: CU_StrDup
*
* DESCRIPTION: Returns a NULL terminated copy of a NULL terminated string
*  in newly allocated memory or NULL if memory error
* 
* PARAMETERS: const char* src: The source string
* 
* RETURNS: char*: ptr to copy of the string; NULL if out of memory
* 
* NOTES: Implementation of standard strdup found on most Unix implementations
* 
*/
EXPORT_PREFIX char* 
CU_StrDup(const char* src)
{
  tUInt4 len;
  char* ret_val;

  if (src == NULL)
    return NULL;

  /* Find length of string and allocate memory */
  len = strlen(src);
  ret_val = (char*) CU_Malloc(len + 1);
  if ( ret_val == NULL )
    return NULL;

  /* Copy string and terminate it */
  if ( len > 0 )
    CU_MemCpy(ret_val, src, len);
  ret_val[len] = '\0';

  return ret_val;
}


/***************************************************************************
* 
* FUNCTION: CU_StrCaseCmp
*
* DESCRIPTION: Do a case-insensitive compare of two strings and return 0 if they're
*  the same, 1 if the first string is lexically greater than the second, or -1
*  if the first is lexically less than the second
* 
* PARAMETERS: 
*  const char* s1: 1st string
*  const char* s2: 2nd string
* 
* RETURNS: 0 if equal, 1 if 1st string > 2nd, or -1 if 1st string < 2nd
* 
* NOTES: Implementation could be sped up, but this will work
* 
*/
EXPORT_PREFIX tInt4 
CU_StrCaseCmp(const char* s1, const char *s2)
{
  tUInt4 len1;
  tUInt4 len2;
  tUInt4 i;
  
  /* Trivial cases */
  if ((s1 == NULL) && (s2 == NULL))
    return 0;
  if (s1 == NULL)
    return -1;
  if (s2 == NULL)
    return 1;

  /* Set up */
  len1 = strlen(s1);
  len2 = strlen(s2);

  /* Compare */
  for (i = 0; ((i < len1) && (i < len2)); i++) {
	if (toupper(s1[i]) < toupper(s2[i]))
      return -1;
 	else if (toupper(s1[i]) > toupper(s2[i]))
      return 1;
  }
  
  /* All matches up to min(len1,len2) so... */
  if (len1 == len2)
    return 0;
  else if (len1 < len2)
    return -1;
  else
    return 1;
}



/***************************************************************************
* 
* FUNCTION: CU_StrNCaseCmp
*
* DESCRIPTION: Do a case-insensitive compare of two strings up to the given 
*  length and return 0 if they're the same, 1 if the first string is lexically 
*  greater than the second, or -1 if the first is lexically less than the second
* 
* PARAMETERS: 
*  const char* s1: 1st string
*  const char* s2: 2nd string
*  tUInt4 max_len: Maximum length of compare
* 
* RETURNS: 0 if equal, 1 if 1st string > 2nd, or -1 if 1st string < 2nd
* 
* NOTES: Implementation could be sped up, but this will work
* 
*/
EXPORT_PREFIX tInt4 
CU_StrNCaseCmp(const char* s1, const char *s2, tUInt4 max_len)
{
  tUInt4 len1;
  tUInt4 len2;
  tUInt4 i;
  tUInt4 cmp_len;
  
  /* Trivial cases */
  if ((s1 == NULL) && (s2 == NULL))
    return 0;
  if (s1 == NULL)
    return -1;
  if (s2 == NULL)
    return 1;

  /* Set up */
  len1 = strlen(s1);
  len2 = strlen(s2);
  cmp_len = CU_MIN(max_len, CU_MIN(len1, len2));

  /* Compare */
  for (i = 0; i < cmp_len; i++) {
    if (toupper(s1[i]) < toupper(s2[i]))
      return -1;
    else if (toupper(s1[i]) > toupper(s2[i]))
      return 1;
  }

  /* If hit max len, they're equal */
  if ( i == max_len )
    return 0;

  /* All matches up to min(len1,len2) so... */
  if (len1 == len2)
    return 0;
  else if (len1 < len2)
    return -1;
  else
    return 1;
}


/****************************************************************************
* FUNCTION: 
*
* CU_StrDeleteChar()
*
* DESCRIPTION: 
*
* Delete the given range of characters from a string.
*
* PARAMETERS: 
*
* char *                 -- The string; modified in place
* tUInt4                 -- Starting position in string
* tUInt4                 -- Length of range to remove
*
* RETURNS:
*
* void.
*
*/

void CU_StrDeleteChar(char* string, tUInt4 start, tUInt4 length) 
{
  char* temp;
  tUInt4 temp_int;
  tUInt4 slen;

  GOODMEM(string);
  slen = strlen(string);
  ASSERT(slen >= start + length);

  /* Trivial case */
  if (length == 0)
    return;

  /* Remove char */
  temp = string + start + length;
  temp_int = slen - length; 
  CU_MemMove(string + start, temp, slen - length - start + 1);
  string[temp_int] = '\0';
}

/***************************************************************************
* 
* FUNCTION: CU_StripString
*
* DESCRIPTION: Strips all leading and trailing whitespace from given string
* 
* PARAMETERS: char* str: string to strip
* 
* RETURNS: nothing
* 
* NOTES: modifies string in-place
* 
*/
void CU_StripString(char *str)
{
  tInt4 pos = 0;
  tInt4 first_char;

  GOODMEM(str);

  // Find beginning of string
  while ( isspace(str[pos]) ) {
    pos++;
  }

  // Left shift and find end of string
  first_char = pos;
  while ( str[pos] != 0 ) {
    if ( first_char != 0 ) {
      str[pos - first_char] = str[pos];
    }
    pos++;
  }
  if ( first_char != 0 ) {
    pos -= first_char;
    str[pos] = '\0';
  }

  // Eliminate whitespace at end of string
  while ( pos > 0 ) {
    pos--;
    if ( isspace(str[pos]) ) {
      str[pos] = '\0';
    } else {
      break;
    }
  }  
}

  
/****************************************************************************
* FUNCTION: 
*
* CU_TXCreate()
*
* DESCRIPTION: 
*
* Create an empty Text structure with default buffer block size.
*
* PARAMETERS: 
*
* None.
*
* RETURNS:
*
* CUt_Text *       -- The newly allocated string
*
*/

EXPORT_PREFIX CUt_Text * 
CU_TXCreate()
{
  CUt_Text *retval = NULL;

  /* Create string */
  retval = (CUt_Text*) CU_Malloc(sizeof(CUt_Text));
  if (retval == NULL)
    return NULL;

  /* Init */
  CU_TXInit(retval);

  /* Done */
  return retval;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXInit()
*
* DESCRIPTION: 
*
* Init a Text structure to empty with default buffer block size.
*
* PARAMETERS: 
*
* CUt_Text *       -- The Text
*
* RETURNS:
*
* None.
*
* NOTES:
*
* If previously used, called CU_TXClear() first.
*
* Can also be used in conjunction with CU_TXGetTextPtr() or 
* CU_TXGetCStringPtr() to steal the text buffer for use elsewhere.
*
*/

EXPORT_PREFIX void 
CU_TXInit(CUt_Text *str)
{
  GOODMEM(str);

  str->text = NULL;
  str->length = 0;
  str->buffer_size = 0;
  str->block_size = CUk_DefTextBlockSize;
  str->borrowed = FALSE;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXClear()
*
* DESCRIPTION: 
*
* Clear a used Text structure to empty.  Buffer block size is left
* unchanged.
*
* PARAMETERS: 
*
* CUt_Text *       -- The Text
*
* RETURNS:
*
* None.
*
* NOTES:
*
* Do not use unless CU_TXCreate() or CU_TXInit() have been called.
*
*/

EXPORT_PREFIX void 
CU_TXClear(CUt_Text *str)
{
  GOODMEM(str);

  if ((str->text != NULL) && !(str->borrowed))
    CU_Free(str->text);
  str->text = NULL;
  str->length = 0;
  str->buffer_size = 0;
  str->borrowed = FALSE;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXDestroy()
*
* DESCRIPTION: 
*
* Destroy an allocated Text structure.
*
* PARAMETERS: 
*
* CUt_Text *       -- The Text
*
* RETURNS:
*
* None.
*
* NOTES:
*
* Do not use unless CU_TXCreate() was used to create the structure.
*
*/

EXPORT_PREFIX void
CU_TXDestroy(CUt_Text *str)
{
  GOODMEM(str);
  CU_TXClear(str);
  CU_Free(str);
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXSetBlockSize()
*
* DESCRIPTION: 
*
* Destroy an allocated Text structure.
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text
* tUInt4              -- The new block size
*
* RETURNS:
*
* None.
*
*/

EXPORT_PREFIX void
CU_TXSetBlockSize(CUt_Text *str, tUInt4 block_size)
{
  GOODMEM(str);
  ASSERT(block_size > 0);
  str->block_size = block_size;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXSetText()
*
* DESCRIPTION: 
*
* Set the value to the given text.
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text
* const char *        -- The text; a copy is made
* tUInt4              -- The text length
*
* RETURNS:
*
* tError              -- kNoError or kOutOfMemory
*
* NOTES:
*
* Embedded NULL characters in the string within the given length are
* ignored, and will cause problems later if the string is converted
* to a C string.
*
*/

EXPORT_PREFIX tError
CU_TXSetText(CUt_Text *str, const char *text, tUInt4 len)
{
  GOODMEM(str);
  CU_TXClear(str);
  return CU_TXAppendText(str, text, len);
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXSetTextPtr()
*
* DESCRIPTION: 
*
* Set the value pointer to the given text.
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text
* const char *        -- The text pointer; memory ownership is relinquished
*                        by caller
* tUInt4              -- The text length
* tUInt4              -- The text buffer length (>= text length + 1)
*
* RETURNS:
*
* None.
*
* NOTES:
*
* Embedded NULL characters in the string within the given length are
* ignored, and will cause problems later if the string is converted
* to a C string.
*
*/

EXPORT_PREFIX void
CU_TXSetTextPtr(CUt_Text *str, char *text, tUInt4 len, tUInt4 buffer)
{
  GOODMEM(str);
  GOODMEM(text);
  ASSERT(buffer >= len + 1);

  CU_TXClear(str);

  str->text = text;
  if (str->text[len] != '\0')
    str->text[len] = '\0';
  str->length = len;
  str->buffer_size = buffer;
  str->borrowed = FALSE;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXUseCString()
*
* DESCRIPTION: 
*
* Set the value pointer to the given C string (NULL terminated).
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text
* char *              -- The C string; memory ownership is retained
*                        by caller but see notes below
*
* RETURNS:
*
* None.
*
* NOTES:
*
* The given char * will not be altered; if the CUt_Text is edited
* subsequently, a copy of the char * will be made first.
*
* Caller must take care that the given pointer is not deallocated while
* this CUt_Text is still in use; the CUt_Text should either be cleared
* or deallocated before the given C String is deallocated or popped
* off of the stack.
*
* This will be problematic for passing CUt_Text structures across
* the network using IIOP under CORBA!  [SRAD]
*
*/

EXPORT_PREFIX void
CU_TXUseCString(CUt_Text *str, char *text)
{
  GOODMEM(str);
  GOODMEM(text);

  CU_TXClear(str);

  str->text = text;
  str->length = strlen(text);
  str->buffer_size = strlen(text) + 1;
  str->borrowed = TRUE;
}


/****************************************************************************
* FUNCTION:
*
* CU_TXUseTextPtr()
*
* DESCRIPTION:
*
* Set the value pointer to the given string and length
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text
* char *              -- The C string; memory ownership is retained
*                        by caller but see notes below
* tUInt4              -- Length of the string
*
* RETURNS:
*
* None.
*
* NOTES:
*
* The given char * will not be altered; if the CUt_Text is edited
* subsequently, a copy of the char * will be made first.
*
* Caller must take care that the given pointer is not deallocated while
* this CUt_Text is still in use; the CUt_Text should either be cleared
* or deallocated before the given C String is deallocated or popped
* off of the stack.
*
*/

void
CU_TXUseTextPtr(CUt_Text *str, char *text, tUInt4 length)
{
  GOODMEM(str);
  GOODMEM(text);

  CU_TXClear(str);

  str->text = text;
  str->length = length;
  str->buffer_size = length + 1;
  str->borrowed = TRUE;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXAppendChar()
*
* DESCRIPTION: 
*
* Append the given character to the Text
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text
* const char          -- The character
*
* RETURNS:
*
* tError              -- kNoError or kOutOfMemory
*
*/

EXPORT_PREFIX tError
CU_TXAppendChar(CUt_Text *str, const char the_char)
{
  tUInt4 needed_size;
  tError error;

  GOODMEM(str);

  /* Check if buffer is full */
  needed_size = str->length + 2;
  if ( needed_size >= str->buffer_size ) {
    error = CU_TXExpandBuffer(str, needed_size);
    if ( error != kNoError )
      return error;
  }

  /* Add the character */
  ASSERT(!str->borrowed);
  str->text[str->length] = the_char;
  ++str->length;
  str->text[str->length] = '\0';

  return kNoError;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXAppendText()
*
* DESCRIPTION: 
*
* Append the given text string to the text
*
* PARAMETERS: 
*
* CUt_Text *          -- The CU Text
* const char *        -- The text to append; a copy is made
* len                 -- The text length
*
* RETURNS:
*
* tError              -- kNoError or kOutOfMemory
*
*/

EXPORT_PREFIX tError
CU_TXAppendText(CUt_Text *str, const char *text, tUInt4 len)
{
  tError error;
  tUInt4 len_needed;

  GOODMEM(str);
  if (len == 0)
    return kNoError;
  GOODMEM(text);

  /* Check if buffer is full */
  len_needed = str->length + len + 1;
  if ( len_needed >= str->buffer_size ) {
    error = CU_TXExpandBuffer(str, len_needed);
    if ( error != kNoError )
      return error;
  }

  /* Append the string */
  ASSERT(!str->borrowed);
  CU_MemCpy(str->text + str->length, text, len);
  str->length += len;
  str->text[str->length] = '\0';

  /* Done */
  return kNoError;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXAppendCString()
*
* DESCRIPTION: 
*
* Append the given C string (NULL terminated) to the Text
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text
* const char *        -- The C string
*
* RETURNS:
*
* tError              -- kNoError or kOutOfMemory
*
*/

EXPORT_PREFIX tError 
CU_TXAppendCString(CUt_Text *str, const char *cstr)
{
  if (cstr != NULL)
    return CU_TXAppendText(str, cstr, strlen(cstr));
  else
    return kNoError;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXAppendCUText()
*
* DESCRIPTION: 
*
* Append the given CU Text to the Text
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text to append to
* const CUt_Text *    -- The Text to append; a copy is made
*
* RETURNS:
*
* tError              -- kNoError or kOutOfMemory
*
*/

EXPORT_PREFIX tError 
CU_TXAppendCUText(CUt_Text *str, const CUt_Text *dstr)
{
  GOODMEM(str);
  GOODMEM(dstr);
  return CU_TXAppendText(str, CU_TXGetCStringPtr(dstr), CU_TXGetLength(dstr));
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXGetLength()
*
* DESCRIPTION: 
*
* Get the current length of the Text's text.
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text
*
* RETURNS:
*
* tUInt4              -- The length
*
*/

EXPORT_PREFIX tUInt4
CU_TXGetLength (const CUt_Text *str)
{
  GOODMEM(str);
  return str->length;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXGetText()
*
* DESCRIPTION: 
*
* Get the value of the CUt_Text
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text.
* char *              -- Buffer space for the return value.
* tUInt4 *            -- On input: The length of provided buffer space
*                        On output: The length of the text; truncated to
*                        available space if necessary.
*
* RETURNS:
*
* The text and length value are copied into the provided buffers.
*
*/

EXPORT_PREFIX void 
CU_TXGetText(CUt_Text *str, char *text, tUInt4 *len)
{
  GOODMEM(str);
  GOODMEM(text);
  GOODMEM(len);

  *len = CU_MIN(*len, str->length);
  CU_MemCpy(text, str->text, *len);
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXGetTextPtr()
*
* DESCRIPTION: 
*
* Get pointer to the value of the CUt_Text
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text.
* char **             -- Buffer space for the returned pointer;
*                        memory ownership is retained by the CUt_Text.
* tUInt4 *            -- Buffer space for the returned length.
*
* RETURNS:
*
* The text pointer and length value are copied into the provided buffers.
*
*/

EXPORT_PREFIX void 
CU_TXGetTextPtr(CUt_Text *str, char **text, tUInt4 *len)
{
  GOODMEM(str);
  GOODMEM(text);
  GOODMEM(len);

  *text = str->text;
  *len = str->length;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXGetCString()
*
* DESCRIPTION: 
*
* Get a copy of the C string (NULL terminated) equivalent of the given text, 
* truncated to the given maximum length
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text.
* char *              -- Buffer space for return value; must be allocated
*                        to accomodate given maximum length plus one for
*                        NULL terminator.
* tUInt4              -- Maximum string length.
*
* RETURNS:
*
* The string is copied into the given buffer space.
*
* NOTES:
*
* If the text has embedded NULL characters, the resulting C string
* will also contain multiple NULL characters.
*
*/

EXPORT_PREFIX void 
CU_TXGetCString(const CUt_Text *str, char *cstr, tUInt1 max_len)
{
  /* Error check */
  GOODMEM(str);
  GOODMEM(cstr);

  /* Do it */
  memcpy((void *) cstr, CU_TXGetCStringPtr(str), 
         CU_MIN(max_len, CU_TXGetLength(str)));
  cstr[CU_MIN(255, CU_TXGetLength(str))] = '\0';
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXGetCStringPtr()
*
* DESCRIPTION: 
*
* Get a pointer to the C string (NULL terminated) equivalent of the string
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text.
*
* RETURNS:
*
* const char *        -- The C string (NULL terminated);
*                        memory ownership is retained by the Text.
*
* NOTES:
*
* If the text contains embedded NULL characters, the resulting C
* string will also contain multiple NULL characters.
*
*/

EXPORT_PREFIX const char * 
CU_TXGetCStringPtr(const CUt_Text *str)
{
  GOODMEM(str);
  return str->text;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXTextToCString()
*
* DESCRIPTION: 
*
* Create a new C language style string from the given CUt_Text;
* supports conversion of empty text.
*
* PARAMETERS: 
*
* const CUt_Text *    -- The Text.
* char **             -- Buffer space for returned char *; memory
*                        ownership is relinquished to caller
*
* RETURNS:
*
* tError              -- kNoError or kOutOfMemory
*
* NOTES:
*
* If the text contains embedded NULL characters, the resulting
* C string will also contain multiple NULL characters.
*
*/

tError CU_TXTextToCString(const CUt_Text *text, char **str)
{
  const char *source;
  tUInt4 len;

  GOODMEM(text);
  GOODMEM(str);

  /* Compute source info */
  source = CU_TXGetCStringPtr(text);
  if (source == NULL)
    len = 4;
  else
    len = strlen(source) + 1;
  
  /* Allocate return value */
  *str = (char*) CU_Malloc(len);
  if (*str == NULL)
    return kOutOfMemory;

  /* Copy data */
  if (source != NULL)
    strcpy(*str, source);
  else
    (*str)[0] = '\0';

  /* Success */
  return kNoError;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXGetPString()
*
* DESCRIPTION: 
*
* Get the Pascal string (0 = len byte) equivalent of the string,
* truncated to the given maximum length
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text.
* char *              -- Buffer space for return value; must be allocated
*                        to accomodate given maximum length.
* tUInt1              -- Maximum string length (<= 255).
*
* RETURNS:
*
* The string is copied into the given buffer space.
*
*/

EXPORT_PREFIX void 
CU_TXGetPString(const CUt_Text *str, char *pstr, tUInt1 max_len)
{
  /* Error check */
  GOODMEM(str);
  GOODMEM(pstr);
  ASSERT(max_len <= 255);

  /* Do it */
  memcpy((void *) &pstr[1], CU_TXGetCStringPtr(str), 
         CU_MIN(max_len, CU_TXGetLength(str)));
  pstr[0] = (char) CU_MIN(255, CU_TXGetLength(str));
}


/***************************************************************************
* 
* FUNCTION: 
*
* CU_TXCompareText()
* 
* DESCRIPTION: 
* 
* Compares two texts.  Returns TRUE if they are the same; FALSE if not.
* 
* PARAMETERS: 
* 
* const CUt_Text *           -- The CU text
* const char *               -- The text to compare with
* tUInt4                     -- Length of compare text
*
* RETURNS:
*
* tBoolean                   -- TRUE if same; FALSE if not.
* 
*/

EXPORT_PREFIX tBoolean 
CU_TXCompareText(const CUt_Text *txt1, const char *str, tUInt4 len)
{
  tUInt2 ret_val;

  GOODMEM(txt1);

  /* Degenerate cases */
  if (( CU_TXGetCStringPtr(txt1) == NULL ) && (str == NULL))
    return TRUE;
  if (CU_TXGetCStringPtr(txt1) == NULL)
    return FALSE;
  if (str == NULL)
    return FALSE;

  /* Check lengths */
  if (CU_TXGetLength(txt1) != len)
    return FALSE;

  /* Check text */
  ret_val = (tUInt2) CU_MemCmp(CU_TXGetCStringPtr(txt1), str, (size_t) len);
  if (ret_val == 0)
    return TRUE;
  else
    return FALSE;

}


/***************************************************************************
* 
* FUNCTION: 
*
* CU_TXCompareCString()
* 
* DESCRIPTION: 
* 
* Compares the CU text with the given C string.  Returns TRUE if they are 
* the same; FALSE if not.
* 
* PARAMETERS: 
* 
* const CUt_Text *           -- The CU text
* const char *               -- The C string to compare with
*
* RETURNS:
*
* tBoolean                   -- TRUE if same; FALSE if not.
* 
*/

EXPORT_PREFIX tBoolean 
CU_TXCompareCString(const CUt_Text *txt1, const char *str)
{
  GOODMEM(txt1);

  /* Degenerate cases */
  if ((str == NULL) && (txt1->text == NULL))
    return TRUE;
  if (str == NULL)
    return FALSE;
  if (txt1->text == NULL)
    return FALSE;

  /* Compare */
  return CU_TXCompareText(txt1, str, strlen(str));

}


/***************************************************************************
* 
* FUNCTION: 
*
* CU_TXComparePString()
* 
* DESCRIPTION: 
* 
* Compares the CU text with the given Pascal string.  Returns TRUE if they 
* are the same; FALSE if not.
* 
* PARAMETERS: 
* 
* const CUt_Text *           -- The CU text
* const char *               -- The pascal string to compare with
*
* RETURNS:
*
* tBoolean                   -- TRUE if same; FALSE if not.
* 
*/

EXPORT_PREFIX tBoolean 
CU_TXComparePString(const CUt_Text *txt1, const char *pstr)
{
  /*
  * Error check 
  */ 

  GOODMEM(txt1);

  /* Degenerate case */
  if ((pstr == NULL) && (txt1->text == NULL))
    return TRUE;
  if (pstr == NULL)
    return FALSE;
  if (txt1->text == NULL)
    return FALSE;

  /* Compare */
  return CU_TXCompareText(txt1, &pstr[1], (tUInt4) ((tUInt1) pstr[0]));

}


/***************************************************************************
* 
* FUNCTION: 
*
* CU_TXCompareCUText()
* 
* DESCRIPTION: 
* 
* Compares two CU texts.  Returns TRUE if they are the same; FALSE if not.
* 
* PARAMETERS: 
* 
* const CUt_Text *           -- One text
* const CUt_Text *           -- The other text
*
* RETURNS:
*
* tBoolean                   -- TRUE if same; FALSE if not.
* 
*/

EXPORT_PREFIX tBoolean 
CU_TXCompareCUText(const CUt_Text *txt1, const CUt_Text *txt2)
{
  GOODMEM(txt1);
  GOODMEM(txt2);

  /* Degenerate cases */
  if ((txt1->text == NULL) && (txt2->text == NULL))
    return TRUE;
  if (txt1->text == NULL)
    return FALSE;
  if (txt2->text == NULL)
    return FALSE;

  /* Compare */
  return CU_TXCompareText(txt1, CU_TXGetCStringPtr(txt2), 
                          CU_TXGetLength(txt2));
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXTruncate()
*
* DESCRIPTION: 
*
* Truncate the text to the given length (does nothing if length given
* is greater than current length).
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text to truncate
* tUInt4              -- The new length
*
* RETURNS:
*
* tError              -- kNoError or kOutOfMemory; can fail because
*                        in order to truncate it may in some cases
*                        have to make a copy of a "borrowed" C string
*                        pointer.
*
*/

EXPORT_PREFIX tError 
CU_TXTruncate(CUt_Text *text, tUInt4 len)
{
  GOODMEM(text);
  
  /* Degenerate case */
  if (len >= text->length)
    return kNoError;

  /* If borrowed pointer, need to make a copy */
  if (text->borrowed) {
    char *new_str;

    /* Copy truncated text */
    new_str = (char*) CU_Malloc(len + 1);
    if (new_str == NULL)
      return kOutOfMemory;
    memcpy(new_str, text->text, len);
    new_str[len] = '\0';

    /* Set values in structure */
    text->text = new_str;
    text->length = len;
    text->buffer_size = len + 1;
    text->borrowed = FALSE;    
  }

  /* Else just do quick adjust */
  else {
    text->length = CU_MIN(len, text->length);
    text->text[text->length] = '\0';
  }

  /* Done */
  return kNoError;
}


/****************************************************************************
* FUNCTION: 
*
* CU_TXExpandBuffer()
*
* DESCRIPTION: 
*
* Expand the string buffer to the given length (using the currently
* set block size)
*
* PARAMETERS: 
*
* CUt_Text *          -- The Text to append to
* tUInt4              -- The new length
*
* RETURNS:
*
* tError              -- kNoError or kOutOfMemory
*
*/

EXPORT_PREFIX tError 
CU_TXExpandBuffer(CUt_Text *str, tUInt4 len)
{
  tUInt4 total_length;
  char *temp_str = NULL;

  ASSERT(str->block_size != 0);

  /* Check if no work to be done; we don't reduce the size */
  if ( len == 0 || len < str->length )
    return kNoError;

  /* Determine total length of buffer based on block size */
  total_length = len + (str->block_size - (len % str->block_size));

  /* Hang onto "borrowed" C string */
  if (str->borrowed) {
    temp_str = str->text;
    str->text = NULL;
    str->borrowed = FALSE;
  }

  /* Allocate memory */
  if ( str->text == NULL ) {
    str->text = (char *) CU_Malloc(total_length);
  } else {
    str->text = (char *) CU_Realloc(str->text, total_length);
  }

  /* Check for errors */
  if ( str->text == NULL ) {
    str->length = str->buffer_size = 0;
    return kOutOfMemory;
  }

  /* Copy in borrowed string, if any */
  if (temp_str != NULL)
    CU_MemCpy(str->text, temp_str, str->length);

  /* Done */
  str->buffer_size = total_length;
  return kNoError;
}


/***************************************************************************
* 
* FUNCTION: CU_TXCopy
*
* DESCRIPTION: 
*
* Copies contents of src text into the dest text structure; dest must be clear
* of any content
* 
* PARAMETERS: 
*
* CUt_Text *dest       -- Destination
* const  CUt_Text *src -- Source
* 
* RETURNS:
*
* tError               -- kNoError or kOutOfMemory 
* 
*/
EXPORT_PREFIX tError 
CU_TXCopy(CUt_Text *dest, const CUt_Text *src )
{
  /* Error check */
  GOODMEM(dest);
  GOODMEM(src);
  ASSERT(dest->length == 0);

  /* If src is empty, simply return */
  if ( src->length == 0 )
    return kNoError;
    
  /* Set up block size in dest and expand its buffer */
  dest->block_size = src->block_size;
  if ( CU_TXExpandBuffer(dest, src->length) == kOutOfMemory )
    return kOutOfMemory;

  /* Memcpy the data fro src to dest and '\0' terminate */
  CU_MemCpy(dest->text, src->text, src->length);
  dest->length = src->length;
  dest->borrowed = FALSE;
  dest->text[dest->length] = '\0';

  return kNoError; /* Success */  
}


/****************************************************************************
* FUNCTION:
*
* CU_TXInsertText()
*
* DESCRIPTION:
*
* Inserts characters into the given text at the a position, right shifting
* existing characters if need be
*
* PARAMETERS:
*
* CUt_Text* text  -- Text to modify
* tUInt4 pos      -- Where to insert the string
* const char* str -- String to insert
* tUInt4 str_len  -- Length of string to insert
* 
* RETURNS:
*
* tError -- kNoError or kOutOfMemory
*
*/
tError
CU_TXInsertText(CUt_Text* text, tUInt4 pos, const char* str, tUInt4 str_len)
{
  tError err;
  tUInt4 i;

  /* Error check */
  GOODMEM(text);
  ASSERT(pos <= text->length);
  if ( str_len == 0 )
    return kNoError;
  GOODMEM(str);

  /* Expand text */
  err = CU_TXExpandBuffer(text, text->length + str_len);
  if ( err != kNoError )
    return err;

  /* Right shift chars */
  for ( i = text->length; i > pos; i-- )
    text->text[(i - 1) + str_len] = text->text[i - 1];
  text->length += str_len;

  /* Copy in str */
  for ( i = 0; i < str_len; i++ )
    text->text[pos + i] = str[i];

  return kNoError;
}


/****************************************************************************
* FUNCTION:
*
* CU_TXDeleteRange()
*
* DESCRIPTION:
*
* Deletes a range of chars from the text buffer
*
* PARAMETERS:
*
* CUt_Text* text   -- Text to modify
* tUInt4 start_pos -- Start of range to delete
* tUInt4 length    -- Length of range to delete
* 
* RETURNS:
*
* tError              -- kNoError or kOutOfMemory; can fail because
*                        in order to truncate it may in some cases
*                        have to make a copy of a "borrowed" C string
*                        pointer.
*
*/
tError
CU_TXDeleteRange(CUt_Text* text, tUInt4 start_pos, tUInt4 length)
{
  tUInt4 i;

  /* Error check */
  GOODMEM(text);
  ASSERT(start_pos + length <= text->length);

  if ( length == 0 )
    return kNoError;

  /* If borrowed pointer, need to make a copy */
  if (text->borrowed) {
    char *new_str;

    /* Copy truncated text */
    new_str = (char*) CU_Malloc(text->length + 1);
    if (new_str == NULL)
      return kOutOfMemory;
    memcpy(new_str, text->text, text->length);
    new_str[text->length] = '\0';

    /* Set values in structure */
    text->text = new_str;
    text->length = text->length;
    text->buffer_size = text->length + 1;
    text->borrowed = FALSE;    
  }

  /* Left shift chars */
  for ( i = start_pos; i + length < text->length; i++ )
    text->text[i] = text->text[i + length];
  text->length -= length;
  text->text[text->length] = '\0';

  return kNoError;
}

/****************************************************************************
* FUNCTION:
*
* CU_TXClearSequence()
*
* DESCRIPTION:
*
* Clears sequence.
*
* PARAMETERS:
*
* CUt_TextSeq* seq   -- Sequence to clear
* 
* RETURNS:
*
*
*/
void
CU_TXClearSequence(CUt_TextSeq* seq)
{
  tUInt4 i;

  /* Error check */
  GOODMEM(seq);

  if ( seq->_length == 0 )
    return;

  for (i = 0; i < seq->_length; i++)
    CU_TXClear(&seq->_buffer[i]);


}


/* End of CU_Text.c */

