/***************************************************************************
*
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE: CU_Thread.c
* 
* AUTHOR: 
* 
* Stephan R.A. Deibel
*
* CREATION DATE: 
* 
* 12/15/92
*
* VERSION: 
* 
* 8/4/94
*
* DESCRIPTION: 
* 
* A cross-platform threads package for multi-stack non-preemptive thread
* multitasking on a variety of hardware and compiler platforms.
* 
* NOTES: 
* 
* This package is written to maintain seperate complete stack environments
* for each thread (including some critical low-memory globals where necessary) 
* allowing allocation of any sort of objects on the stack of each thread.  
* Seperate stacks are swapped in through register manipulation, making 
* context switches very fast (even on machines that don't have OS support
* for threads).
*
* The major shortcoming of this package currently is lack of debug-time
* verification of stack usage.  So for now, you are on your own in making
* sure that you are allocating enough stack space for each thread!
*
* There are four ports of this package that allow it to run in a variety
* of environments.  These are selected with compiler variables defined
* in your prefix:
*
* 1) The initial implementation was with setjmp/longjmp and clever 
*    manipulation of the setjmp_buf.  This runs on the Mac 68K but is hard 
*    to port to some platforms simply because it is often hard to get
*    technical information on the structure of the setjmp_buf.
*    To use this port, #define SETJMPTHREADS.
* 2) Port to SunOS4 was acheived through use of the Light Weight Process
*    (LWP) facility.  To use this port, #define SUNLWP.
* 3) Port to HP/UX and other machines was acheived through use of the
*    QuickThreads package.  To use this port, #define QUICKTHREADS.
* 4) Port to Mac ThreadManager for use on all Macs.  To use this port,
*    #define MACTHREADMANAGER.
*
* Currently, the following platforms and compilers are supported:
*
* - Macintosh System 7.x 68K or PPC architecture with Symantec 7.x C/C++ or
*   Metrowerks DR6.  The MACTHREADMANAGER port is untested on Symantec.
* - SunOS 4.x with gnu C using LWP
* - HP/UX 9.0.5 with gnu C using QuickThreads
* - Other UNIX machines should also work because QuickThreads will run
*   on those machines.  This has not been tested, however.
* - MS Windows 95 and MS Windows NT (Win32) with Borland C++
*
* Notes for all users:
*
* - Make sure you call CU_StartTasking() before calling other functions
*   in this package.  If you fail to do this, you will trigger ASSERT()'s
*   in DEBUG mode and get program errors when DEBUG is off.
* - Make sure you call CU_ApplStatusChanged() to keep the CUg_ApplStatus
*   global up to date in applications that are daemons (always background)
*   or in applications that switch in and out of the foreground.  Default
*   starting state is CUk_InForeground.
* - Make sure you #define the appropriate port selector variable
*   (SETJMPTHREADS, SUNLWP, QUICKTHREADS, or MACTHREADMANAGER).
* - #define macintosh, sunos4, or hpux9 as appropriate.
* - #define PREEMPTIVEMULTI if and only if you are on a preemptive 
*   multitasking system.  As explained below, CU_Thread includes some
*   optimizations for non-preemptive systems that must be disabled on
*   preemptive machines.  [Note that this refers to preemptive process
*   switching and not to preemptive multithreading.  CU_Thread does not
*   and will never support preemptive multithreading.]
*
* Notes for Macintosh Thread Manager users:
*
* - If you are using the Thread Manager on Mac, you need to make sure that
*   MaxApplZone() is called early in the application's execution.
*   This is not included in CU_StartTasking() since that function might
*   be called multiple times in some implementations, and it might be called
*   too late (e.g., after calls to MoreMasters() have already been made).
*
* Notes for Macintosh Symantec C/C++ users:
*
* - Symantec "Defer and combine stack adjust" optimization does not work!
* - Other optimizations may also screw things up; symantec is a bit odd.
* - Possibly doesn't work with THINK profiler; some other threads packages
*   have had problems with this.
* - There may be problems with "4 byte ints" (bus error when freeing stacks 
*   upon call to CU_StopTasking()).
* - Symantec is not actively supported; use Metrowerks instead!
*
* Notes for Sun OS 4.x LWP users:
*
* - This was ported using gnu C but since the light weight process (LWP)
*   library is used, it will probably work on most other compilers.
* - Implementation on Sun OS 4.x replaces use of setjmp() and longjmp()
*   with somewhat non-traditional use of the lwp.  Basic strategy is as 
*   follows:
*
*   1) All threads are set to the same priority so switching between
*      them is non-preemptive.
*   2) Local structures are maintained but lwp_create() and lwp_destroy()
*      as used as well to create/destroy lwp's.  The user procedure is
*      wrappered because our parameter set is different.
*   3) Replace use of setjmp() and longjmp() with calls to lwp_yield()
*      and an extra CU_RestoreEnvironment() call after each yield.
*
* - Note that the Sun port context switches about an order of magnitude 
*   slower than the Mac implementation.  This could probably be optimized
*   but it may be due largely to the way LWP works.
*
* Notes for QuickThreads, Win32 and Mac ThreadManager users:
*
* - These are recent ports so there may still be problems; it has only
*   been tested on HP/UX 9.0.5 so far.
*
* Notes for users on non-preemptive multitasking systems:
*
* This package includes some basic support for yielding time to other
* applications in a non-preemptive environment in a manner that is
* adaptive to:
*
*   - Whether or not the application is in foreground
*   - Whether there are any unblocked threads.
*   - Recent thread activity.
*   - Existence of a thread that is blocked waiting on action of a 
*     local (same CPU) peer process.
*
* Under non-preemptive systems, you must start a thread that is running
* CU_ProcessYielder() (provided with this package) in order to yield
* time to other applications on your system in an intelligent manner.  This
* routine manages inter-process (*not* inter-thread) time-yielding.
*
* In addition to starting the process yielder thread, you must also keep
* up to date the status information that are used by that thread.  To do 
* this you need to:
*
* 1) Call CU_ApplStatusChanged() as needed, when the application switches
*    into and out of the foreground.
* 2) Set the 'local_peer' parameter in CU_BlockThread() accordingly.
*
* Note that on preemptive systems, the CU_ProcessYielder thread will
* exit immediately and calls to associated routines listed about (other
* than CU_ApplStatusChanged) will be ignored.
*
* TO DO:
*
* - Add checking of stack space somehow?
* - Must fill in failure management!
* - On Mac, could preallocate a thread pool and reuse threads in there
*   instead of allocating them individually as we do now.  This would
*   speed up applications that create alot of short-lived threads.
*
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
* 
* $Log: CU_Thread.c,v $
* Revision 1.2  1998/07/15 05:20:22  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
 * Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
 * Converting to CVS
 *
* Revision 1.15  1997/03/12 22:36:32  jrequena
* Minor change to comment
*
* Revision 1.14  1996/07/02  14:11:36  sdeibel
* Fixed use of UNUSED macro which prevented compilation
*
* Revision 1.13  1996/06/25  21:21:11  epattiso
* changed to compile on Mac
*
* Revision 1.12  1996/06/21  19:07:40  epattiso
* changed to compile on Mac
*
* Revision 1.11  1996/06/19  20:02:40  jehresma
* Initial MS Visual C++ port -- There are still a few bugs, though
*
* Revision 1.10  1996/04/12  20:51:43  jehresma
* Win32 threads implementation
*
* Revision 1.9  1996/04/02  22:20:01  sdeibel
* Added code to compile nothing if CUk_NOTHREADS is defined
*
* Revision 1.8  1996/01/23  23:32:43  sdeibel
* Misc changes associated with introducing CU_Text.*
*
 * Revision 1.7  1996/01/12  15:15:27  sdeibel
 * Removed some garbage!
 *
 * Revision 1.6  1996/01/11  18:50:55  sdeibel
 * Ported to Mac Thread Manager so it will run native on PPN
 * (that's PPC) ... and also of course will be safer on 68K Macs
 *
 * Revision 1.5  1996/01/10  20:07:33  sdeibel
 * Port back to Mac with partial port to PPC / Thread Manager
 * (to be continued)
 *
*
* 12/15/93 Stephan   Created file / initial implementation
* 8/4/94   Stephan   Rewrote based on some code from Michael Hecht
*                    and some creative hacking to swap out stacks
*                    rather than copying them on each context switch.
*                    Also applied DSG naming and commenting conventions.
* 8/8/94   Stephan   Ported to SunOS 4.x.
* 8/10/94  Stephan   Augmented blocking facility.
* 9/21/94  Stephan   Added support for smart process yielding on non-
*                    preemptive systems.  Also some formatting/cosmetics.
* 7/18/95  Stephan   Ported to QuickThreads, so it will run on HP/UX 9.0.5
*                    among other systems.
* 
*/

#ifndef CUk_NOTHREADS

#include <string.h>
#include <stdio.h>

#ifdef dos 
#include "CUFail.h"
#include "CUDebug.h"
#include "CUTime.h"
#include "CUThread.h"
#include "CUThrd_p.h"
#else
#include "CU_Failure.h"
#include "CU_Debug.h"
#include "CU_Time.h"
#include "CU_Thread.h"
#include "CU_Thread.prv.h"
#endif /* dos */

#ifdef WIN32THREADS
#include <process.h>
#endif 

#ifdef macintosh
#include <Events.h>
#endif

/***************************************************************************/
/* GLOBALS */

#ifdef WIN32THREADS
CRITICAL_SECTION CUg_Win32ChangeEventSection;
#endif

/****************************************************************************
* FUNCTION:
*
* CU_StartTasking
*
* DESCRIPTION:
*
* Start thread multitasking.
*
* PARAMETERS:
*
* void *         -- User data value for the master thread.
*
* RETURNS:
*
* CUt_ThreadID   -- The master thread ID.
*
* NOTES:
* 
* This function sets up the master thread and starts the process of
* multi-tasking threads.
*
* The master task is assumed to be running already when this function
* is called so no stack is allocated.  The master task keeps running 
* until it calls CU_ThreadYield() at some later point.
*
* Calling this function twice (without calling CU_StopTasking() in
* between) is a program error.
*
*/

CUt_ThreadID 
CU_StartTasking(void *data)
{
  CUt_ThreadEntry *thread = NULL;
#ifdef WIN32THREADS
  HANDLE pseudo_handle;
#endif

  ASSERT(CUg_ThreadControl.master == NULL);

  /* Set local peer block counter */
  CUg_ThreadControl.num_local_peers = 0;
  
  /* Create master thread and fill with 0's */
  thread = (CUt_ThreadEntry *) CU_MallocSet(sizeof(CUt_ThreadEntry), 0);
  if (thread == NULL) {
    UNIMPLEMENTED;
  }
  
  /* Initialize non-0 fields of master thread */
  thread->user_data = data;
  thread->ID = ++CUg_ThreadControl.ID_seed;
  thread->status = CUk_Running;

  /* On Sun, hang onto LWP ID for self as well as internal ID */
#if defined(SUNLWP)
  if (lwp_self(&thread->lwp_ID) != 0)
    UNIMPLEMENTED;
#endif

  /* On Mac with ThreadManager, hang onto thread ID for self as well as internal ID */
#if defined(MACTHREADMANAGER)
  if (GetCurrentThread(&thread->tm_ID) != noErr)
    UNIMPLEMENTED;
#endif

#if defined(macintosh) && defined(SETJMPTHREADS)
  /* Set up Macintosh LM globals needed for seperate stacks */
  thread->env.heap_end = LMGetHeapEnd();
  thread->env.appl_limit = LMGetApplLimit();
  thread->env.hi_heap_mark = LMGetHiHeapMark();
#endif

#ifdef QUICKTHREADS
  /* Set up master stack/thread as NULL for now; will be set later */
  thread->env.stack = NULL;
  thread->qthread = NULL;
#endif
  
#ifdef WIN32THREADS
  /* Get the handle to the current thread and duplicate it so that other threads
     may use it */
  pseudo_handle = GetCurrentThread();
  if ( DuplicateHandle(GetCurrentProcess(), pseudo_handle, GetCurrentProcess(),
                       &thread->win32_thread, 0, FALSE, DUPLICATE_SAME_ACCESS)
       != TRUE )
    UNIMPLEMENTED;
  thread->suspended = FALSE;

  thread->win32_event = CreateEvent(NULL, TRUE, FALSE, NULL);
  if ( thread->win32_event == NULL )
    UNIMPLEMENTED;  

  InitializeCriticalSection(&CUg_Win32ChangeEventSection);
#endif

  /* Set up thread control block and thread list */
  CU_InsertTLEntry(thread);

  /* Return the ID of the master thread */
  return(thread->ID);
}


/****************************************************************************
* FUNCTION:
*
* CU_StopTasking
*
* DESCRIPTION:
*
* Stop thread multitasking.
*
* NOTES:
* 
* This function kills all thread entries including the master thread
* entry and deallocates the threads package control information.  Once
* this function is called, calling other thread package routines
* (including CU_ThreadYield()) is a program error.
*
* This function should only be called from the master thread after 
* CU_StartTasking() has  already been called.  Anything else is a 
* program error.
*
* A subsequent call to CU_StartTasking() is allowed.
*
* The master task is left running.
*
*/

void 
CU_StopTasking()
{
  CUt_ThreadEntry *thread;
  CUt_ThreadEntry *next;
#ifdef WIN32THREADS
  DWORD ret;
#endif
  
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(CUg_ThreadControl.current == CUg_ThreadControl.master);
  
  /* Remove all entries from the thread list and dispose; note that master is
     always at head of thread linked list */
  thread = CUg_ThreadControl.master;  
  while (thread != NULL) {
    /* Save next entry */
    next = thread->next_entry;
    
    /* On SUNLWP, dispose of the LWP (except for master) */
#if defined(SUNLWP)
    if (thread != CUg_ThreadControl.master)
      if (lwp_destroy(thread->lwp_ID) != 0)
        UNIMPLEMENTED;
#endif

    /* On Mac with ThreadManager, dispose of the thread (except for master) */
#if defined(MACTHREADMANAGER)
    if (thread != CUg_ThreadControl.master)
      if (DisposeThread(thread->tm_ID, NULL, FALSE) != noErr)
        UNIMPLEMENTED;
#endif

    /* On WIN32, dispose of the thread (except for master) and release its 
       associated event object */
#if defined(WIN32)
    if (thread != CUg_ThreadControl.master)
      CU_DestroyWin32Thread(thread);

    ret = CloseHandle(thread->win32_event);
    ASSERT(ret == TRUE);
    thread->win32_event = NULL;
/*
    ret = CloseHandle(thread->win32_thread);
    ASSERT(ret == TRUE);
    thread->win32_thread = NULL;
*/
#endif

    /* Remove thread from thread list */
    CU_RemoveTLEntry(thread);

    /* Diispose thread memory if not the master thread.  We don't delete 
       the master entry now because the WIN32 code depends on it being
       around */
    /* Actually env.stack is always NULL for now since we enforce */
    /* this to be called from the master thread only! [is this right? jpe] */
    if ( thread != CUg_ThreadControl.master ) {
      if (thread->env.stack != NULL)
        CU_Free(thread->env.stack);
      CU_Free(thread);
    }

    /* Go to next thread or stop if none left */
    if ( CUg_ThreadControl.current == NULL )
      thread = NULL;
    else
      thread = next;
  }
  
#ifdef WIN32THREADS
  DeleteCriticalSection(&CUg_Win32ChangeEventSection);
#endif

  /* Finally, free the master thread structure and reset fields in the global
     thread control */
  CU_Free(CUg_ThreadControl.master);
  CUg_ThreadControl.master = NULL;

  /* That's all! */
}


/****************************************************************************
* NOTE: CU_CreateThread is maintained for each running environment because
* it is too nasty to read otherwise!
*
* For future ports, could probably use one or the other depending on
* whether a threads package is being used or whether setjmp() and longjmp()
* are used.
****************************************************************************/

#ifdef SETJMPTHREADS
/****************************************************************************
* FUNCTION:
*
* CU_CreateThread
*
* DESCRIPTION:
*
* Create a new thread.
*
* PARAMETERS:
*
* CUt_ThreadProc   -- The procedure for the new thread.
* void *           -- User data value for the new thread.
* size_t           -- Size of stack (in bytes) to allocate for new thread,
*                     or 0 to use default.
*
* RETURNS:
*
* CUt_ThreadID     -- The new thread's ID.
*
* NOTES:
* 
* The task is not run until later when CU_TaskYield is called and
* it comes up as being eligible to run according to the thread
* scheduler function CU_NextScheduledThread().
*
*/

CUt_ThreadID 
CU_CreateThread(CUt_ThreadProc proc, void *data, size_t stack_size)
{
  CUt_ThreadEntry *thread = NULL;
  
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(proc != NULL);
  ASSERT(stack_size >= 0);

  /* Create new thread and fill with 0's */
  thread = (CUt_ThreadEntry *) CU_MallocSet(sizeof(CUt_ThreadEntry), 0);
  if (thread == NULL) {
    UNIMPLEMENTED;
  }

  /* Initialize non-0 fields of new thread */
  thread->ID = ++CUg_ThreadControl.ID_seed;
  thread->proc = proc;
  thread->status = CUk_Ready; /* Doing this now assumes no preemption! */
  thread->user_data = data;
  
  /* Create a stack for the thread */
  if (stack_size == 0)
    stack_size = CU_DefaultStackSize();
#ifdef MS_WIN
  thread->env.stack = _nmalloc(stack_size);
#else
  thread->env.stack = CU_Malloc(stack_size);
#endif /* MS_WIN */
  if (thread->env.stack == NULL) {
    UNIMPLEMENTED;
  }

#ifdef macintosh
  /* Set up Macintosh LM globals needed for seperate stacks */
  thread->env.heap_end = thread->env.stack;
  thread->env.appl_limit = thread->env.stack;
  thread->env.hi_heap_mark = thread->env.stack;
#endif

   /* Save the current environment for later longjmp() */
  if (setjmp(thread->env.setjmp_env) == CUk_RunThread) {
  
    /*
    * Arrive here when the new thread is started (some time
    * after initial execution of this routine has ended and
    * someone has called CU_ThreadYield().
    */
    
    /* Restore the rest of this thread's environment not
    * already set up by longjmp() */
    CU_RestoreEnvironment();
        
    /* Run the thread! */
    CUg_ThreadControl.current->proc(CUg_ThreadControl.current->user_data);
          
    /* When the thread procedure terminates, thread is done so we kill it */
    /* and start the next thread.  Note that not all threads terminate like */
    /* this; some run forever and get killed from the outside. */
    CU_DestroyThreadEntry(CUg_ThreadControl.current);

    /* Above should never return */
    ASSERT(0);
  }
  
  /*
  * Arrive here when setjmp() is initially called and has saved the
  * environment above
  */

  /* Bootstrap!  Since we are actually running using someone else's stack */
  /* right now, we need to fake the environment to use the private stack */
  /* when the thread is actually started later. */
  /* This is Platform-dependent; fill in for each new SETJMP impl platform!! */
#ifdef macintosh
#ifdef __MWERKS__
  thread->env.setjmp_env[a7] = (long *) ((long) thread->env.stack + stack_size);
  thread->env.setjmp_env[a6] = (long *) 0;
#else
  thread->env.setjmp_env[a7] = (long) thread->env.stack + stack_size;
  thread->env.setjmp_env[a6] = 0;
#endif
#endif
#ifdef MS_WIN 
  if ( ((tUInt2)thread->env.stack + (tUInt4)stack_size - (tUInt4)sizeof(long)) > 0x0000FFFF ) {
    ASSERT(0);
  }
  thread->env.setjmp_env[sp] = (int)((long)thread->env.stack + (long)stack_size - (long)sizeof(long));
/*  *((CUt_ThreadProc *) &(thread->env.setjmp_env[pc])) = proc; */
/*  thread->env.setjmp_env[ds] = (int) ((long) ((char far *) &thread) >> 16);  */ 
  /* Save the PC(?) here */
#ifdef DEBUG
  thread->env.pc0 = thread->env.setjmp_env[pc];
  thread->env.pc1 = thread->env.setjmp_env[pc+1];
  thread->env.sp = thread->env.setjmp_env[sp];
  thread->env.env0 = thread->env.setjmp_env[0];
#endif 
#endif /* MS_WIN */

  /* Add new thread to process list */
  CU_InsertTLEntry(thread);
  
  /* Return new thread ID */
  return(thread->ID);
}
#endif /* SETJMPTHREADS */


#if defined(SUNLWP)
/****************************************************************************
* FUNCTION:
*
* CU_CreateThread
*
* DESCRIPTION:
*
* Create a new thread.
*
* PARAMETERS:
*
* CUt_ThreadProc   -- The procedure for the new thread.
* void *           -- User data value for the new thread.
* size_t           -- Size of stack to allocate for new thread,
*                     or 0 to use default.
*
* RETURNS:
*
* CUt_ThreadID     -- The new thread's ID.
*
* NOTES:
* 
* The task is not run until later when CU_TaskYield is called and
* it comes up as being eligible to run according to the thread
* scheduler function CU_NextScheduledThread().
*
*/

CUt_ThreadID CU_CreateThread(CUt_ThreadProc proc,
  void *data, size_t stack_size)
{
  CUt_ThreadEntry *thread = NULL;
  
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(proc != NULL);
  ASSERT(stack_size >= 0);

  /* Create new thread and fill with 0's */
  thread = (CUt_ThreadEntry *) CU_MallocSet(sizeof(CUt_ThreadEntry), 0);
  if (thread == NULL) {
    UNIMPLEMENTED;
  }

  /* Initialize non-0 fields of new thread */
  thread->ID = ++CUg_ThreadControl.ID_seed;
  thread->proc = proc;
  thread->status = CUk_Ready; /* Doing this now assumes no preemption! */
  thread->user_data = data;
  
  /* Create a stack for the thread */
  if (stack_size == 0)
    stack_size = CU_DefaultStackSize();
  thread->env.stack = CU_Malloc(stack_size);
  if (thread->env.stack == NULL) {
    UNIMPLEMENTED;
  }

  /* Create a LWP entry for this thread (create it in suspended state */
  /* and with same priority as all others to avoid preemption!) */
  if (lwp_create(&thread->lwp_ID, CU_SunThreadProc, pod_getmaxpri(), 
                       LWPSUSPEND & LWPNOLASTRITES, STKTOP(thread->env.stack),
                       3, proc, data, thread) != 0)
    UNIMPLEMENTED;
                 
  /* Add new thread to process list */
  CU_InsertTLEntry(thread);
  
  /* Return new thread ID */
  return(thread->ID);
}
#endif /* SUNLWP */


#ifdef QUICKTHREADS
/****************************************************************************
* FUNCTION:
*
* CU_CreateThread
*
* DESCRIPTION:
*
* Create a new thread.
*
* PARAMETERS:
*
* CUt_ThreadProc   -- The procedure for the new thread.
* void *           -- User data value for the new thread.
* size_t           -- Size of stack (in bytes) to allocate for new thread,
*                     or 0 to use default.
*
* RETURNS:
*
* CUt_ThreadID     -- The new thread's ID.
*
* NOTES:
* 
* The task is not run until later when CU_TaskYield is called and
* it comes up as being eligible to run according to the thread
* scheduler function CU_NextScheduledThread().
*
*/

/*
* Local utility to bootstrap the thread on its stack
*/

static void
CU_QTStartThread(void *data, void *thread, qt_userf_t *fct)
{
  (*(CUt_ThreadProc)(fct))(data);
  CU_DestroyThreadEntry((CUt_ThreadEntry *) thread);
  ASSERT(0);
}


CUt_ThreadID CU_CreateThread(CUt_ThreadProc proc,
  void *data, size_t stack_size)
{
  CUt_ThreadEntry *thread = NULL;
  void *aligned_stack;

  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(proc != NULL);
  ASSERT(stack_size >= 0);

  /* Create new thread and fill with 0's */
  thread = (CUt_ThreadEntry *) CU_MallocSet(sizeof(CUt_ThreadEntry), 0);
  if (thread == NULL) {
    UNIMPLEMENTED;
  }

  /* Initialize non-0 fields of new thread */
  thread->ID = ++CUg_ThreadControl.ID_seed;
  thread->proc = proc;
  thread->status = CUk_Ready;
  thread->user_data = data;
  
  /* Create a stack for the thread */
  if (stack_size == 0)
    stack_size = CU_DefaultStackSize();
  thread->env.stack = CU_Malloc(stack_size);
  if (thread->env.stack == NULL) {
    UNIMPLEMENTED;
  }

  /* Create a QuickThread for this thread.  Stack is aligned as */
  /* needed by platform and thread is created then set up for */
  /* initial run with QT_StartThread routine.  Only the thread */
  /* parameter to QT_ARGS is actually used; the other information */
  /* is obtained from the process list when CU_StartThreadEntry */
  /* is called. */
  aligned_stack = CU_STKALIGN(thread->env.stack);
  thread->qthread = QT_SP(aligned_stack, stack_size - QT_STKALIGN);
  thread->qthread = QT_ARGS(thread->qthread, data, thread, proc, 
                            CU_QTStartThread);

  /* Add new thread to process list */
  CU_InsertTLEntry(thread);
  
  /* Return new thread ID */
  return(thread->ID);
}
#endif /* QUICKTHREADS */


#if defined(MACTHREADMANAGER)
/****************************************************************************
* FUNCTION:
*
* CU_CreateThread
*
* DESCRIPTION:
*
* Create a new thread.
*
* PARAMETERS:
*
* CUt_ThreadProc   -- The procedure for the new thread.
* void *           -- User data value for the new thread.
* size_t           -- Size of stack to allocate for new thread,
*                     or 0 to use default.
*
* RETURNS:
*
* CUt_ThreadID     -- The new thread's ID.
*
* NOTES:
* 
* The task is not run until later when CU_TaskYield is called and
* it comes up as being eligible to run according to the thread
* scheduler function CU_NextScheduledThread().
*
*/

CUt_ThreadID CU_CreateThread(CUt_ThreadProc proc,
  void *data, size_t stack_size)
{
  CUt_ThreadEntry *thread = NULL;
  CUt_MacThreadProcData *tm_data = NULL;
  
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(proc != NULL);
  ASSERT(stack_size >= 0);

  /* Create new thread and fill with 0's */
  thread = (CUt_ThreadEntry *) CU_MallocSet(sizeof(CUt_ThreadEntry), 0);
  if (thread == NULL) {
    UNIMPLEMENTED;
  }

  /* Initialize non-0 fields of new thread */
  thread->ID = ++CUg_ThreadControl.ID_seed;
  thread->proc = proc;
  thread->status = CUk_Ready; /* Doing this now assumes no preemption! */
  thread->user_data = data;
  
  /* Create data structure for CU_MacThreadProc call */
  tm_data = (CUt_MacThreadProcData *) CU_Malloc(sizeof(CUt_MacThreadProcData));
  if (tm_data == NULL)
    UNIMPLEMENTED;
  tm_data->proc = proc;
  tm_data->data = data;
  tm_data->thread = thread;
    
  /* Create a ThreadManager thread for this thread */
  if (stack_size == 0)
    stack_size = CU_DefaultStackSize();
  if (NewThread
        (kCooperativeThread,
         (pascal void * (*)(void *)) CU_MacThreadProc, 
         tm_data, 
         stack_size,
         (kCreateIfNeeded + kNewSuspend), 
         NULL, 
         &thread->tm_ID) != noErr)
    UNIMPLEMENTED;
    
  /* Set thread up ready to run later */
  if (SetThreadState(thread->tm_ID, kReadyThreadState, 
                     CUg_ThreadControl.current->tm_ID) != noErr)
    UNIMPLEMENTED;
    
  /* Stack pointer isn't used with the ThreadManager */
  thread->env.stack = NULL;
                 
  /* Add new thread to process list */
  CU_InsertTLEntry(thread);
  
  /* Return new thread ID */
  return(thread->ID);
}
#endif /* MACTHREADMANAGER */


#if defined(WIN32)
/***************************************************************************
* 
* FUNCTION: CU_Win32StartThreadProc
*
* DESCRIPTION: 
*
* Helper function to start a win32 thread and call the CU start procedure
* 
* PARAMETERS: 
*
* void * param -- win32's generic ptr; used here to store ptr to CUt_ThreadEntry
* 
* RETURNS: nothing
* 
* NOTES: 
*
* Since win32 threads must be destroyed from within themselves, we need to
* make sure this thread has not already been destroyed.
* 
*/
#ifdef _BORLANDC_
void _USERENTRY
CU_Win32StartThreadProc(void* param)
#endif
#ifdef MS_VC
unsigned __stdcall
CU_Win32StartThreadProc(void* param)
#endif
{
  CUt_ThreadEntry* thread;

  /* Error check and cast the void param to a thread entry pointer */
  GOODMEM(param);
  thread = (CUt_ThreadEntry*) param;
  GOODMEM(thread->proc);

  /* Check to make sure we're not already marked for destruction */
  if ( thread->status == CUk_Destroyed ) {
#ifdef _BORLANDC_
    _endthread();
#endif
#ifdef MS_VC
    _endthreadex(0);
#endif
  }
  
  /* Destroy any thread marked for destruction */
  if ( CUg_ThreadControl.destroy != NULL ) {
    CU_DestroyWin32Thread(CUg_ThreadControl.destroy);
    ASSERT(CUg_ThreadControl.destroy->env.stack != NULL);
    CU_Free(CUg_ThreadControl.destroy);
    CUg_ThreadControl.destroy = NULL;
  }

  /* Set the thread's status and global current thread */
  CUg_ThreadControl.current = thread;
  thread->status = CUk_Running;

  /* Call the CU user callback */
  (*thread->proc)(thread->user_data);

#ifdef MS_VC
  return 0L;
#endif
}




/****************************************************************************
* FUNCTION:
*
* CU_CreateThread
*
* DESCRIPTION:
*
* Create a new thread.
*
* PARAMETERS:
*
* CUt_ThreadProc   -- The procedure for the new thread.
* void *           -- User data value for the new thread.
* size_t           -- Size of stack to allocate for new thread,
*                     or 0 to use default.
*
* RETURNS:
*
* CUt_ThreadID     -- The new thread's ID.
*
* NOTES:
* 
* The task is not run until later when CU_TaskYield is called and
* it comes up as being eligible to run according to the thread
* scheduler function CU_NextScheduledThread().
*
*/

CUt_ThreadID  CU_CreateThread(CUt_ThreadProc proc,
  void *data, size_t stack_size)
{
  CUt_ThreadEntry *thread = NULL;
#ifdef __BORLANDC__
  WORD win32_id;
#endif
#ifdef MS_VC
  DWORD win32_id;
#endif
  
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(proc != NULL);
  ASSERT(stack_size >= 0);

  /* Create new thread and fill with 0's */
  thread = (CUt_ThreadEntry *) CU_MallocSet(sizeof(CUt_ThreadEntry), 0);
  if (thread == NULL) {
    UNIMPLEMENTED;
  }

  /* Initialize non-0 fields of new thread */
  thread->ID = ++CUg_ThreadControl.ID_seed;
  thread->proc = proc;
  thread->status = CUk_Ready; /* Doing this now assumes no preemption! */
  thread->user_data = data;
  
  /* Create a WIN32 thread in suspended mode for this thread */
  if (stack_size == 0)
    stack_size = CU_DefaultStackSize();
#ifdef _BORLANDC_
  thread->win32_thread = (HANDLE) _beginthreadNT(CU_Win32StartThreadProc,
                                                 stack_size, thread, NULL,
                                                 CREATE_SUSPENDED, &win32_id);
  if ( thread->win32_thread == -1 )
    UNIMPLEMENTED;
#endif
#ifdef MS_VC
  thread->win32_thread =
    (HANDLE) _beginthreadex(NULL, stack_size, CU_Win32StartThreadProc,
                   thread, CREATE_SUSPENDED, &win32_id);
  if ( thread->win32_thread == 0 )
    UNIMPLEMENTED;
#endif

  thread->suspended = TRUE;

  /* Create the win32 "event" to signal this thread when it should run */
  thread->win32_event = CreateEvent(NULL, TRUE, FALSE, NULL);
  if ( thread->win32_event == NULL )
    UNIMPLEMENTED;

  /* Stack pointer isn't used with the WIN32 */
  thread->env.stack = NULL;
                 
  /* Add new thread to process list */
  CU_InsertTLEntry(thread);
  
  /* Return new thread ID */
  return(thread->ID);
}
#endif /* WIN32 */


/****************************************************************************
* FUNCTION:
*
* CU_DestroyThread
*
* DESCRIPTION:
*
* Kill the given thread, remove it from the thread list, and deallocate
* it.
*
* PARAMETERS:
*
* CUt_ThreadID     -- The thread ID.
*
* NOTES:
*
* The master thread is assumed to be killed only after all other threads
* are gone.  Use CU_StopTasking() to kill all threads and halt the tasking
* process.
*
*/

void 
CU_DestroyThread(CUt_ThreadID ID)
{
  CUt_ThreadEntry * thread;

  thread = CU_FindTLEntry(ID);
  ASSERT(thread != NULL);
  CU_DestroyThreadEntry(thread);
}


/****************************************************************************
* FUNCTION:
*
* CU_GetActiveThread
*
* DESCRIPTION:
*
* Get the ID of the currently active thread.
*
* RETURNS:
*
* CUt_ThreadID    -- The ID.
*
*/

CUt_ThreadID 
CU_GetActiveThread()
{
  ASSERT(CUg_ThreadControl.current != NULL);
  return CUg_ThreadControl.current->ID;
}


/****************************************************************************
* FUNCTION:
*
* CU_GetMasterThread
*
* DESCRIPTION:
*
* Get the ID of the master thread.
*
* RETURNS:
*
* CUt_ThreadID    -- The ID.
*
*/

CUt_ThreadID 
CU_GetMasterThread()
{
  ASSERT(CUg_ThreadControl.master != NULL);
  return CUg_ThreadControl.master->ID;
}


/****************************************************************************
* FUNCTION:
*
* CU_GetThreadData
*
* DESCRIPTION:
*
* Get the user data for the given thread.
*
* PARAMETERS:
*
* CUt_ThreadID     -- The thread ID.
*
* RETURNS:
*
* void *           -- The user data.
*
*/

void * 
CU_GetThreadData(CUt_ThreadID ID)
{
  CUt_ThreadEntry * thread;

  thread = CU_FindTLEntry(ID);
  ASSERT(thread != NULL);
  return thread->user_data;
}


/****************************************************************************
* FUNCTION:
*
* CU_SetThreadData
*
* DESCRIPTION:
*
* Set the user data for the given thread.
*
* PARAMETERS:
*
* CUt_ThreadID     -- The thread ID.
* void *           -- The user data.
*
* RETURNS:
*
* void *           -- The previous user data (user may need to deallocate
*                     this pointer if it was allocated for the thread
*                     to begin with).
*
*/

void * 
CU_SetThreadData(CUt_ThreadID ID, void *user_data)
{
  CUt_ThreadEntry * thread;
  void *old_ptr;

  thread = CU_FindTLEntry(ID);
  ASSERT(thread != NULL);
  old_ptr = thread->user_data;
  thread->user_data = user_data;
  
  return old_ptr;
}


/****************************************************************************
* FUNCTION:
*
* CU_GetThreadStackSpace
*
* DESCRIPTION:
*
* Get the space remaining on the given thread's stack.  Use this
* instead of StackSpace() trap since that won't work in thread
* context.
*
* PARAMETERS:
*
* CUt_ThreadID     -- The thread ID.
*
* RETURNS:
*
* size_t           -- The space left on stack, in bytes.
*
*/

size_t 
CU_GetThreadStackSpace(CUt_ThreadID ID)
{
#if defined(macintosh) && defined(SETJMPTHREADS)
  tUInt4 result;
  Ptr stack_bottom;
  jmp_buf registers;
  CUt_ThreadEntry * thread;
  
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(ID != 0);
  
  /* Find thread for ID */
  result = 0;
  thread = CU_FindTLEntry(ID);
  ASSERT(thread != NULL);
  
  /* Find stack's bottom -- master thread stack not allocated by us */
  if (thread == CUg_ThreadControl.master)
    stack_bottom = GetApplLimit();
  else
    stack_bottom = (Ptr) thread->env.stack;
  
  /* Compute stack space left -- if running, need to read registers */
  if (thread->status == CUk_Running) {
    (void) setjmp(registers);
    ASSERT((Ptr) registers[a7] >= stack_bottom);
    result = (Ptr) registers[a7] - stack_bottom;
  }
  else {
#ifdef __MWERKS__
    ASSERT((Ptr) thread->env.setjmp_env[a7] >= stack_bottom);
    result = (Ptr) thread->env.setjmp_env[a7] - stack_bottom;
#else
    ASSERT((Ptr) thread->env.setjmp_env[a7] >= stack_bottom);
    result = (Ptr) thread->env.setjmp_env[a7] - stack_bottom;
#endif
  }

  ASSERT(result >= 0);
  return(result);
#endif /* macintosh with SETJMPTHREADS */

#if defined(MACTHREADMANAGER)
  unsigned long result;
  CUt_ThreadEntry * thread;

  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(ID != 0);
  
  /* Find thread for ID */
  result = 0;
  thread = CU_FindTLEntry(ID);
  ASSERT(thread != NULL);
  
  /* Just ask the thread manager */
  if (ThreadCurrentStackSpace(thread->tm_ID, &result) != noErr)
    UNIMPLEMENTED;
    
  ASSERT(result >= 0);
  return(result);
#endif /* Mac with Thread Manager */

#if defined(SUNLWP)
  UNIMPLEMENTED;
  return 0;
#endif

#if defined(QUICKTHREADS)
  UNIMPLEMENTED;
  return 0;
#endif

#if defined(WIN32THREADS)
  UNTESTED;
#ifdef __BORLANDC__
  return stackavail();
#endif
  UNIMPLEMENTED;
  return 0;
#endif
}


/****************************************************************************
* FUNCTION:
*
* CU_CountActiveThreads
*
* DESCRIPTION:
*
* Count the number of threads currently active (that is, those that are
* running or ready to run).
*
* RETURNS:
*
* tUInt4           -- The number of threads.
*
*/

tUInt4 
CU_CountActiveThreads()
{
  tUInt4 count = 0;
  CUt_ThreadEntry *curr_entry;
  CUt_ThreadEntry *stop_entry;
  tUInt4 curr_ticks;
  
  ASSERT(CUg_ThreadControl.current != NULL);

  /* Check all non-current threads and count up number ready */
  curr_entry = CUg_ThreadControl.current->next_entry;
  stop_entry = CUg_ThreadControl.current;
  curr_ticks = CU_CurrentTicks();
  while (curr_entry != stop_entry) {
    if (CU_ThreadEntryReady(curr_entry, curr_ticks) != FALSE)
      count++;
    curr_entry = curr_entry->next_entry;
  }
  
  /* Add one if current thread is runnable (usually is) */
  if (CU_ThreadEntryReady(curr_entry, curr_ticks) != FALSE)
    count++;
  
  /* Return value */
  return count;
}


/* PROBABLY NOT NEEDED ! */
#if 0
/****************************************************************************
* FUNCTION:
*
* CU_GuessNextRunTime
*
* DESCRIPTION:
*
* Guess the time when some thread will next be ready to run.  This value
* is based on scheduled wake-up times and also guesses based on load in
* the recent past.
*
* PARAMETERS:
*
* tBoolean        -- TRUE when the current thread should be ignored in
*                    the estimate.
*
* RETURNS:
*
* tUInt4          -- Time of next run (as given by CU_CurrentTicks()).
*
*/

tUInt4 CU_GuessNextRunTime(tBoolean exclude_current)
{
  CUt_ThreadEntry *curr_entry;
  CUt_ThreadEntry *stop_entry;
  tUInt4 curr_ticks;
  tUInt4 result;
  
  ASSERT(CUg_ThreadControl.current != NULL);
  
  /* Check all non-current threads and count up number ready */
  curr_entry = CUg_ThreadControl.current->next_entry;
  stop_entry = CUg_ThreadControl.current;
  curr_ticks = CU_CurrentTicks();
  result = kMaxUInt4;
  
  /* Check current thread first (usually running, so best to do 1st) */
  if (!exclude_current) {
    if (CU_ThreadEntryReady(stop_entry, curr_ticks) != FALSE) 
      return curr_ticks;
    else if (stop_entry->block_opt == CUk_Delay) 
      result = CU_MIN(result, stop_entry->wake_time);
    else 
      result = CU_MIN(result, curr_ticks + (curr_ticks - stop_entry->last_run));
  }
   
  /* Check other threads */
  while (curr_entry != stop_entry) {
    if (CU_ThreadEntryReady(curr_entry, curr_ticks) != FALSE) 
      return curr_ticks;
    else if (curr_entry->block_opt == CUk_Delay) 
      result = CU_MIN(result, curr_entry->wake_time);
    else 
      result = CU_MIN(result, curr_ticks + (curr_ticks - curr_entry->last_run));
    curr_entry = curr_entry->next_entry;
  }
  
  /* Return value */
  return result;
}
#endif


/****************************************************************************
* FUNCTION:
*
* CU_BlockThread
*
* DESCRIPTION:
*
* Block the given thread according to various criteria.
*
* PARAMETERS:
*
* CUt_ThreadID         -- The thread ID.
* CUt_BlockOpt         -- Blocking options (describes remaining params)
* void *               -- Parameter one; use depends on blocking option.
* void *               -- Parameter two; use depends on blocking option.
* void *               -- Parameter three; use depends on blocking option.
* void *               -- Parameter four; reserved for future expansion.
* tBoolean             -- TRUE when thread is blocked waiting for an event
*                         in a local (same machine) peer process.  FALSE
*                         in all other cases.
*                         
*
* p1 through p4 are always of type pointer.  Thus, instead of passing a
* value, a pointer to the value is always passed.  E.g., if one parameter
* is a tUInt4 value, a (tUInt4 *) is passed.  Although this is a bit
* annoying at the point of call, it is the only way to ensure that this
* code is portable.
*
* Acceptible block options include:
*
* 0) CUk_UnBlock:
*
*    Unblock thread.  Execution will be continued the next time the thread
*    is given a time slice.  p1, p2, p3, and p4 should be NULL.
*
* 1) CUk_Delay: 
*
*    Block thread until given number of ticks has passed.  Ptr to number 
*    of ticks (tUInt4) is passed in p1.  This value may be kMaxUInt4 to
*    delay the thread indefinately but should not be other large values
*    that will cause value overflow when added to current ticks.  p2, 
*    p3, and p4 should be NULL.
*
* 2) CUk_WhileFALSE: 
*
*    Block while given boolean * is set to FALSE.  Ptr to the boolean
*    (tBoolean *) is passed in p1.  Note that this pointer must continue
*    to point to valid memory until the thread is unblocked.  p2, p3, and 
*    p4 should be NULL.
*
* 3) CUk_UntilFALSE: 
*
*    Block until given boolean * changes to FALSE.  Ptr to the boolean
*    (tBoolean *) is passed in p1.  Note that this pointer must continue
*    to point to valid memory until the thread is unblocked.  p2, p3, and 
*    p4 should be NULL.
*
* 4) CUk_WhileEquals: 
*
*    Block while a given memory range equals a given value.  Ptr to
*    the memory range (any type *) is passed in p1.  Ptr to comparison
*    range (the target value) is passed in p2.  Note that p1 and p2 must 
*    continue to point to valid memory until the thread is unblocked.  
*    Ptr to length of the memory ranges in bytes (tUInt4 *) is passed 
*    in p3.  p4 should be NULL.
*
* 5) CUk_UntilEquals: 
*
*    Block until a given memory range changes to a given value.  Ptr to
*    the memory range (any type *) is passed in p1.  Ptr to comparison
*    range (the target value) is passed in p2.  Note that p1 and p2 must 
*    continue to point to valid memory until the thread is unblocked.  
*    Ptr to length of the memory ranges in bytes (tUInt4 *) is passed 
*    in p3.  p4 should be NULL.
*
* 6) CUk_WhileFctFALSE: 
*
*    Block while a given function returns FALSE.  Ptr to the function
*    (CUt_ThreadBlockProc) is passed in p1.  Extra parameters for
*    the block proc (passed to the proc's second parameter) is passed
*    in p2 (any type *).  Note that p1 and p2 must continue to point
*    to valid proc and memory, respectively, until the thread is
*    unblocked.  Ptr to maximum frequency in ticks of calls to the block 
*    procedure (tUInt4 *) is passed in p3.  This value should not contain
*    large values that will overflow when added to current ticks (there
*    is no checking for this).  p4 should be NULL.
*
* 7) CUk_UntilFctFALSE: 
*
*    Block until a given function returns FALSE.  Ptr to the function
*    (CUt_ThreadBlockProc) is passed in p1.  Extra parameters for
*    the block proc (passed to the proc's second parameter) is passed
*    in p2 (any type *).  Note that p1 and p2 must continue to point
*    to valid proc and memory, respectively, until the thread is
*    unblocked.  Ptr to maximum frequency in ticks of calls to the block 
*    procedure (tUInt4 *) is passed in p3.  This value should not contain
*    large values that will overflow when added to current ticks (there
*    is no checking for this).  p4 should be NULL.
*
* NOTES:
*
* The blocking function is passed the thread user data as its first
* parameter and the params given here as the second parameter.
*
* Ticks are approximately 1/60 second.
*
* If the given thread is the current thread, it will not be stopped
* until CU_ThreadYield() is called.
*
* If this function is called repeatedly, newer calls overwrite 
* any existing block criteria.
*
*/

void 
CU_BlockThread(CUt_ThreadID ID, CUt_BlockOpt opt, void *p1, void *p2, 
                    void *p3, void *p4, tBoolean local_peer)
{
  CUt_ThreadEntry * thread;
  
  ASSERT(ID != 0);
  
  thread = CU_FindTLEntry(ID);
  ASSERT(thread != NULL);
  CU_BlockThreadEntry(thread, opt, p1, p2, p3, p4, local_peer);
}


/****************************************************************************
* FUNCTION:
*
* CU_ThreadYield
*
* DESCRIPTION:
*
* Yield processing time to other threads.  Threads are expected to call
* this function relatively often.
*
*/

void 
CU_ThreadYield()
{
  CUt_ThreadEntry *next = NULL;
  
  ASSERT(CUg_ThreadControl.master != NULL);
  
  /* Get next thread and run it.  This is a loop because when all */
  /* threads are blocked, we need to wait until something unblocks. */
  /* The logic of the thread that called this function expects that */
  /* it not return until it is un-blocked!  So here we sit... */
  /* Note that on non-preemptive systems, the process yielder thread */
  /* is woken up right away when there is nothing else to be done! */
  while (next == NULL) {
    next = CU_NextScheduledThreadEntry();
    if (next != NULL) {
      if (!CU_StartThreadEntry(next))
        UNIMPLEMENTED;
    }
    else 
      CU_WakeUpProcessYielder();
  }
}


/****************************************************************************
* FUNCTION:
*
* CU_StartThread
*
* DESCRIPTION:
*
* Switch context and start the given thread processing.  Does nothing
* if it is already running.
*
* PARAMETERS:
*
* CUt_ThreadID         -- The thread ID.
*
* RETURNS:
* 
* tBoolean             -- TRUE when thread was run (by the time this returns
*                         it has already run) and FALSE when it is not
*                         ready, is sleeping, or blocked.  This function
*                         also returns FALSE when passed the current thread.
*
* NOTES:
*
* This is provided to allow faster 'conversations' between two threads
* by allowing them to start each other up but developers should take
* care not to block out other threads for long periods of time when this
* function is used instead of the usual call to CU_ThreadYield().
*
*/

tBoolean 
CU_StartThread(CUt_ThreadID ID)
{  
  CUt_ThreadEntry * thread;
  
  ASSERT(ID != 0);
  
  thread = CU_FindTLEntry(ID);
  ASSERT(thread != NULL);
  return CU_StartThreadEntry(thread);
}


/****************************************************************************
* FUNCTION:
*
* CU_ApplStatusChanged()
*
* DESCRIPTION:
*
* Set current application status.  The global CUg_ApplStatus can be used
* to access current status, which is useful in some cases for scheduling
* and/or blocking threads.
*
* PARAMETERS:
*
* CUt_ApplStatus        -- Status (one of CUk_InBackground, CUk_InForeground,
*                          or CUk_Quitting).
*
* NOTES:
*
* Note that you *must* call this function under all systems because
* application status is used to determine duration and frequency
* of yields to other processes (where applicable), among other things.
*
*/

void 
CU_ApplStatusChanged(CUt_ApplStatus status)
{
  if (CUg_ApplStatus != status) {
    CU_WakeUpProcessYielder();
    CUg_ApplStatus = status;
  }
}


/****************************************************************************
* FUNCTION:
*
* CU_ProcessYielder()
*
* DESCRIPTION:
*
* Thread procedure for yielding time to other processes in a non-preemptive
* multitasking environment.  This procedure uses information about currently
* running threads, the nature of existing thread blocks, and application
* status (foreground/interactive or background) to determine the best 
* frequency and duration of yields to other processes in the system.
*
* PARAMETERS:
*
* void *       -- Required data pointer.   Unused.
*
* NOTES:
*
* To use, call CU_StartThread(CU_ProcessYielder, NULL, CUk_UseDefaultStackSize) 
* and make sure you maintain this package's datastructures with all required 
* calls to:
*
* 1) CU_ApplStatusChanged() when the application switches into and out 
*    of the foreground.
* 2) Set flag local_peer accordingly when calling CU_BlockThread().
*
* This process will cause immediate termination of the thread under 
* preemptive multitasking systems (where it is not needed); thus can be
* used safely on all systems.
*
* THIS IS STILL EXPERIMENTAL!  But seems to work OK.  The main improvement
* that could be made is to switch to calling WaitNextEvent() on Mac instead
* of EventAvail() when running in the foreground and when there has been
* no recent thread activity.  The current implementation causes apps
* to hog alot of time even when completely idle in the foreground.
*
*/

void 
CU_ProcessYielder(void *data)
{
#ifdef PREEMPTIVEMULTI
  UNUSED(data);
  return;
#else
  register tUInt4 curr_ticks;
  tBoolean running_threads;
  tBoolean blocked_on_local_peer;
  tUInt4 old_proc_yield_freq;
  tUInt4 temp_long;
#ifdef macintosh
  EventRecord mac_event;
  tUInt4 temp_ticks;
#endif

  UNUSED(data);
  
  /* Store this thread's table entry in global for external wakeup */
  CUg_ProcYielderThread = CUg_ThreadControl.current;
  ASSERT(CUg_ProcYielderThread != NULL);
  
  /* Run this thread! */
  while (CUg_ApplStatus != CUk_Quitting) {
    /* Compute current ticks only once */
    curr_ticks = CU_CurrentTicks();

    /* Determine time and duration of next yield based on current status */
    /* if status changed */
    if ((CUg_ProcYieldStatusChanged) || (curr_ticks >= CUg_ProcYieldRedoTime)) { 
      /* Reset status flag */
      CUg_ProcYieldStatusChanged = FALSE;
      CUg_ProcYieldRedoTime = curr_ticks + CUk_ProcYieldRedoDelay;
      
      /* Precomputed values */
      old_proc_yield_freq = CUg_ProcYieldFreq;
      running_threads = (CU_CountActiveThreads() > 1);
      blocked_on_local_peer = (CUg_ThreadControl.num_local_peers > 0);
      
      /* Set frequency and duration */
      if (running_threads) {
        if (CUg_ApplStatus == CUk_InForeground) {
          if (blocked_on_local_peer) {
            /* FG, Running, blocked on local peer */
            /* Give up a bit more since we have a local peer! */
            CUg_ProcYieldFreq = 1;
            CUg_ProcYieldDuration = 1;
          }
          else {
            /* FG, Running, No local peer */
            CUg_ProcYieldFreq = 1;
            CUg_ProcYieldDuration = 0;
          }
        }
        else {
          if (blocked_on_local_peer) {
            /* BG, Running, blocked on local peer */
            /* Give up a bit more since we have a local peer! */
            CUg_ProcYieldFreq = 0;
            CUg_ProcYieldDuration = 1;
          }
          else {
            /* BG, Running, No local peer */
            /* In this case we have to be more hoggy since Mac is already */
            /* being mighty stingy with CPU time! */
            CUg_ProcYieldFreq = 10;
            CUg_ProcYieldDuration = 0;
          }
        }
      } 
      else {
        /* No threads running */
        /* Will of course stop earlier usually! */
        CUg_ProcYieldFreq = 0;
        CUg_ProcYieldDuration = 10000;
      }
      
      /* printf("*** Freq = %i, Dur = %i\n", CUg_ProcYieldFreq, CUg_ProcYieldDuration); */
      
      /* Adjust next yield time if frequency of yields just changed */
      /* Don't recompute from scratch, though, because this should be */
      /* relative to the last time we actually did it! */
      if (old_proc_yield_freq != CUg_ProcYieldFreq) {
        CUg_NextProcYield = CUg_NextProcYield + 
                            (CUg_ProcYieldFreq - old_proc_yield_freq);
      }
    }

    /* Yield time to other processes if necessary */
    if (curr_ticks >= CUg_NextProcYield) {
      CUg_NextProcYield = curr_ticks + CUg_ProcYieldFreq;
#ifdef macintosh
      /* NOTE: Cannot use WaitNextEvent() here because it swallows */
      /* events no matter how you use it (not sure why or how) but */
      /* can use EventAvail(), which also yields time to the system */
      /* in small amounts.  We call it repeatedly for desired effect */
      /* until time is up, or forever until a thread is ready to run. */
      temp_ticks = CU_CurrentTicks();
      do {
        EventAvail(everyEvent, &mac_event);
      } while ((CU_CurrentTicks() - temp_ticks < CUg_ProcYieldDuration) &&
               (CU_CountActiveThreads() <= 1));
      /* printf("--- Yielded %i ticks\n", CU_CurrentTicks() - temp_ticks); */
#endif 
#ifdef MS_WIN
      UNFINISHED;
#endif /* MS_WIN */
    }
    
    /* Block until next yield or state change (wake-up is explicit in */
    /* latter case).  Note that we purposely ignore time elapsed during */
    /* above process yield in order to avoid starving this app when */
    /* some other process is being a hog! */
    temp_long = CU_MAX(0, CUg_NextProcYield - curr_ticks);
    CU_BlockThreadEntry(CUg_ProcYielderThread, CUk_Delay, &temp_long, 
                   NULL, NULL, NULL, FALSE);

    /* Give up time to other threads! */
    CU_ThreadYield();
  }
#endif /* !PREEMPTIVEMULTI */
}



/**************************************************************************/
/**************************************************************************/
/* PRIVATE GLOBALS AND FUNCTIONS */

/**************************************************************************/
/* GLOBALS */

/*
* The thread control structure - contains information about the state of
* the thread package, including the list of all threads.
*/

static CUt_ThreadControl CUg_ThreadControl = { 0, 0, NULL, NULL, NULL };

/* 
* Application status -- either in foreground (interacting with user),
* in background, or quitting
*/

CUt_ApplStatus CUg_ApplStatus = CUk_InForeground;

/*
* Process dependency globals for smart time slicing  to other processes in 
* cooperative multitasking environments.
*/

#ifndef PREEMPTIVEMULTI
tUInt4 CUg_NextProcYield = 0;
tUInt4 CUg_ProcYieldFreq = 1;
tUInt4 CUg_ProcYieldDuration = 1;
tBoolean CUg_ProcYieldStatusChanged = TRUE;
tUInt4 CUg_ProcYieldRedoTime = 0;
CUt_ThreadEntry *CUg_ProcYielderThread = NULL;
#endif


/****************************************************************************
* FUNCTION:
*
* CU_InsertTLEntry
*
* DESCRIPTION:
*
* Insert a thread entry into the thread list.
*
* PARAMETERS:
*
* CUt_ThreadEntry *     -- The thread entry to add.
*
* NOTES:
*
* This is written to assume that the first insert will be the master
* thread, which is OK because this whole package is written with that
* assumption.
*
*/

void CU_InsertTLEntry(CUt_ThreadEntry *thread)
{
  ASSERT(thread != NULL);
  
  /* Insert first entry and set 'current' so subsequent inserts work */
  if (CUg_ThreadControl.current == NULL) {
    CUg_ThreadControl.current = thread;
    CUg_ThreadControl.master = thread;
    thread->prev_entry = thread;
    thread->next_entry = thread;
  }
  
  /* Enter additional entries just before current thread */
  else {
    ASSERT(CUg_ThreadControl.current != NULL);
    
    thread->prev_entry = CUg_ThreadControl.current->prev_entry;
    thread->next_entry = CUg_ThreadControl.current;
    if (CUg_ThreadControl.current->next_entry == CUg_ThreadControl.current) 
      CUg_ThreadControl.current->next_entry = thread;
    else 
      CUg_ThreadControl.current->prev_entry->next_entry = thread;
    CUg_ThreadControl.current->prev_entry = thread;
  }
}


/****************************************************************************
* FUNCTION:
*
* CU_RemoveTLEntry
*
* DESCRIPTION:
*
* Remove a thread entry from the thread list.
*
* PARAMETERS:
*
* CUt_ThreadEntry *     -- The thread entry to remove.
*
*/

void CU_RemoveTLEntry(CUt_ThreadEntry *thread)
{
  ASSERT(thread != NULL);
  
  /* Remove last entry */
  if (thread->next_entry == thread) {
    CUg_ThreadControl.current = NULL;
  }
  
  /* Remove non-last entry */
  else {
    ASSERT(thread->next_entry != NULL);
    ASSERT(thread->prev_entry != NULL);
    thread->next_entry->prev_entry = thread->prev_entry;
    thread->prev_entry->next_entry = thread->next_entry;
  }
  
  /* Reset thread entry pointers */
  thread->next_entry = NULL;
  thread->prev_entry = NULL;
}


/****************************************************************************
* FUNCTION:
*
* CU_FindTLEntry
*
* DESCRIPTION:
*
* Find the thread entry for the given thread ID.
*
* PARAMETERS:
*
* CUt_ThreadID       -- The ID
*
* RETURNS:
*
* CUt_ThreadEntry *  -- The thread list entry, or NULL if not found.
*
*/

CUt_ThreadEntry *CU_FindTLEntry(register CUt_ThreadID ID)
{
  register CUt_ThreadEntry *thread;

  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(ID >= 0);
  ASSERT(ID <= CUg_ThreadControl.ID_seed);
  
  /* Find the thing */
  thread = CUg_ThreadControl.current->next_entry;
  while ((thread != CUg_ThreadControl.current) && (thread->ID != ID))
    thread = thread->next_entry;
  if (thread->ID != ID)
    thread = NULL;

   /* Return it */
  return(thread);
}


/****************************************************************************
* FUNCTION:
*
* CU_DestroyThreadEntry
*
* DESCRIPTION:
*
* Kill the given thread, remove it from the thread list, and deallocate
* it.
*
* PARAMETERS:
*
* CUt_ThreadEntry *     -- The thread entry.
*
* NOTES:
*
* The master thread is assumed to be killed only after all other threads
* are gone.  Use CU_StopTasking() to kill all threads and halt the tasking
* process.
*
*/

/*
* Local dummy function required by QuickThreads ... in our case, it does
* nothing because we already have a clean-up mechanism
*/
#ifdef QUICKTHREADS
static void *
CU_QTAbortDummy(qt_t *sp, void *old, void *unused)
{
  return NULL;
}
#endif

void CU_DestroyThreadEntry(CUt_ThreadEntry *thread)
{
  CUt_ThreadEntry *next = NULL;
  tBoolean temp_bool = FALSE;
  tBoolean is_running = FALSE;
#ifdef WIN32THREADS
  DWORD ret;
#endif

  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(thread != NULL);
  
  /* 
  * Find next entry
  */
  
  if (thread == CUg_ThreadControl.master) {
    /* SHOULD WE KILL OTHER THREADS HERE ?? */
    /* For now, enforce disposal of master last */
    ASSERT(thread->next_entry == thread);
    next = NULL;
  }
  else if (thread->status == CUk_Running) {
    is_running = TRUE;
    temp_bool = FALSE;
    CU_BlockThreadEntry(thread, CUk_WhileFALSE, &temp_bool, 
                        NULL, NULL, NULL, FALSE);
    next = NULL;
    while (next == NULL)
      next = CU_NextScheduledThreadEntry();
  } 
  
  /* 
  * Remove thread from list 
  */
  
  CU_RemoveTLEntry(thread);
  
  /*
  * If disposing current thread, need to start the next one, if any
  * and wait until later to deallocate the stack (which is in use).
  * (NOTE: if disposing running thread without a next one, then don't 
  * need to do this.)
  */
  
  if ((is_running == TRUE) && (thread != CUg_ThreadControl.master)) {
  
    /* Hang onto thread for later destruction */
    CUg_ThreadControl.destroy = thread;
    thread->status = CUk_Destroyed;

    /* Start next thread */
    CUg_ThreadControl.current = next;
    next->status = CUk_Running;
#ifdef SETJMPTHREADS
#ifdef MS_WIN
    UNFINISHED;
#endif
    longjmp(next->env.setjmp_env, CUk_RunThread);
#endif
#if defined(SUNLWP)
    if (lwp_yield(next->lwp_ID) != 0)
      UNIMPLEMENTED;
#endif
#if defined(QUICKTHREADS)
    QT_ABORT(CU_QTAbortDummy, thread->qthread, NULL, next->qthread);
#endif
#if defined(MACTHREADMANAGER)
    if (YieldToThread(next->tm_ID) != noErr)
      UNIMPLEMENTED;
#endif
#if defined(WIN32)
    CU_StartWin32Thread(next);
#endif    
    /* Above jump/yields should never return because this thread */
    /* will be destroyed in subsequent CU_RestoreEnvironment(). */
    ASSERT(0);
  }
  
  /*
  * If thread is not current or is the master thread, just 
  * destroy it now.
  */
  
  else {
#ifdef SUNLWP
    if (thread != CUg_ThreadControl.master) {
      if (lwp_destroy(thread->lwp_ID) != 0)
        UNIMPLEMENTED;
    }
#endif
#ifdef QUICKTHREADS
    /* Set master thread stack info to NULL; can't dealloc it! */
    if (thread == CUg_ThreadControl.master) {
      thread->qthread = NULL;
      thread->env.stack = NULL;
    }
    /* In any case, don't want to dealloc thread->qthread; it's just a */
    /* pointer into thread->env.stack! */
#endif
#ifdef MACTHREADMANAGER
    if (thread != CUg_ThreadControl.master) {
      if (DisposeThread(thread->tm_ID, NULL, FALSE) != noErr)
        UNIMPLEMENTED;
    }
#endif
#ifdef WIN32THREADS
    if (thread != CUg_ThreadControl.master) {
      CU_DestroyWin32Thread(thread);
    }

    ret = CloseHandle(thread->win32_event);
    ASSERT(ret == TRUE);
    thread->win32_event = NULL;
#endif

    if (thread->env.stack != NULL) {
      CU_Free(thread->env.stack);
    }
    CU_Free(thread);
  }
}


/****************************************************************************
* FUNCTION:
*
* CU_MinStackSize
*
* DESCRIPTION:
*
* Get the minimum suggested stack size.
*
* RETURNS:
*
* size_t    -- The size, in bytes.
*
*/

size_t CU_MinStackSize()
{
#ifdef QUICKTHREADS
  return CUk_QThreadMinStack;
#else
#if defined(macintosh) && defined(SETJMPTHREADS)
  return(LMGetMinStack()); 
#elif defined(MACTHREADMANAGER)
  return(LMGetMinStack());
#elif defined( WIN32 )
  return( CUk_Win32ThreadDefaultStack );
#elif defined( sunos4 )
  return(MINSTACKSZ * sizeof(stkalign_t));
#elif defined( hpux9 )
  return CUk_QThreadMinStack;
#endif
#endif /* QUICKTHREADS */ 
}

/****************************************************************************
* FUNCTION:
*
* CU_DefaultStackSize
*
* DESCRIPTION:
*
* Get the suggested default stack size.
*
* RETURNS:
*
* size_t    -- The size, in bytes.
*
*/

size_t CU_DefaultStackSize()
{
#ifdef QUICKTHREADS
  return CUk_QThreadDefaultStack;
#else
#if defined(macintosh) && defined(SETJMPTHREADS)
  return(LMGetDefltStack()); 
#elif defined(MACTHREADMANAGER)
  size_t result;
  if (GetDefaultThreadStackSize(kCooperativeThread, (Size *) &result) != noErr) {
    BARF;
    result = 0;
  }
  return(result);
#elif defined( WIN32 )
  return( CUk_Win32ThreadDefaultStack );
#elif defined( sunos4 )
  return(MINSTACKSZ * 2 * sizeof(stkalign_t));
#elif defined( hpux9 )
  return CUk_QThreadDefaultStack;
#endif
#endif /* QUICKTHREADS */
}


#ifndef PREEMPTIVEMULTI
/****************************************************************************
* FUNCTION:
*
* CU_WakeUpProcessYielder
*
* DESCRIPTION:
*
* Wake up the process yielder thread, if not already ready to run.
*
* NOTES:
*
* Does not use CU_BlockThreadEntry() to avoid infinite looping!
*
*/

void CU_WakeUpProcessYielder()
{
  if (CUg_ProcYielderThread != NULL) {
    CUg_ProcYieldStatusChanged = TRUE;
    CUg_ProcYielderThread->status = CUk_Ready;
  }
}
#endif /* !PREEMPTIVEMULTI */


/****************************************************************************
* FUNCTION:
*
* CU_SaveEnvironment
*
* DESCRIPTION:
*
* Save the environment of the current thread (other than register and stack
* environment).
*
*/

void CU_SaveEnvironment()
{
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(CUg_ThreadControl.current != NULL);

#if defined(macintosh) && defined(SETJMPTHREADS)
  CUg_ThreadControl.current->env.heap_end = LMGetHeapEnd();
  CUg_ThreadControl.current->env.appl_limit = LMGetApplLimit();
  CUg_ThreadControl.current->env.hi_heap_mark = LMGetHiHeapMark();
#endif
}

/****************************************************************************
* FUNCTION:
*
* CU_RestoreEnvironment
*
* DESCRIPTION:
*
* Restore the environment for the current thread (other than register and
* stack environment).
*
*/

void CU_RestoreEnvironment()
{
  register CUt_ThreadEntry *thread;
  
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(CUg_ThreadControl.current != NULL);
  
  /* Optimization by using registers (many compilers may do this anyway) */
  thread = CUg_ThreadControl.current;

  /* Reset Macintosh LM globals needed for thread */
#if defined(macintosh) && defined(SETJMPTHREADS)
  LMSetHeapEnd((Ptr) thread->env.heap_end);
  LMSetApplLimit((Ptr) thread->env.appl_limit);
  LMSetHiHeapMark((Ptr) thread->env.hi_heap_mark);
#endif

  /* Destroy previous thread if it was killed while running */
  if (CUg_ThreadControl.destroy != NULL) {
#ifdef SUNLWP
    if (lwp_destroy(CUg_ThreadControl.destroy->lwp_ID) != 0)
      UNIMPLEMENTED;
#endif
#ifdef QUICKTHREADS
    /* Don't have to do anything since destroy->qthread just points */
    /* into destroy->stack, which is disposed below */
    ASSERT(CUg_ThreadControl.destroy != CUg_ThreadControl.master);
#endif
#ifdef MACTHREADMANAGER
    if (DisposeThread(CUg_ThreadControl.destroy->tm_ID, NULL, FALSE) != noErr)
      UNIMPLEMENTED;
#endif

    if (CUg_ThreadControl.destroy->env.stack != NULL) 
      CU_Free(CUg_ThreadControl.destroy->env.stack);
    CU_Free(CUg_ThreadControl.destroy);
    CUg_ThreadControl.destroy = NULL;
  }

/* PROBABLY NOT NEEDED! */
#if 0
  /* Set last run time statistic */
  thread->last_run = CU_CurrentTicks();
#endif 
}


/****************************************************************************
* FUNCTION:
*
* CU_BlockThreadEntry
*
* DESCRIPTION:
*
* Block the given thread according to various criteria.
*
* PARAMETERS:
*
* CUt_ThreadEntry *    -- The thread entry.
* CUt_BlockOpt         -- Blocking options (describes remaining params)
* void *               -- Parameter one; use depends on blocking option.
* void *               -- Parameter two; use depends on blocking option.
* void *               -- Parameter three; use depends on blocking option.
* void *               -- Parameter four; reserved for future expansion.
* tBoolean             -- TRUE when thread is blocked waiting for an event
*                         in a local (same machine) peer process.  FALSE
*                         in all other cases.
*
* p1 through p4 are always of type pointer.  Thus, instead of passing a
* value, a pointer to the value is always passed.  E.g., if one parameter
* is a tUInt4 value, a (tUInt4 *) is passed.  Although this is a bit
* annoying at the point of call, it is the only way to ensure that this
* code is portable.
*
* Acceptible block options include:
*
* 0) CUk_UnBlock:
*
*    Unblock thread.  Execution will be continued the next time the thread
*    is given a time slice.  p1, p2, p3, and p4 should be NULL.
*
* 1) CUk_Delay: 
*
*    Block thread until given number of ticks has passed.  Ptr to number 
*    of ticks (tUInt4) is passed in p1.  This value may be kMaxUInt4 to
*    delay the thread indefinately but should not be other large values
*    that will cause value overflow when added to current ticks.  p2, 
*    p3, and p4 should be NULL.
*
* 2) CUk_WhileFALSE: 
*
*    Block while given boolean * is set to FALSE.  Ptr to the boolean
*    (tBoolean *) is passed in p1.  Note that this pointer must continue
*    to point to valid memory until the thread is unblocked.  p2, p3, and 
*    p4 should be NULL.
*
* 3) CUk_UntilFALSE: 
*
*    Block until given boolean * changes to FALSE.  Ptr to the boolean
*    (tBoolean *) is passed in p1.  Note that this pointer must continue
*    to point to valid memory until the thread is unblocked.  p2, p3, and 
*    p4 should be NULL.
*
* 4) CUk_WhileEqual: 
*
*    Block while a given memory range equals a given value.  Ptr to
*    the memory range (any type *) is passed in p1.  Ptr to comparison
*    range (the target value) is passed in p2.  Note that p1 and p2 must 
*    continue to point to valid memory until the thread is unblocked.  
*    Ptr to length of the memory ranges in bytes (tUInt4 *) is passed 
*    in p3.  p4 should be NULL.
*
* 5) CUk_UntilEqual: 
*
*    Block until a given memory range changes to a given value.  Ptr to
*    the memory range (any type *) is passed in p1.  Ptr to comparison
*    range (the target value) is passed in p2.  Note that p1 and p2 must 
*    continue to point to valid memory until the thread is unblocked.  
*    Ptr to length of the memory ranges in bytes (tUInt4 *) is passed 
*    in p3.  p4 should be NULL.
*
* 6) CUk_WhileFctFALSE: 
*
*    Block while a given function returns FALSE.  Ptr to the function
*    (CUt_ThreadBlockProc) is passed in p1.  Extra parameters for
*    the block proc (passed to the proc's second parameter) is passed
*    in p2 (any type *).  Note that p1 and p2 must continue to point
*    to valid proc and memory, respectively, until the thread is
*    unblocked.  Ptr to maximum frequency in ticks of calls to the block 
*    procedure (tUInt4 *) is passed in p3.  This value should not contain
*    large values that will overflow when added to current ticks (there
*    is no checking for this).  p4 should be NULL.
*
* 7) CUk_UntilFctFALSE: 
*
*    Block until a given function returns FALSE.  Ptr to the function
*    (CUt_ThreadBlockProc) is passed in p1.  Extra parameters for
*    the block proc (passed to the proc's second parameter) is passed
*    in p2 (any type *).  Note that p1 and p2 must continue to point
*    to valid proc and memory, respectively, until the thread is
*    unblocked.  Ptr to maximum frequency in ticks of calls to the block 
*    procedure (tUInt4 *) is passed in p3.  This value should not contain
*    large values that will overflow when added to current ticks (there
*    is no checking for this).  p4 should be NULL.
*
* NOTES:
*
* The blocking function is passed the thread user data as its first
* parameter and the params given here as the second parameter.
*
* Ticks are approximately 1/60 second.
*
* If the given thread is the current thread, it will not be stopped
* until CU_ThreadYield() is called.
*
* If this function is called repeatedly, newer calls overwrite 
* any existing block criteria.
*
*/

void CU_BlockThreadEntry(CUt_ThreadEntry *thread, CUt_BlockOpt opt, void *p1, 
                         void *p2, void *p3, void *p4, tBoolean local_peer)
{
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(thread != NULL);

  thread->status = CUk_Blocked;
  thread->block_opt = opt;

  switch (opt) {
  
    /*
    * Unblock thread
    */
    
    case CUk_UnBlock:
    
      ASSERT(p1 == NULL);
      ASSERT(p2 == NULL);
      ASSERT(p3 == NULL);
      ASSERT(p4 == NULL);
      
      thread->status = CUk_Ready;
      
      if (thread->local_peer) {
         thread->local_peer = FALSE;
         CUg_ThreadControl.num_local_peers--;
         ASSERT(CUg_ThreadControl.num_local_peers >= 0);
      }
      CU_WakeUpProcessYielder();

      break;
      
    /* 
    * Delay style block 
    */
    
    case CUk_Delay:
    
      ASSERT(p1 != NULL);
      ASSERT(p2 == NULL);
      ASSERT(p3 == NULL);
      ASSERT(p4 == NULL);
  
      if (*((tUInt4 *) p1) == kMaxUInt4)
        thread->wake_time = kMaxUInt4;
      else 
        thread->wake_time = CU_CurrentTicks() + *((tUInt4 *) p1);
        
      if (local_peer) {
        thread->local_peer = TRUE;
        CUg_ThreadControl.num_local_peers++;
      }
      CU_WakeUpProcessYielder();

      break;
    
    /* 
    * While/Until FALSE style block 
    */
    
    case CUk_WhileFALSE:
    case CUk_UntilFALSE:
    
      ASSERT(p1 != NULL);
      ASSERT(p2 == NULL);
      ASSERT(p3 == NULL);
      ASSERT(p4 == NULL);
  
      thread->blocked_bool = (tBoolean *) p1;
        
      if (local_peer) {
        thread->local_peer = TRUE;
        CUg_ThreadControl.num_local_peers++;
      }
      CU_WakeUpProcessYielder();
      
      break;
    
    /* 
    * While/Until memory equals style block 
    */
    
    case CUk_WhileEqual:
    case CUk_UntilEqual:
    
      ASSERT(p1 != NULL);
      ASSERT(p2 != NULL);
      ASSERT(p3 != NULL);
      ASSERT(p4 == NULL);

      thread->blocked_mem = p1;
      thread->target_mem = p2;
      thread->mem_len = *((tUInt4 *) p3);
        
      if (local_peer) {
        thread->local_peer = TRUE;
        CUg_ThreadControl.num_local_peers++;
      }
      CU_WakeUpProcessYielder();
      
      break;

    /* 
    * While/Until function FALSE style block 
    */
    
    case CUk_WhileFctFALSE:
    case CUk_UntilFctFALSE:
    
      ASSERT(p1 != NULL);
      ASSERT(p3 != NULL);
      ASSERT(p4 == NULL);

      thread->blocked_fct = (CUt_ThreadBlockProc) p1;
      thread->blocked_params = p2;
      thread->wake_time = CU_CurrentTicks() + *((tUInt4 *) p3);
      thread->check_freq = *((tUInt4 *) p3);
        
      if (local_peer) {
        thread->local_peer = TRUE;
        CUg_ThreadControl.num_local_peers++;
      }
      CU_WakeUpProcessYielder();
      
      break;
    
    default:
      BARF;
  }
}


/****************************************************************************
* FUNCTION:
*
* CU_ThreadEntryReady
*
* DESCRIPTION:
*
* Check if the given thread is ready to run at the given time.
*
* PARAMETERS:
*
* CUt_ThreadEntry *    -- The thread entry.
* tUInt4               -- The current time in ticks (mainly here for optimization
*                         purposes when this function is called from a loop).
*
* RETURNS:
*
* tBoolean             -- TRUE if ready to run (or is already running)
*
*/

tBoolean CU_ThreadEntryReady(CUt_ThreadEntry *thread, tUInt4 ticks)
{
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(thread != NULL);

  switch (thread->status) {
    case CUk_Ready:
    case CUk_Running:
      return TRUE;

    case CUk_Blocked:
      switch (thread->block_opt) {
        case CUk_Delay:
    
          if (ticks >= thread->wake_time) {
            thread->status = CUk_Ready;
            if (thread->local_peer) {
              thread->local_peer = FALSE;
              CUg_ThreadControl.num_local_peers--;
              ASSERT(CUg_ThreadControl.num_local_peers >= 0);
            }
            CU_WakeUpProcessYielder();
            return TRUE;
          }
          else
            return FALSE;
            
          break;
    
        case CUk_WhileFALSE:
    
          if (*((tBoolean *) thread->blocked_bool) != FALSE) {
            thread->status = CUk_Ready;
            if (thread->local_peer) {
              thread->local_peer = FALSE;
              CUg_ThreadControl.num_local_peers--;
              ASSERT(CUg_ThreadControl.num_local_peers >= 0);
            }
            CU_WakeUpProcessYielder();
            return TRUE;
          }
          else
            return FALSE;
        
          break;
    
        case CUk_UntilFALSE:
    
          if (*((tBoolean *) thread->blocked_bool) == FALSE) {
            thread->status = CUk_Ready;
            if (thread->local_peer) {
              thread->local_peer = FALSE;
              CUg_ThreadControl.num_local_peers--;
              ASSERT(CUg_ThreadControl.num_local_peers >= 0);
            }
            CU_WakeUpProcessYielder();
            return TRUE;
          }
          else
            return FALSE;
        
          break;
    
        case CUk_WhileEqual:
    
          if (CU_MemCmp(thread->blocked_mem, thread->target_mem, 
                     thread->mem_len) != 0) {
            thread->status = CUk_Ready;
            if (thread->local_peer) {
              thread->local_peer = FALSE;
              CUg_ThreadControl.num_local_peers--;
              ASSERT(CUg_ThreadControl.num_local_peers >= 0);
            }
            CU_WakeUpProcessYielder();
            return TRUE;
          }
          else
            return FALSE;
          
          break;
    
        case CUk_UntilEqual:
    
          if (CU_MemCmp(thread->blocked_mem, thread->target_mem, 
                     thread->mem_len) == 0) {
            thread->status = CUk_Ready;
            if (thread->local_peer) {
              thread->local_peer = FALSE;
              CUg_ThreadControl.num_local_peers--;
              ASSERT(CUg_ThreadControl.num_local_peers >= 0);
            }
            CU_WakeUpProcessYielder();
            return TRUE;
          }
          else
            return FALSE;
          
          break;
    
        case CUk_WhileFctFALSE:
        
          if (ticks >= thread->wake_time)
            if (thread->blocked_fct(thread->user_data, 
                                    thread->blocked_params) != FALSE) {
              thread->status = CUk_Ready;
              if (thread->local_peer) {
                thread->local_peer = FALSE;
                CUg_ThreadControl.num_local_peers--;
                ASSERT(CUg_ThreadControl.num_local_peers >= 0);
              }
              CU_WakeUpProcessYielder();
              return TRUE;
            } else {
              thread->wake_time = ticks + thread->check_freq;
              return FALSE;
            }
            else
              return FALSE;

          break;
          
        case CUk_UntilFctFALSE:
        
          if (ticks >= thread->wake_time)
            if (thread->blocked_fct(thread->user_data, 
                            thread->blocked_params) == FALSE) {
              thread->status = CUk_Ready;
              if (thread->local_peer) {
                thread->local_peer = FALSE;
                CUg_ThreadControl.num_local_peers--;
                ASSERT(CUg_ThreadControl.num_local_peers >= 0);
              }
              CU_WakeUpProcessYielder();
              return TRUE;
            } else {
              thread->wake_time = ticks + thread->check_freq;
              return FALSE;
            }
            else
              return FALSE;

          break;
          
        default:
          BARF;
      }
      
      /* Above case statement should always return */
      BARF;
    
    /* All other thread states imply it is not ready */
    default:
      return FALSE;
  }
}


/****************************************************************************
* FUNCTION:
*
* CU_NextScheduledThreadEntry
*
* DESCRIPTION:
*
* Find next thread to be run after the current thread.
*
* RETURNS:
*
* CUt_ThreadEntry *    -- The thread entry.
*
* NOTES:
*
* Each thread in is run in a round-robin manner, skipping over
* any that are delayed or blocked.  If no threads are ready
* to run, then NULL is returned.
*
*/

CUt_ThreadEntry *CU_NextScheduledThreadEntry()
{
  register CUt_ThreadEntry *current;
  register CUt_ThreadEntry *next;
  register tUInt4 ticks;
  
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(CUg_ThreadControl.current != NULL);
  
  /* Look for the next thread that is ready to run */
#ifdef macintosh  /* For optimization */
  ticks = LMGetTicks();
#else
  ticks = CU_CurrentTicks();
#endif
  current = CUg_ThreadControl.current;
  next = current->next_entry;
  while ((next != current) && !CU_ThreadEntryReady(next, ticks))
    next = next->next_entry;
    
  /* If no threads are ready to run, return NULL */
  if ( (next == current) && !CU_ThreadEntryReady(next, ticks) ) 
    next = NULL;

  return(next);
}


/****************************************************************************
* NOTE: CU_StartThreadEntry is maintained for each running environment because
* it is too nasty to read otherwise!
*
* For future ports, could probably use one or the other depending on
* whether a threads package is being used or whether setjmp() and longjmp()
* are used.
****************************************************************************/

#ifdef SETJMPTHREADS
/****************************************************************************
* FUNCTION:
*
* CU_StartThreadEntry
*
* DESCRIPTION:
*
* Switch context and start the given thread processing.  Does nothing
* if the given thread is already running or is blocked or sleeping.
*
* PARAMETERS:
*
* CUt_ThreadEntry *    -- The thread entry.
*
* RETURNS:
* 
* tBoolean             -- TRUE when thread was run (by the time this returns
*                         it has already run) and FALSE when it is not
*                         ready, is sleeping, or blocked.  This function
*                         also returns FALSE when passed the current thread.
*
*/

tBoolean CU_StartThreadEntry(CUt_ThreadEntry * thread)
{  
  tUInt4 curr_ticks;

  ASSERT(thread != NULL);
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(CUg_ThreadControl.current != NULL);

#ifdef macintosh
  curr_ticks = LMGetTicks();
#else
  curr_ticks = CU_CurrentTicks();
#endif
  if (CU_ThreadEntryReady(thread, curr_ticks)) {
    if (setjmp(CUg_ThreadControl.current->env.setjmp_env) == CUk_RunThread) {
    
      /* 
      * Arrive here when a longjmp() has been called to switch in a
      * new thread.  So we restore the rest of the thread's context.
      */
      
      CU_RestoreEnvironment();
      
      return TRUE;
    }
    
    else {
    
      /* 
      * Arrive here when the above setjmp() has saved the current thread's
      * environment in preparation for running the next thread.
      */
#ifdef MS_WIN 
  
      if ( ((tUInt2)CUg_ThreadControl.current->env.stack + (tUInt4)CU_DefaultStackSize() - (tUInt4)sizeof(long)) > 0x0000FFFF ) {
        ASSERT(0);
      }
//      CUg_ThreadControl.current->env.setjmp_env[sp] = (int)((long)CUg_ThreadControl.current->env.stack + (long)CU_DefaultStackSize() - (long)sizeof(long));
//      CUg_ThreadControl.current->env.setjmp_env[0] = CUg_ThreadControl.current->env.env0; 

/*       thread->env.setjmp_env[pc] = thread->env.pc0;
      thread->env.setjmp_env[pc+1] = thread->env.pc1;  
*/
/*  *((CUt_ThreadProc *) &(thread->env.setjmp_env[pc])) = proc; */
/* CUg_ThreadControl.current->env.setjmp_env[ds] = (int) ((long) ((char far *) CUg_ThreadControl.current) >> 16); */

#endif /* MS_WIN */

      /* Save rest of environment for current thread */
      CU_SaveEnvironment();
      
      /* Reset status of current thread */
      if (CUg_ThreadControl.current->status == CUk_Running)
        CUg_ThreadControl.current->status = CUk_Ready;

      /* Set new thread and jump to its environment */
      CUg_ThreadControl.current = thread;
      thread->status = CUk_Running;
      longjmp(thread->env.setjmp_env, CUk_RunThread);
      
      /* longjmp() never returns ... it is a goto! */
      ASSERT(0);
    }
  }
  
  return FALSE;
}
#endif /* SETJMPTHREADS */


#ifdef SUNLWP
/****************************************************************************
* FUNCTION:
*
* CU_StartThreadEntry
*
* DESCRIPTION:
*
* Switch context and start the given thread processing.  Does nothing
* if the given thread is already running or is blocked or sleeping.
*
* PARAMETERS:
*
* CUt_ThreadEntry *    -- The thread entry.
*
* RETURNS:
* 
* tBoolean             -- TRUE when thread was run (by the time this returns
*                         it has already run) and FALSE when it is not
*                         ready, is sleeping, or blocked.  This function
*                         also returns FALSE when passed the current thread.
*
*/

tBoolean CU_StartThreadEntry(CUt_ThreadEntry * thread)
{  
  ASSERT(thread != NULL);
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(CUg_ThreadControl.current != NULL);

  if (CU_ThreadEntryReady(thread, CU_CurrentTicks())) {
    /* Save environment for current thread */
    CU_SaveEnvironment();
    
    /* Reset status of current thread */
    if (CUg_ThreadControl.current->status == CUk_Running)
      CUg_ThreadControl.current->status = CUk_Ready;

    /* Set new thread running status */
    CUg_ThreadControl.current = thread;
    thread->status = CUk_Running;
    
    /* Yield time to the selected thread */
    if (lwp_yield(thread->lwp_ID) != 0)
      UNIMPLEMENTED;
      
    /* Restore environment to continue executing this thread */
    CU_RestoreEnvironment();
    
    /* Success */
    return TRUE;
  }
  
  return FALSE;
}
#endif /* SUNLWP */


#if QUICKTHREADS
/****************************************************************************
* FUNCTION:
*
* CU_StartThreadEntry
*
* DESCRIPTION:
*
* Switch context and start the given thread processing.  Does nothing
* if the given thread is already running or is blocked or sleeping.
*
* PARAMETERS:
*
* CUt_ThreadEntry *    -- The thread entry.
*
* RETURNS:
* 
* tBoolean             -- TRUE when thread was run (by the time this returns
*                         it has already run) and FALSE when it is not
*                         ready, is sleeping, or blocked.  This function
*                         also returns FALSE when passed the current thread.
*
* NOTES:
*
* Note that the master thread is treated specially by marking
* the current stack frame location with master_stack and setting the
* qthread and env.stack values before yielding.
*
*/

/*
* Local yield function required by QuickThreads ... in our case, it just
* saves the stack information for the previously run thread so it can be
* restored later.
*/
#ifdef QUICKTHREADS
static void *
CU_QTYieldHelp(qt_t *sp, void *old, void *unused)
{
  ((CUt_ThreadEntry *) old)->qthread = sp;
  return NULL;
}
#endif

tBoolean CU_StartThreadEntry(CUt_ThreadEntry * thread)
{ 
  CUt_ThreadEntry *prev_thread;

  ASSERT(thread != NULL);
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(CUg_ThreadControl.current != NULL);

  if (CU_ThreadEntryReady(thread, CU_CurrentTicks())) {
    /* Save environment for current thread */
    CU_SaveEnvironment();
    
    /* Reset status of current thread */
    if (CUg_ThreadControl.current->status == CUk_Running)
      CUg_ThreadControl.current->status = CUk_Ready;

    /* Set new thread running status */
    prev_thread = CUg_ThreadControl.current;
    CUg_ThreadControl.current = thread;
    thread->status = CUk_Running;

    /* Yield time to the selected thread */
    QT_BLOCK(CU_QTYieldHelp, prev_thread, NULL, thread->qthread);
      
    /* Restore environment to continue executing this thread */
    CU_RestoreEnvironment();
    
    /* Success */
    return TRUE;
  }
  
  return FALSE;
}
#endif /* QUICKTHREADS */


#ifdef MACTHREADMANAGER
/****************************************************************************
* FUNCTION:
*
* CU_StartThreadEntry
*
* DESCRIPTION:
*
* Switch context and start the given thread processing.  Does nothing
* if the given thread is already running or is blocked or sleeping.
*
* PARAMETERS:
*
* CUt_ThreadEntry *    -- The thread entry.
*
* RETURNS:
* 
* tBoolean             -- TRUE when thread was run (by the time this returns
*                         it has already run) and FALSE when it is not
*                         ready, is sleeping, or blocked.  This function
*                         also returns FALSE when passed the current thread.
*
*/

tBoolean CU_StartThreadEntry(CUt_ThreadEntry * thread)
{  
  ASSERT(thread != NULL);
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(CUg_ThreadControl.current != NULL);

  if (CU_ThreadEntryReady(thread, CU_CurrentTicks())) {
    /* Save environment for current thread */
    CU_SaveEnvironment();
    
    /* Reset status of current thread */
    if (CUg_ThreadControl.current->status == CUk_Running)
      CUg_ThreadControl.current->status = CUk_Ready;

    /* Set new thread running status */
    CUg_ThreadControl.current = thread;
    thread->status = CUk_Running;
    
    /* Yield time to the selected thread */
    if (YieldToThread(thread->tm_ID) != noErr)
      UNIMPLEMENTED;
      
    /* Restore environment to continue executing this thread */
    CU_RestoreEnvironment();
    
    /* Success */
    return TRUE;
  }
  
  return FALSE;
}
#endif /* MACTHREADMANAGER */


#ifdef WIN32THREADS
/****************************************************************************
* FUNCTION:
*
* CU_StartThreadEntry
*
* DESCRIPTION:
*
* Switch context and start the given thread processing.  Does nothing
* if the given thread is already running or is blocked or sleeping.
*
* PARAMETERS:
*
* CUt_ThreadEntry *    -- The thread entry.
*
* RETURNS:
* 
* tBoolean             -- TRUE when thread was run (by the time this returns
*                         it has already run) and FALSE when it is not
*                         ready, is sleeping, or blocked.  This function
*                         also returns FALSE when passed the current thread.
*
*/

tBoolean CU_StartThreadEntry(CUt_ThreadEntry * thread)
{
  DWORD ret;
  CUt_ThreadEntry *old;

  ASSERT(thread != NULL);
  ASSERT(CUg_ThreadControl.master != NULL);
  ASSERT(CUg_ThreadControl.current != NULL);

  if (CU_ThreadEntryReady(thread, CU_CurrentTicks())) {

    CU_StartWin32Thread(thread);

    /* Success */
    return TRUE;
  }

  return FALSE;
}
#endif /* WIN32 */


#ifdef WIN32THREADS
/***************************************************************************
* 
* FUNCTION: CU_StartWin32Thread
*
* DESCRIPTION: 
*
* Handles the switching from one win32 thread to another and/or deleting
* a win32 thread.  Deletion is handled here because deletion can only be done
* safely from within the thread being deleted (Why? because that's the way
* the runtime libs are written).
* 
* PARAMETERS: 
*
* CUt_ThreadEntry* start_thread -- Thread to start.
* tBoolean block_this           -- Whether this thread should 
* 
* RETURNS: nothing
* 
* NOTES: 
*
* Because win32 threads are realy preemptive, more than one thread may be running 
* concurrently in this function between the time we start another and begin to 
* block the current one.
* 
*/

void
CU_StartWin32Thread(CUt_ThreadEntry* start_thread)
{
  CUt_ThreadEntry* this_thread;
  DWORD ret;

  GOODMEM(start_thread);
  ASSERT(start_thread->status != CUk_Destroyed);

  /* Get pointer to the thread we're in */
  this_thread = CUg_ThreadControl.current;

  /* Reset status of current thread */
  if (this_thread->status == CUk_Running)
    this_thread->status = CUk_Ready;

  /*
   * Start another thread by changing which event win32 is signaled within a win32 critical section.
   * The critical section ensures that only one thread is doing this at
   * the same time.  Once the event is signaled, the thread will start.
   */

  /* Begin critical section */
  EnterCriticalSection(&CUg_Win32ChangeEventSection);

  /* Start the win32 thread, either through ResumeThread if it has never run
     or by signalling its event */
  if ( start_thread->suspended ) {
    /* printf("starting start_thread #%d\n", start_thread->ID);*/
    ret = ResumeThread(start_thread->win32_thread);
    ASSERT(ret == 1);
    start_thread->suspended = FALSE;
  }
  else {
    /* printf("resuming start_thread #%d\n", start_thread->ID);*/
    ret = SetEvent(start_thread->win32_event);
    ASSERT(ret == TRUE);
  }

  /* If the thread that we're starting is not being destroyed, set the event 
     for this thread to unsignaled so that it blocks until another thread
     signals it to start */
  ret = ResetEvent(this_thread->win32_event);
  ASSERT(ret == TRUE);

  /* End critical section */
  LeaveCriticalSection(&CUg_Win32ChangeEventSection);

  /*
   * Block this thread until some other thread signals us to to start. 
   * Note that even if another thread has already
   * signaled this thread to start we're OK; we just won't wait at all
   */
  ret = WaitForSingleObject(this_thread->win32_event, INFINITE);
  ASSERT(ret == WAIT_OBJECT_0);

  /*
   * The thread switch has now been completed and this thread now either has
   * to return to continue running or destroy itself
   */

  /* If this thread has been destroyed, terminate it */
  if ( this_thread->status == CUk_Destroyed ) {
    /* printf("destroying thread #%d\n", old->ID);*/
#ifdef _BORLANDC_
    _endthread();
#endif
#ifdef MS_VC
    _endthreadex(0);
#endif
  }

  /* Destroy any thread marked for destruction */
  if ( CUg_ThreadControl.destroy != NULL ) {
    CU_DestroyWin32Thread(CUg_ThreadControl.destroy);
    ASSERT(CUg_ThreadControl.destroy->env.stack != NULL);
    CU_Free(CUg_ThreadControl.destroy);
    CUg_ThreadControl.destroy = NULL;
  }

  /* Set CU flags to indicate this thread is current and running */
  CUg_ThreadControl.current = this_thread;
  this_thread->status = CUk_Running;
}


/***************************************************************************
* 
* FUNCTION: CU_DestroyWin32Thread
*
* DESCRIPTION: 
*
* Destroys the native win32 thread.  Must be called from another thread when
* the thread to be destroyed is stopped.  Will not return until thread is 
* destroyed.  
* 
* PARAMETERS: 
*
* CUt_ThreadEntry* thread -- Thread to destroy
* 
* RETURNS: nothing
* 
* NOTES: 
* 
* Currently can not destroy the master thread
*
*/

void
CU_DestroyWin32Thread(CUt_ThreadEntry* thread)
{
  DWORD ret;

  ASSERT(thread != CUg_ThreadControl.current);
  ASSERT(thread != CUg_ThreadControl.master);

  thread->status = CUk_Destroyed;

  /* Start the win32 thread, either through ResumeThread if it has never run
     or by signalling its event */
  if ( thread->suspended ) {
    ret = ResumeThread(thread->win32_thread);
    ASSERT(ret == 1);
    thread->suspended = FALSE;
  }
  else {
    /* printf("resuming start_thread #%d\n", start_thread->ID);*/
    ret = SetEvent(thread->win32_event);
    ASSERT(ret == TRUE);
  }
  
  /* Block until thread has been destroyed.  Note that we don't check the return
     value here because this will return a failure
     code if the thread has already been destroyed */
  WaitForSingleObject(thread->win32_thread, INFINITE);
}
#endif /* WIN32 */


#ifdef SUNLWP
/****************************************************************************
* FUNCTION:
*
* CU_SunThreadProc
*
* DESCRIPTION:
*
* Wrapper for regular thread proc for use with the Sun OS lightweight
* process library (LWP).
*
* PARAMETERS:
*
* CUt_ThreadProc     -- The procedure for the thread.
* void *             -- User data value for the thread.
* CUt_ThreadEntry *  -- The thread's thread table entry.
*
*/

void CU_SunThreadProc(CUt_ThreadProc proc, void *data, CUt_ThreadEntry *thread)
{
  GOODMEM(data);
  GOODMEM(thread);
  
  /* Restore environment for this thread, run the thread, and destroy */
  /* it when done. */
  CU_RestoreEnvironment();
  proc(data);
  CU_DestroyThreadEntry(thread);
}
#endif /* SUNLWP */


#ifdef MACTHREADMANAGER
/****************************************************************************
* FUNCTION:
*
* CU_MacThreadProc
*
* DESCRIPTION:
*
* Wrapper for regular thread proc for use with the Mac ThreadManager.
*
* PARAMETERS:
*
* CUt_MacThreadProcData *   -- The procedure, data, and thread table entry
*                              for the thread; in struct format.
*
*/

pascal void *CU_MacThreadProc(CUt_MacThreadProcData *data)
{
  GOODMEM(data);
  
  /* Restore environment for this thread, run the thread, and destroy */
  /* it when done. */
  CU_RestoreEnvironment();
  data->proc(data->data);
  CU_Free(data->data);
  CU_DestroyThreadEntry(data->thread);
  return NULL;
}
#endif /* MACTHREADMANAGER */


#endif /* CUk_NOTHREADS */


/* end of CU_Thread.c */
