/***************************************************************************
* 
* Copyright (c) 1998, Decision Systems Group
*
* Permission is hereby granted, free of charge, to any person obtaining
* a copy of this software and associated documentation files (the
* "Software"), to deal in the Software without restriction, including
* without limitation the rights to use, copy, modify, merge, publish,
* distribute, sublicense, and/or sell copies of the Software, and to
* permit persons to whom the Software is furnished to do so, subject to
* the following conditions:
*
* The above copyright notice and this permission notice shall be
* included in all copies or substantial portions of the Software.
*
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
* EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
* MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
* IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
* CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
* TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
* SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
* 
* ==========================================================================
* 
* FILE: CU_Tree.c
* 
* AUTHOR: 
*
* Jonathan Traum
* Stephan R.A. Deibel
* YongJoon Lee
* Karen Dintino
* 
* CREATION DATE: 
*
* June 1994
*
* DESCRIPTION: 
*
* Generic tree implementation
* 
* NOTES: 
* 
* 
* MODIFICATIONS: 
* --------------------------------------------------------------------------
* Date     Name      Description of modification
* --------------------------------------------------------------------------
*
* $Log: CU_Tree.c,v $
* Revision 1.2  1998/07/15 05:20:23  sdeibel
* Changed all copyright notices to match MIT license release plus misc
* prep for 0.8.3 release
*
 * Revision 1.1.1.1  1997/10/24  16:48:44  sdeibel
 * Converting to CVS
 *
* Revision 1.8  1997/02/18 21:38:48  jrequena
* more sibling support
*
* Revision 1.7  1997/01/15  21:08:27  jrequena
* Added sibling support.
*
* Revision 1.6  1996/12/18  15:38:23  jrosenbe
* Ditched ^L & ^J at start of file
*
* Revision 1.5  1996/06/18  21:05:14  epattiso
* made changes to compile on Mac
*
* Revision 1.4  1996/06/03  03:53:43  sdeibel
* Clean up / added functions
* 
* 
*/

/**************************************************************************/
/* INCLUDES */

#ifdef dos 
#include "CUBasics.h"
#include "CUDebug.h"
#include "CUTree.h"
#else 
#include "CU_Basics.h"
#include "CU_Debug.h"
#include "CU_Tree.h"
#endif /* dos */


/***************************************************************************
* 
* FUNCTION: 
* 
* CU_TN_Create()
* 
* DESCRIPTION: 
*
* Create an empty tree node
* 
* PARAMETERS: 
* 
* None.
* 
* RETURNS: 
* 
* CUt_TreeNode *               -- The new node; or NULL if out of memory.
* 
*/

CUt_TreeNode*  
CU_TN_Create()
{
  CUt_TreeNode *retval;

  /* Allocate */
  retval = (CUt_TreeNode*) CU_Calloc(1, sizeof(CUt_TreeNode));
  if (retval == NULL)
    return NULL;

  /* Clear */
  retval->parent = NULL;
  retval->first_child = NULL;
  retval->prev_sib = NULL;
  retval->next_sib = NULL;
  retval->data = NULL;

  /* Done */
  return retval;
}


/***************************************************************************
* 
* FUNCTION: 
*
* CU_TN_Destroy()
* 
* DESCRIPTION: 
*
* Free up the memory occupied by a tree node
* 
* PARAMETERS: 
* 
* CUt_TreeNode *                -- The node.
* 
* NOTES: 
*
* This function does NOT alter the tree of which this node might be an
* element. In order to remove a node from a tree, you need to call
* CU_Tree_RemoveNode() first.
*
* This function does not free data in the "data" field of CUt_TreeNode.
* 
*/

void  
CU_TN_Destroy(CUt_TreeNode* node)
{
  GOODMEM(node);
  CU_Free(node);
}


/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_AddSibling()
* 
* DESCRIPTION: 
*
* Add new_node to the tree as the next sibling of the given node
* 
* PARAMETERS: 
* 
* CUt_TreeNode *               -- The node to add
* CUt_TreeNode *               -- The sibling 
* 
* RETURNS: 
* 
* void
* 
*/

void  
CU_Tree_AddSibling(CUt_TreeNode *new_node, CUt_TreeNode *sibling)
{
  /* Error check */
  GOODMEM(sibling);
  GOODMEM(new_node);

  /* Insert the new node as sibling of given node */
  new_node->next_sib = sibling->next_sib;
  new_node->prev_sib = sibling;
  sibling->next_sib = new_node;
  new_node->parent = sibling->parent;

  /* Done */
  return;
}

/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_AppendChild()
* 
* DESCRIPTION: 
*
* Add a child to the end of the given node's child-list
* 
* PARAMETERS: 
* 
* CUt_TreeNode *               -- The node to add
* CUt_TreeNode *               -- The parent to add to
* 
* RETURNS: 
* 
* void
* 
*/

void  
CU_Tree_AppendChild(CUt_TreeNode *new_node, CUt_TreeNode *parent)
{
  CUt_TreeNode* last_child;

  /* Error check */
  GOODMEM(parent);
  GOODMEM(new_node);

  /* Insert node as first child */
  if (parent->first_child == NULL) {
    new_node->parent = parent;
    parent->first_child = new_node;
    new_node->next_sib = NULL;
    new_node->prev_sib = NULL;
    return;
  }

  /* Insert child as Nth child */
  else {
    new_node->parent = parent;
    for (last_child = parent->first_child; last_child->next_sib != NULL; 
         last_child = last_child->next_sib)
      ;
    last_child->next_sib = new_node;
    new_node->prev_sib = last_child;
    new_node->next_sib = NULL;
  }

  /* Done */
  return;
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CU_Tree_PrependChild()
* 
* DESCRIPTION: 
* 
* Add a new child as the given node's first child
* 
* PARAMETERS: 
* 
* CUt_TreeNode *new_node     - the new node to add
* CUt_TreeNode *parent       - The newly assigned parent for new_node. 
* 
* RETURNS: 
* 
* void
* 
*/


void  
CU_Tree_PrependChild(CUt_TreeNode *new_node, CUt_TreeNode *parent)
{
  /* Error check */
  GOODMEM(new_node);
  GOODMEM(parent);
  
  /* Prepend */
  if (parent->first_child != NULL)
    parent->first_child->prev_sib = new_node;
  new_node->prev_sib = NULL;
  new_node->next_sib = parent->first_child;
  parent->first_child = new_node;
  new_node->parent = parent;

  /* Done */
  return;    
}

/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_MoveRight()
* 
* DESCRIPTION: 
*
* Move a node and it's subtree to the right, i.e. make it the last child of its
* prev_sib, which should be non-null.
*
* PARAMETERS: 
* 
* CUt_TreeNode *               -- The node
* 
* RETURNS: 
* 
* void
* 
*/

void  
CU_Tree_MoveRight(CUt_TreeNode *node)
{
  CUt_TreeNode *new_parent;
  CUt_TreeNode *new_sib;

  /* Error check */
  GOODMEM(node);

  /* The previous sibling for the node is going to be its new parent */
  new_parent = node->prev_sib;
  if (new_parent == NULL)
    return;

  /* Remove node from current position in tree */
  CU_Tree_RemoveNode(node);

  /* Find new previous sibling node, if any */
  new_sib = new_parent->first_child;
  if (new_sib != NULL)
    while (new_sib->next_sib != NULL)
      new_sib = new_sib->next_sib;

  /* Move the node to new position - if found a sibling, add node after it or */
  /* use AddChild when new parent does not yet have any children. */
  if (new_sib != NULL)
    CU_Tree_AddSibling(node, new_sib);
  else
    CU_Tree_PrependChild(node, new_parent);

  /* Done */
  return;
}

/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_MoveLeft()
* 
* DESCRIPTION: 
*
* Move the given node and its subtree to the left, i.e. make it the next sibling of
* itsParent (which should be non-null)
*
* PARAMETERS: 
* 
* CUt_TreeNode *               -- The node
* 
* RETURNS: 
* 
* void
* 
*/

void  
CU_Tree_MoveLeft(CUt_TreeNode *node)
{
  CUt_TreeNode* its_parent;

  /* Error check */
  GOODMEM(node);

  /* Make sure that this node is not already at root level */
  if (node->parent == NULL)
    return;

  /* Find parent of this node */
  its_parent = node->parent;

  /* Remove node from current position in tree */
  CU_Tree_RemoveNode(node);

  /* Place the node into its new position */
  CU_Tree_AddSibling(node, its_parent);

  /* Done */
  return;
}

/***************************************************************************
* 
* FUNCTION: 
* 
* CU_Tree_InsertChildAfter()
* 
* DESCRIPTION: 
* 
* Add a new child after the given sibling node
* 
* PARAMETERS: 
* 
* CUt_TreeNode *                 -- The new node to add
* CUt_TreeNode *                 -- The sibling after which to add;
*                                   must have a valid parent
* 
* RETURNS: 
* 
* void
* 
*/


void  
CU_Tree_InsertChildAfter(CUt_TreeNode *new_node, CUt_TreeNode *prev_sibling)
{
  /* Error check */
  GOODMEM(new_node);
  GOODMEM(prev_sibling);
  GOODMEM(prev_sibling->parent);

  /* Insert */
  if (prev_sibling->next_sib != NULL)
    prev_sibling->next_sib->prev_sib = new_node;
  new_node->next_sib = prev_sibling->next_sib;
  prev_sibling->next_sib = new_node;
  new_node->parent = prev_sibling->parent;

  /* Done */
  return;
}


/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_RemoveNode()
* 
* DESCRIPTION: 
*
* Remove a node and its descendents from its tree.
* 
* PARAMETERS: 
*
* CUt_TreeNode *                -- The node to remove
* 
* RETURNS: 
* 
* void
* 
* NOTES: 
* 
* Must call CU_TN_Destroy() to free the node memory.
* 
*/

void
CU_Tree_RemoveNode(CUt_TreeNode *the_node)
{
  GOODMEM(the_node);

  /* Remove from sibling list */
  if (the_node->prev_sib != NULL)
    the_node->prev_sib->next_sib = the_node->next_sib;
  if (the_node->next_sib != NULL)
    the_node->next_sib->prev_sib = the_node->prev_sib;

  /* Remove from parent/child list */
  if ((the_node->parent != NULL) && 
      (the_node->parent->first_child == the_node))
    the_node->parent->first_child = the_node->next_sib;

  /* Blank out node's connection pointers */
  the_node->parent = NULL;
  the_node->prev_sib = NULL;
  the_node->next_sib = NULL;

  /* Done */
  return;
}


/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_IsDescendentOf
* 
* DESCRIPTION: 
*
* Find out whether a node is a descendent of another node
* 
* PARAMETERS: 
*
* CUt_TreeNode *               -- The node to check for
* CUt_TreeNode *               -- The potential ancestor
* 
* RETURNS: 
*
* tBoolean                     -- TRUE iff node is a descendent of
*                                 given ancestor
* 
*/

tBoolean
CU_Tree_IsDescendentOf(CUt_TreeNode *the_node, CUt_TreeNode *ancestor)
{
  CUt_TreeNode *curr_node;
  
  /* Look for ancestory */
  for (curr_node = the_node; curr_node != NULL; 
       curr_node = curr_node->parent) {
    if (curr_node == ancestor)
      return TRUE;    
  }

  /* Not found */
  return FALSE;
  
}

/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_IsSiblingOf
* 
* DESCRIPTION: 
*
* Returns TRUE if the_node is a sibling of root_node.  Nodes are considered to be
* siblings of themselves.
*
* PARAMETERS: 
*
* CUt_TreeNode *               -- The node to check for
* CUt_TreeNode *               -- The potential sibling
* 
* RETURNS: 
*
* tBoolean                     -- TRUE if node is a sibling of
*                                 given root_node
* 
*/

tBoolean
CU_Tree_IsSiblingOf(CUt_TreeNode *the_node, CUt_TreeNode *root_node)
{
  CUt_TreeNode *curr_node;
  
  /* Check before rootNode */
  curr_node = root_node;
  while (curr_node != NULL && curr_node != the_node)
    curr_node = curr_node->prev_sib;
  if (curr_node == the_node)
    return TRUE;

  /* Check after rootNode */
  curr_node = root_node->next_sib;
  while (curr_node != NULL && curr_node != the_node)
    curr_node = curr_node->next_sib;
  if (curr_node == the_node)
    return TRUE;

  /* Not found */
  return FALSE;
  
}


/***************************************************************************
* 
* FUNCTION: 
* 
* CU_Tree_GetNodeDepth()
* 
* DESCRIPTION: 
* 
* Compute the number of generations from the root to the given node
*
* PARAMETERS: 
* 
* CUt_TreeNode *                   -- The node.
*
* RETURNS: 
*
* tUInt4                           -- The depth (root = 1).
* 
*/

tUInt4
CU_Tree_GetNodeDepth(CUt_TreeNode *the_node)
{
  tUInt4 depth = 0;
  CUt_TreeNode *cur_node;

  GOODMEM(the_node);

  /* Count depth */
  for (cur_node = the_node; cur_node != NULL; cur_node = cur_node->parent)
    ++depth;

  /* Done */
  return depth;
}


/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_TraverseChildren()
* 
* DESCRIPTION: 
*
* Traverse the children of the given node
* 
* PARAMETERS: 
*
* CUt_TreeNode *                  -- The parent
* CUt_TraversalFn                 -- The traversal function (should
*                                    return TRUE to halt traversal)
* void *                          -- User data to be sent to traversal fct
* 
* RETURNS: 
*
* CUt_TreeNode *                  -- Node selected by the traversal
*                                    function or NULL if end of tree
*                                    was reached.
*
* NOTES:
*
* The traversal function may remove and delete the node it is sent,
* if this is desired.
* 
*/

CUt_TreeNode *
CU_Tree_TraverseChildren(CUt_TreeNode *parent, CUt_TraversalFn Action, 
                         void* data)
{
  CUt_TreeNode *node, *nextNode;

  /* Error check */
  GOODMEM(parent);
  
  /* Traverse siblings */
  for (node = parent->first_child; node != NULL; node = nextNode) {

    /* Get next node now so action may delete the current node */
    nextNode = node->next_sib;

    /* Do action with current node */
    if (Action(node, data))
      return node;
  }

  /* Reached end of tree */
  return NULL;
}

/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_TraverseSiblings()
* 
* DESCRIPTION: 
*
* Traverse the siblings of the given node
* 
* PARAMETERS: 
*
* CUt_TreeNode *                  -- The node
* CUt_TraversalFn                 -- The traversal function (should
*                                    return TRUE to halt traversal)
* void *                          -- User data to be sent to traversal fct
* 
* RETURNS: 
*
* CUt_TreeNode *                  -- Node selected by the traversal
*                                    function or NULL if end of tree
*                                    was reached.
*
* NOTES:
*
* The traversal function may remove and delete the node it is sent,
* if this is desired.
* 
*/

CUt_TreeNode *
CU_Tree_TraverseSiblings(CUt_TreeNode *node, CUt_TraversalFn Action, 
                         void* data)
{
  CUt_TreeNode *nextNode;

  /* Error check */
  GOODMEM(node);
  
  /* Traverse siblings */
  while (node != NULL) {
    /* Get next node now so action may delete the current node */
    nextNode = node->next_sib;

    /* Do action with current node */
    if (Action(node, data))
      return node;

    node = nextNode;
  }

  /* Reached end of tree */
  return NULL;
}

/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_TraverseInOrder()
* 
* DESCRIPTION: 
* 
* Traverse tree recursively with "in order" traversal.
* 
* PARAMETERS: 
*
* CUt_TreeNode *                -- The root/subtree node
* CUt_TraversalFn               -- The traversal function (should
*                                  return TRUE to halt traversal)
* void *                        -- User data to be sent to traversal fct
* 
* RETURNS: 
* 
* CUt_TreeNode *                -- Node selected by the traversal
*                                  function or NULL if end of tree
*                                  was reached.
*
* NOTES: 
* 
* The traversal function may remove and delete the node it is sent,
* if this is desired.
* 
*/

CUt_TreeNode *  
CU_Tree_TraverseInOrder(CUt_TreeNode *root, CUt_TraversalFn Action, void *data)
{
  CUt_TreeNode *thisNode, *nextNode, *retNode;

  /* Check if at end of traversal */
  if (root == NULL) 
    return NULL;

  /* Hang onto next node so the action may delete current node */
  thisNode = root->first_child;

  /* Perform action before continuing traversal; stop if it returns TRUE */
  if (Action(root, data))
    return root;

  /* Traverse siblings at this level */
  while (thisNode != NULL)
  {
    /* Hang onto next node so that the action may delete current node */
    nextNode = thisNode->next_sib;

    /* Recurse to do subtree belonging to this sibling */
    retNode = CU_Tree_TraverseInOrder(thisNode, Action, data);
    if (retNode != NULL) 
      return retNode;
    
    /* Next sibling */
    thisNode = nextNode;
  }

  /* Reached end of tree */
  return NULL;
}


/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_TraverseGuided
* 
* DESCRIPTION: 
* 
* Traverse the tree, moving downward as guided by the traversal
* function
* 
* PARAMETERS: 
*
* CUt_TreeNode *           -- The root/subtree node
* CUt_GuideFn              -- The guide function; selects
*                             a child to move to for each node it is
*                             passed; or returns NULL to stop traversal
*                             at current node.
* void *                   -- Data to be passed to the guide function
* 
* RETURNS: 
* 
* CUt_TreeNode *           -- The node at which the guide function
*                             returned NULL
* 
* NOTES:
*
* The guide function may remove and delete the given node, so long as it
* returns a valid node to continue with.
*
*/

CUt_TreeNode *  
CU_Tree_TraverseGuided(CUt_TreeNode *root, CUt_GuideFn Guide, void *data)
{
  CUt_TreeNode *thisNode, *prevNode;

  GOODMEM(root);

  /* Set up */
  prevNode = NULL;
  thisNode = root;

  /* Traverse */
  while (thisNode != NULL) {
    prevNode = thisNode;
    thisNode = Guide(thisNode, data);
  }

  /* Done */
  return prevNode;
}


/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_NextInOrderNode
* 
* DESCRIPTION: 
* 
* Return the node that would come after the given node in an
* in-order tree traversal.
* 
* PARAMETERS: 
*
* CUt_TreeNode *           -- The node.
* 
* RETURNS: 
* 
* CUt_TreeNode *           -- The next node in order.
* 
*/

CUt_TreeNode *  
CU_Tree_NextInOrderNode(CUt_TreeNode *node)
{
  CUt_TreeNode *tmp_node;

  /* Error check */
  GOODMEM(node);

  /* Go to first child first */
  if (node->first_child != NULL)
    return node->first_child;

  /* If no children, try next sibling */
  else if (node->next_sib != NULL)
    return node->next_sib;

  /* If no next sibling, try nearest ancestor's next sibling */
  else {
    tmp_node = node->parent;
    while (tmp_node != NULL && tmp_node->next_sib == NULL)
      tmp_node = tmp_node->parent;
    if (tmp_node != NULL)
      return tmp_node->next_sib;
    else
      return NULL;
  }
}

/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_PrevInOrderNode
* 
* DESCRIPTION: 
* 
* Return the node that would come before the given node in an
* in-order tree traversal.
* 
* PARAMETERS: 
*
* CUt_TreeNode *           -- The node.
* 
* RETURNS: 
* 
* CUt_TreeNode *           -- The previous node in order.
* 
*/

CUt_TreeNode *  
CU_Tree_PrevInOrderNode(CUt_TreeNode *node)
{
  CUt_TreeNode *tmp_node;

  /* Error check */
  GOODMEM(node);

  /* If no previous siblings, then parent is previous */
  if (node->prev_sib == NULL)
    return node->parent;

  /* else, last in subtree of previous sibling is previous */
  else {
    tmp_node = node->prev_sib;
    if (tmp_node != NULL && tmp_node->first_child != NULL)
      return CU_Tree_LastInSubTree(tmp_node->first_child);
    else
      return tmp_node;
  }
}

/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_LastSibling()
* 
* DESCRIPTION: 
*
* Return the last sibling of the given node, or the node itself if it has
* no siblings after itself.
*
* PARAMETERS: 
* 
* CUt_TreeNode *               -- The node
* 
* RETURNS: 
* 
* CUt_TreeNode*
* 
*/

CUt_TreeNode*
CU_Tree_LastSibling(CUt_TreeNode *node)
{
  CUt_TreeNode* sibling;

  /* Error check */
  GOODMEM(node);

  sibling = node;
  while (sibling->next_sib != NULL)
    sibling = sibling->next_sib;
  /* Done */
  return sibling;
}

/***************************************************************************
* 
* FUNCTION: 
*
* CU_Tree_LastInSubTree()
* 
* DESCRIPTION: 
*
* Return the last (inorder) node in the subTree that is defined by the given
* node and its siblings.  Returns the root_node itself if the given node has
* no children or siblings.
*
* PARAMETERS: 
* 
* CUt_TreeNode *               -- The node
* 
* RETURNS: 
* 
* CUt_TreeNode*
* 
*/

CUt_TreeNode*
CU_Tree_LastInSubTree(CUt_TreeNode *root_node)
{
  CUt_TreeNode* last_node;

  /* Error check */
  GOODMEM(root_node);

  last_node = CU_Tree_LastSibling(root_node);
  if (last_node->first_child != NULL)
    last_node = CU_Tree_LastInSubTree(last_node->first_child);

  /* Done */
  return last_node;
}

/* end of CU_Tree.c */
